/* cm_PR.c */
/* rational polynomial routines used in cmR.c.*/

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

extern MPR *R[];
extern POLYR PR[];

void ADD_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\n+\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp = PR[l];
	PR[l] = ADDPR(PR[j], PR[k]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", l);
	PRINTPR(PR[l]);
	printf("\n");
	GetReturn();
	return;
}

void SUB_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\n-\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp = PR[l];
	PR[l] = SUBPR(PR[j], PR[k]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", l);
	PRINTPR(PR[l]);
	printf("\n");
	GetReturn();
	return;
}

void MULT_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\n*\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp = PR[l];
	PR[l] = MULTPR(PR[j], PR[k]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", l);
	PRINTPR(PR[l]);
	printf("\n");
	GetReturn();
	return;
}

void POWER_PR()
{
	unsigned int j, k, n;
	POLYR Tmp;

	if (scanf12(&n, &j, &k) == 0)
		return;
	printf("The %u-th power of PR[%u] = ", n, j);
	PRINTPR(PR[j]);
	Tmp = PR[k];
	PR[k] = POWERPR(PR[j], n);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	printf("\n");
	GetReturn();
	return;
}

void DIV_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	if (PR[k] == NULL)
		printf("PR[%u] = 0 and so PR[%u] div PR[%u] does not exist!\n", k, j, k);
	else
	{
		printf("PR[%u] = ", j);
		PRINTPR(PR[j]);
		printf("\ndiv\nPR[%u] = ", k);
		PRINTPR(PR[k]);
		Tmp = PR[l];
		PR[l] = DIVPR(PR[j], PR[k]);
		DELETE_PR(Tmp);
		printf("\n->\nPR[%u] = ", l);
		PRINTPR(PR[l]);
		printf("\n");
	}
	GetReturn();
	return;
}

void MOD_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	if (PR[k] == NULL)
		printf("PR[%u] = 0 and so PR[%u] mod PR[%u] does not exist!\n", k, j, k);
	else
	{
		printf("PR[%u] = ", j);
		PRINTPR(PR[j]);
		printf("\nmod\nPR[%u] = ", k);
		PRINTPR(PR[k]);
		Tmp = PR[l];
		PR[l] = MODPR(PR[j], PR[k]);
		DELETE_PR(Tmp);
		printf("\n->\nPR[%u] = ", l);
		PRINTPR(PR[l]);
		printf("\n");
	}
	GetReturn();
	return;
}

void SCALAR_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("R[%u] = ", j);
	PRINTR(R[j]);
	printf("\n*\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp = PR[l];
	PR[l] = SCALARPR(R[j], PR[k]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", l);
	PRINTPR(PR[l]);
	printf("\n");
	GetReturn();
	return;
}

void VAL_PR()
{
	unsigned int j, k, l;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\nevaluated at\nR[%u] = ", k);
	PRINTR(R[k]);
	FREEMPR(R[l]);
	R[l] = VALPR(PR[j], R[k]);
	printf("\n->\nR[%u] = ", l);
	PRINTR(R[l]);
	printf("\n");
	GetReturn();
	return;
}

void DERIV_PR()
{
	unsigned int j, k;
	POLYR Tmp;

	if (scanf02(&j, &k) == 0)
		return;
	printf("The derivative of PR[%u] = ", j);
	PRINTPR(PR[j]);
	Tmp = PR[k];
	PR[k] = DERIVPR(PR[j]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	printf("\n");
	GetReturn();
	return;
}

void GCD_PR()
{
	unsigned int j, k, l;
	POLYR Tmp;

	if (scanf03(&j, &k, &l) == 0)
		return;
	printf("gcd of PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\nand\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp = PR[l];
	PR[l] = GCDPR(PR[j], PR[k]);
	DELETE_PR(Tmp);
	printf("\n->\nPR[%u] = ", l);
	PRINTPR(PR[l]);
	printf("\n");
	GetReturn();
	return;
}

void EUCLID_PR()
{
	unsigned int j, k, l, m, n;
	POLYR Tmp1, Tmp2, Tmp3;

	if (scanf05(&j, &k, &m, &n, &l) == 0)
		return;
	printf("gcd of PR[%u] = ", j);
	PRINTPR(PR[j]);
	printf("\nand\nPR[%u] = ", k);
	PRINTPR(PR[k]);
	Tmp1 = PR[l];
	Tmp2 = PR[m];
	Tmp3 = PR[n];
	PR[l] = EUCLIDPR(PR[j], PR[k], &PR[m], &PR[n]);
	DELETE_PR(Tmp1);
	DELETE_PR(Tmp2);
	DELETE_PR(Tmp3);
	printf("\n->\nPR[%u] = PR[%u]*PR[%u]+PR[%u]*PR[%u] = ", l, m, j, n, k);
	PRINTPR(PR[l]);
	printf("\nwhere\nPR[%u] = ", m);
	PRINTPR(PR[m]);
	printf("\nand\n");
	printf("PR[%u] = ", n);
	PRINTPR(PR[n]);
	printf("\n");
	GetReturn();
	return;
}

void EUCLIDS_PR()
{
	unsigned int e, n, i, A[M0];
	POLYR *P, G, *H;

	printf("enter n (2 <= n <= %u) :  the number of polynomials PR[a], PR[b],...\n", M0);
	if (scanf10(&n) == 0)
		return;
	if (n < 2)
	{
		printf("n < 2\n");
		return;
	}
	e = 0;
	printf("enter %u array integers a, b, ... (< %u) :", n, M0);
	for (i = 0; i < n; i++)
		e += scanf("%u", &A[i]);
	Flush();
	if (e != n)
	{
		printf("\n\ninput error : entered only %u integers successfully\n", e);	
		GetReturn();
		return;
	}
	for (i = 0; i < n; i++)
	{
		if (A[i] >= M0)
		{
			printf("\n\ninput error : %u-th array integer >= %u\n", i + 1, M0);	
			GetReturn();
			return;
		}
	}
	P = (POLYR *)mmalloc (n * sizeof(POLYR));
	for (i = 0; i < n; i++)
		P[i] = COPYPR(PR[A[i]]);
	G = EUCLIDSPR(P, &H, n);
	for (i = 0; i < n; i++)
		DELETE_PR(P[i]);
	ffree((char *)P, n * sizeof(POLYR));
	PRINTPR(G);
	printf(" = gcd of polynomials ");
	for (i = 0; i < n; i++)
		printf("PR[%u], ", A[i]);
	printf("\n\n");
	DELETE_PR(G);
	printf(" = ");
	for (i = 0; i < n - 1; i++)
		printf("H[%u]*PR[%u]+", i, A[i]);
		printf("H[%u]*PR[%u]\n\n", i, A[i]);
	printf("where\n\n");
	for (i = 0; i < n; i++)
	{
		printf("H[%u] = ", i + 1);
		PRINTPR(H[i]);	
		DELETE_PR(H[i]);
		printf("\n");
		GetReturn();
	}
	ffree((char *)H, n * sizeof(POLYR));
	return;
}

void SQUAREFREE_PR()
{
	unsigned int j;

	if (scanf01(&j) == 0)
		return;
	if (DEGREEPR(PR[j]) <= 0)
		printf("degree PR[%u] <= 0 and a squarefree decomposition is not defined.\n", j);
	else
	{
		printf("the squarefree decomposition of PR[%u] =\n", j);
		PRINTPR(PR[j]);
		printf("\nis given by\n");
		SQUAREFREEPR(PR[j]);
	}
	GetReturn();
	return;
}

void KALTOFEN_PR()
{
	unsigned int i, j, *mults, n;
	int m;
	POLYR *G;

	if (scanf01(&j) == 0)
		return;
	m = DEGREEPR(PR[j]);
	if (m <= 0)
		printf("degree PR[%u] <= 0 and factorization is not interesting!.\n", j);
	else
	{
		printf("factorizing PR[%u] = ", j);
		PRINTPR(PR[j]);
		printf(" :\n");
		G = KALTOFENPR(PR[j], &mults, &n);
		for (i = 0; i < n; i++)
		{
			printf("irreducible factor G[%u] = ", i);
			PRINTPR(G[i]);
			DELETE_PR(G[i]);
			printf(": mult'y = %u\n", mults[i]);
		}
		ffree((char *)G, (unsigned int)m * sizeof(POLYR));
		ffree((char *)mults, (unsigned int)m * sizeof(unsigned int));
	}
	GetReturn();
	return;
}

void RESULTANT_PI()
{
	unsigned int j, k, l;
	POLYI P1, P2;
	MPI *T;
	POLYR CURSOR;

	if (scanf03(&j, &k, &l) == 0)
		return;
	if (DEGREEPR(PR[j]) <= 0)
	{
		printf("The degree of PR[%u] is %d\n", j, DEGREEPR(PR[j]));
		GetReturn();
		return;
	}
	if (DEGREEPR(PR[k]) <= 0)
	{
		printf("The degree of PR[%u] is %d\n", k, DEGREEPR(PR[k]));
		GetReturn();
		return;
	}
	CURSOR = PR[j];
	while (CURSOR != NULL)
	{
		if (!EQONEI((CURSOR->COEF)->D))
		{
			printf("The coefficient of x^%d in PR[%u] is not an integer\n", CURSOR->DEG, j);
			GetReturn();
			return;
		}
		CURSOR = CURSOR->NEXT;
	}
	CURSOR = PR[k];
	while (CURSOR != NULL)
	{
		if (!EQONEI((CURSOR->COEF)->D))
		{
			printf("The coefficient of x^%d in PR[%u] is not an integer\n", CURSOR->DEG, k);
			GetReturn();
			return;
		}
		CURSOR = CURSOR->NEXT;
	}
	P1 = COPYPR_TO_PI(PR[j]);
	P2 = COPYPR_TO_PI(PR[k]);
	T = SUBRESULTANT(P1, P2);
	printf("the resultant of ");
	PRINTPI(P1);
	printf("\nand\n");
	PRINTPI(P2);
	printf("\n -> R[%u]=", l);
	PRINTI(T);
	printf("\n\n");
	DELETE_PI(P1);
	DELETE_PI(P2);
	FREEMPR(R[l]);
	R[l] = BUILDMPR();
	R[l]->N = T;
	R[l]->D = ONEI();
	GetReturn();
	return;
}

void DISCRIMINANT_PI()
{
	unsigned int j, k;
	POLYI P1;
	MPI *T;
	POLYR CURSOR;

	if (scanf02(&j, &k) == 0)
		return;
	if (DEGREEPR(PR[j]) <= 1)
	{
		printf("The degree of PR[%u] is %d\n", j, DEGREEPR(PR[j]));
		GetReturn();
		return;
	}
	CURSOR = PR[j];
	while (CURSOR != NULL)
	{
		if (!EQONEI((CURSOR->COEF)->D))
		{
			printf("The coefficient of x^%d in PR[%u] is not an integer\n", CURSOR->DEG, j);
			GetReturn();
			return;
		}
		CURSOR = CURSOR->NEXT;
	}
	P1 = COPYPR_TO_PI(PR[j]);
	T = DISCRIMINANTPI(P1);
	printf("the discriminant of ");
	PRINTPI(P1);
	printf("\n -> R[%u]=", k);
	PRINTI(T);
	printf("\n\n");
	DELETE_PI(P1);
	FREEMPR(R[k]);
	R[k] = BUILDMPR();
	R[k]->N = T;
	R[k]->D = ONEI();
	GetReturn();
	return;
}

void CYCLOTOMIC_PR()
{
	unsigned int n, j;
	POLYI H;

	if (scanf11(&n, &j) == 0)
		return;
	clearscreen();
	if (n < 1)
		printf("n < 1\n");
	else
	{
		DELETE_PR(PR[j]);
		H = CYCLOTOMIC(n);
		PR[j] = COPYPI_TO_PR(H);
		DELETE_PI(H);
		printf("%uth cyclotomic polynomial \n", n);
		printf("\n->\nPR[%u] = ", j);
		PRINTPR(PR[j]);
		printf("\n");
	}
	GetReturn();
	return;
}
