/* cm_Pm.c */
/* modular arithmetic polynomial routines used in cmm.c.*/

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

extern POLYm Pm[];
extern MPMATm *mM[];

void ADD_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	printf("Pm[%u] = ", j);
	PRINTPm(Pm[j]);
	printf("\n +\nPm[%u] = ", k);
	PRINTPm(Pm[k]);
	Tmp = Pm[l];
	Pm[l] = ADDPm(Pm[j], Pm[k], m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, l);
	PRINTPm(Pm[l]);
	printf("\n");
	GetReturn();
	return;
}

void SUB_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	printf("Pm[%u] = ", j);
	PRINTPm(Pm[j]);
	printf("\n-\nPm[%u] = ", k);
	PRINTPm(Pm[k]);
	Tmp = Pm[l];
	Pm[l] = SUBPm(Pm[j], Pm[k], m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, l);
	PRINTPm(Pm[l]);
	printf("\n");
	GetReturn();
	return;
}

void MULT_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	printf("Pm[%u] = ", j);
	PRINTPm(Pm[j]);
	printf("\n*\nPm[%u] = ", k);
	PRINTPm(Pm[k]);
	Tmp = Pm[l];
	Pm[l] = MULTPm(Pm[j], Pm[k], m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, l);
	PRINTPm(Pm[l]);
	printf("\n");
	GetReturn();
	return;
}

void POWER_Pm()
{
	unsigned int j, k, n, m;
	POLYm Tmp;

	if (scanf121(&n, &j, &k, &m) == 0)
		return;
	printf("The %u-th power of Pm[%u] = ", n, j);
	PRINTPm(Pm[j]);
	Tmp = Pm[k];
	Pm[k] = POWERPm(Pm[j], n, m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, k);
	PRINTPm(Pm[k]);
	printf("\n");
	GetReturn();
	return;
}

void DIV_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	if (Pm[k] == NULL)
		printf("Pm[%u] = 0 and so Pm[%u] div Pm[%u] does not exist!\n", k, j, k);
	else
	{
		printf("Pm[%u] = ", j);
		PRINTPm(Pm[j]);
		printf("\ndiv\nPm[%u] = ", k);
		PRINTPm(Pm[k]);
		Tmp = Pm[l];
		Pm[l] = DIVPm(Pm[j], Pm[k], m);
		DELETE_Pm(Tmp);
		printf("\n-> (mod %u)\nPm[%u] = ", m, l);
		PRINTPm(Pm[l]);
		printf("\n");
	}
	GetReturn();
	return;
}

void MOD_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	if (Pm[k] == NULL)
		printf("Pm[%u] = 0 and so Pm[%u] mod Pm[%u] does not exist!\n", k, j, k);
	else
	{
		printf("Pm[%u] = ", j);
		PRINTPm(Pm[j]);
		printf("\nmod\nPm[%u] = ", k);
		PRINTPm(Pm[k]);
		Tmp = Pm[l];
		Pm[l] = MODPm(Pm[j], Pm[k], m);
		DELETE_Pm(Tmp);
		printf("\n-> (mod %u)\nPm[%u] = ", m, l);
		PRINTPm(Pm[l]);
		printf("\n");
	}
	GetReturn();
	return;
}

void SCALAR_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf121(&j, &k, &l, &m) == 0)
		return;
	printf("%u * Pm[%u] = ", j, k);
	PRINTPm(Pm[k]);
	Tmp = Pm[l];
	Pm[l] = SCALARPm(j, Pm[k], m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, l);
	PRINTPm(Pm[l]);
	printf("\n");
	GetReturn();
	return;
}

void VAL_Pm()
{
	unsigned int j, k, l, m;

	if (scanf012(&j, &k, &m) == 0)
		return;
	printf("Pm[%u] = ", j);
	PRINTPm(Pm[j]);
	l = VALPm(Pm[j], k, m);
	printf("\nevaluated (mod %u) at %u\n->%u\n", m, k, l);
	GetReturn();
	return;
}

void DERIV_Pm()
{
	unsigned int j, k, m;
	POLYm Tmp;

	if (scanf021(&j, &k, &m) == 0)
		return;
	printf("The derivative of Pm[%u] = ", j);
	PRINTPm(Pm[j]);
	Tmp = Pm[k];
	Pm[k] = DERIVPm(Pm[j], m);
	DELETE_Pm(Tmp);
	printf("\n-> (mod %u)\nPm[%u] = ", m, k);
	PRINTPm(Pm[k]);
	printf("\n");
	GetReturn();
	return;
}

void GCD_Pm()
{
	unsigned int j, k, l, m;
	POLYm Tmp;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	printf("gcd (mod %u) of Pm[%u] = ", m, j);
	PRINTPm(Pm[j]);
	printf("\nand\nPm[%u] = ", k);
	PRINTPm(Pm[k]);
	Tmp = Pm[l];
	Pm[l] = GCDPm(Pm[j], Pm[k], m);
	DELETE_Pm(Tmp);
	printf("\n->\nPm[%u] = ", l);
	PRINTPm(Pm[l]);
	printf("\n");
	GetReturn();
	return;
}

void EUCLID_Pm()
{
	unsigned int j, k, l, m, n, M;
	POLYm Tmp1, Tmp2, Tmp3;

	if (scanf051(&j, &k, &m, &n, &l, &M) == 0)
		return;
	printf("gcd (mod %u) of Pm[%u] = ", M, j);
	PRINTPm(Pm[j]);
	printf("\nand\nPm[%u] = ", k);
	PRINTPm(Pm[k]);
	Tmp1 = Pm[l];
	Tmp2 = Pm[m];
	Tmp3 = Pm[n];
	Pm[l] = EUCLIDPm(Pm[j], Pm[k], &Pm[m], &Pm[n], M);
	DELETE_Pm(Tmp1);
	DELETE_Pm(Tmp2);
	DELETE_Pm(Tmp3);
	printf("\n->\nPm[%u] = Pm[%u]*Pm[%u]+Pm[%u]*Pm[%u] = ", l, m, j, n, k);
	PRINTPm(Pm[l]);
	printf("\nwhere\nPm[%u] = ", m);
	PRINTPm(Pm[m]);
	printf("\nand\n");
	printf("Pm[%u] = ", n);
	PRINTPm(Pm[n]);
	printf("\n");
	GetReturn();
	return;
}

void EUCLIDS_Pm()
{
	unsigned int e, n, i, A[M0], m;
	POLYm *P, G, *H;

	if (scanf10(&m) == 0)
		return;
	printf("enter n (2 <= n <= %u)\n", M0);
	printf("where n is the number of (mod %u) polynomials Pm[a], Pm[b],...\n", m);
	if (scanf10(&n) == 0)
		return;
	if (n < 2)
	{
		printf("n < 2\n");
		return;
	}
	e = 0;
	printf("enter %u array integers a, b, ... (< %u) :", n, M0);
	for (i = 0; i < n; i++)
		e += scanf("%u", &A[i]);
	Flush();
	if (e != n)
	{
		printf("\n\ninput error : entered only %u integers successfully\n", e);	
		GetReturn();
		return;
	}
	for (i = 0; i < n; i++)
	{
		if (A[i] >= M0)
		{
			printf("\n\ninput error : %u-th array integer >= %u\n", i + 1, M0);	
			GetReturn();
			return;
		}
	}
	P = (POLYm *)mmalloc (n * sizeof(POLYm));
	for (i = 0; i < n; i++)
		P[i] = COPYPm(Pm[A[i]], m);
	G = EUCLIDSPm(P, &H, n, m);
	for (i = 0; i < n; i++)
		DELETE_Pm(P[i]);
	ffree((char *)P, n * sizeof(POLYm));
	PRINTPm(G);
	printf(" = gcd of polynomials ");
	for (i = 0; i < n; i++)
		printf("Pm[%u], ", A[i]);
	printf("\n\n");
	DELETE_Pm(G);
	printf(" = ");
	for (i = 0; i < n - 1; i++)
		printf("H[%u]*Pm[%u]+", i, A[i]);
		printf("H[%u]*Pm[%u]\n\n", i, A[i]);
	printf("where\n\n");
	for (i = 0; i < n; i++)
	{
		printf("H[%u] = ", i + 1);
		PRINTPm(H[i]);	
		DELETE_Pm(H[i]);
		printf("\n");
		GetReturn();
	}
	ffree((char *)H, n * sizeof(POLYm));
	return;
}

void BERLEKAMP_Pm()
{
	unsigned int j, k, m;

	if (scanf021(&j, &k, &m) == 0)
		return;
	if (DEGREEPm(Pm[j]) <= 0)
		printf("degree Pm[%u] <= 0 and the Berlekamp matrix is not defined.\n", j);
	else
	{
		FREEMATm(mM[k]);
		printf("the Berlekamp Matrix (mod %u) of Pm[%u] = \n", m, j);
		PRINTPm(Pm[j]);
		printf("\n");
		printf("is mM[%u] = \n", k);
		mM[k] = BERLEKAMPm(Pm[j], m);
		PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
	}
	GetReturn();
	return;
}

void SQUAREFREE_Pm()
{
	unsigned int i, j, m, n, *C, number;
	POLYm *p;

	if (scanf011(&j, &m) == 0)
		return;
	if (DEGREEPm(Pm[j]) <= 0)
		printf("degree Pm[%u] <= 0 and a squarefree decomposition is not defined.\n", j);
	else
	{
		printf("the squarefree decomposition of Pm[%u] =\n", j);
		PRINTPm(Pm[j]);
		printf("\nis given by\n");
		n = (unsigned int)(Pm[j]->DEG); 
		p = YUNSQUAREFREEPm(Pm[j], &C, &number, m);
		for (i = 0; i <= number - 1; i++)
		{
			printf("p[%u] = ", i);
			PRINTPm(p[i]);
			printf(" : multiplicity = %u\n", C[i]);
			DELETE_Pm(p[i]);
		}
		ffree((char *)p, n * sizeof(POLYm));
		ffree((char *)C, n * sizeof(unsigned int));
	}
	GetReturn();
	return;
}

void FACTORS_Pm()
{
	unsigned int i, j, *mults, n, M, lead;
	int m;
	POLYm *G;

	if (scanf011(&j, &M) == 0)
		return;
	m = DEGREEPm(Pm[j]);
	if (m <= 0)
		printf("degree Pm[%u] <= 0 and factorization is not interesting!.\n", j);
	else
	{
		printf("factorizing (mod %u) Pm[%u] = ", M, j);
		PRINTPm(Pm[j]);
		printf(" :\n");
		G = FACTORSPm(Pm[j], &lead, &mults, &n, M);
		printf("leading coefficient = %u\n", lead);
		for (i = 0; i < n; i++)
			{
				printf("irreducible factor G[%u] = ", i);
				PRINTPm(G[i]);
				DELETE_Pm(G[i]);
				printf(": multiplicity = %u\n", mults[i]);
			}
		ffree((char *)G, (unsigned int)m * sizeof(POLYm));
		ffree((char *)mults, (unsigned int)m * sizeof(unsigned int));
	}
	GetReturn();
	return;
}

void POWER0_Pm()
{
	unsigned int j, k, l, n, m;
	POLYm Tmp;

	if (scanf131(&n, &j, &k, &l, &m) == 0)
		return;
	if (Pm[k] == NULL)
		printf("Pm[%u] = 0!\n", k);
	else
	{
		printf("The %u-th power of Pm[%u] = ", n, j);
		PRINTPm(Pm[j]);
		printf("\n(mod  Pm[%u]) = ", k);
		PRINTPm(Pm[k]);
		Tmp = Pm[l];
		Pm[l] = POWER0Pm(Pm[j], n, Pm[k], m);
		DELETE_Pm(Tmp);
		printf("\n-> (mod %u)\nPm[%u] = ", m, l);
		PRINTPm(Pm[l]);
		printf("\n");
	}
	GetReturn();
	return;
}

void IRR_Pm()
{
	unsigned int j, n, p;
	POLYm temp;

	if (scanf012(&j, &n, &p) == 0)
		return;
	if (n <= 1 || p <= 1)
	{
		printf("n <= 1 or p <= 1, try again\n");
		GetReturn();
		return;
	}
	else
	{
		temp = Pm[j];
		Pm[j] = IRRPm(n, p);
		DELETE_Pm(temp);
		printf("found irreducible polynomial of degree %u (mod %u)\n", n, p);
		printf("\n->\nPm[%u] = ", j);
		PRINTPm(Pm[j]);
		printf("\n");
		GetReturn();
		return;
	}
}

void CYCLOTOMIC_Pm()
{
	unsigned int j, n, p;
	POLYm temp;
	POLYI tempI;

	if (scanf012(&j, &n, &p) == 0)
		return;
	if (n <= 1 || p <= 1)
	{
		printf("n <= 1 or p <= 1, try again\n");
		GetReturn();
		return;
	}
	else
	{
		tempI = CYCLOTOMIC(n);
		temp = Pm[j];
		Pm[j] = CHANGETOPm(tempI, p);
		DELETE_Pm(temp);
		DELETE_PI(tempI);
		printf("%uth cyclotomic polynomial (mod %u)\n", n, p);
		printf("\n->\nPm[%u] = ", j);
		PRINTPm(Pm[j]);
		printf("\n");
		GetReturn();
		return;
	}
}
