/* cm_mM2.c */

/* modular arithmetic routines used in cmm.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

extern MPMATm *mM[];
extern POLYm Pm[];

extern char cmM[];
extern char cPmM[];
extern char cPm[];

void CHAR_POLY_m()
{
	unsigned int j, k, m;

	if (scanf021(&j, &k, &m) == 0)
		return;
	if (mM[j]->R != mM[j]->C)
		printf("Matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else
	{
		DELETE_Pm(Pm[k]);
		printf("the characteristic polynomial (mod %u) of mM[%u] = \n\n", m, j);
		PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
		printf("is Pm[%u] = ", k);
		Pm[k] = MCWORTERm(mM[j], m);
		PRINTPm(Pm[k]);
		printf("\n");
	}
	GetReturn();
	return;
}
	
void MINIMUM_POLY_m()
{
	unsigned int j, k, m;

	if (scanf021(&j, &k, &m) == 0)
		return;
	if (mM[j]->R != mM[j]->C)
		printf("Matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else
	{
		DELETE_Pm(Pm[k]);
		printf("the minimum polynomial of mM[%u] = \n\n", j);
		PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
		printf("is Pm[%u] = ", k);
		Pm[k] = MINIMUM_POLYm(mM[j], m);
		PRINTPm(Pm[k]);
		printf("\n");
	}
	GetReturn();
	return;
}

void ADJ_m()
{
	unsigned int j, k, m;
	MPMATm *Temp;

	if (scanf021(&j, &k, &m) == 0)
		return;
	if (mM[j]->R != mM[j]->C)
		printf("Matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else
	{
		printf("adj (mod %u) of mM[%u] = \n", m, j);
		PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
		Temp = mM[k];
		mM[k] = ADJOINTm(mM[j], m);
		FREEMATm(Temp);
		printf("->\n");
		printf("mM[%u] = \n", k);
		PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
	}
	GetReturn();
	return;
}

void DET_m()
{
	unsigned int j, det, m;

	if (scanf011(&j, &m) == 0)
		return;
	if (mM[j]->R != mM[j]->C)
		printf("Matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else
	{
		printf("det (mod%u) of mM[%u] = \n", m, j);
		PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
		det = DETm(mM[j], m);
		printf("-> %u\n", det);
	}
	GetReturn();
	return;
}

void INVERSE_MAT_m()
{
	unsigned int j, k, m;
	MPMATm *N;

	if (scanf021(&j, &k, &m) == 0)
		return;
	if (mM[j]->R != mM[j]->C)
		printf("Matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else
	{
		printf("inverse (mod %u) of mM[%u] = \n", m, j);
		PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
		N = INVERSEMATm(mM[j], m);
		if (N == NULL)
			printf("does not exist.\n");
		else
		{
			FREEMATm(mM[k]);
			mM[k] = N;
			printf("->\n");
			printf("mM[%u] = \n", k);
			PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
		}
	}
	GetReturn();
	return;
}

void TRANSFORM_MAT_m()
{
	unsigned int j, k, l, m;
	MPMATm *Temp1, *Temp2, *N;

	if (scanf031(&j, &k, &l, &m) == 0)
		return;
	if (mM[k]->C != mM[k]->R)
		printf("matrix mM[%u] is not square\n", k);
	else if (mM[j]->C != mM[j]->R)
		printf("matrix mM[%u] is not square\n", j);
	else if (mM[j]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", j);
	else if (mM[k]->R == 1)
		printf("Matrix mM[%u] is 1 x 1\n", k);
	else if (mM[j]->R != mM[k]->R)
		printf("matrices mM[%u] and mM[%u] have different sizes\n", j, k);
	else
	{
		N = INVERSEMATm(mM[j], m);
		if (N == NULL)
			printf("mM[%u] inverse does not exist.\n", j);
		else
		{
			printf("The mM[%u] transform of mM[%u] -> \n", j, k);
			Temp1 = MULTMATm(N, mM[k], m);
			Temp2 = mM[l];
			mM[l] = MULTMATm(Temp1, mM[j], m);
			FREEMATm(Temp1);
			FREEMATm(Temp2);
			FREEMATm(N);
			printf("mM[%u] = \n", l);
			PRINTMATm(0, mM[l]->R - 1, 0, mM[l]->C - 1, mM[l]);
		}
	}
	GetReturn();
	return;
}

void SOLVE_m()
{
	unsigned int j, m;

	if (scanf011(&j, &m) == 0)
		return;
	printf("Solving (mod %u) the system with augmented matrix mM[%u] =  \n\n", m, j);
	PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
	printf("\n");
	EQUATIONSm(mM[j], m);
	GetReturn();
	return;
}

void ROW_REDUCE_m()
{
	unsigned int j, k, m, r, *D;
	MPMATm *Temp;

	if (scanf021(&j, &k, &m) == 0)
		return;
	printf("The reduced row-echelon form (mod %u) of mM[%u] = \n", m, j);
	PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
	Temp = mM[k];
	mM[k] = GAUSS_JORDANm(mM[j], &r, &D, m);
	FREEMATm(Temp);
	printf("->\n");
	printf("mM[%u] = \n", k);
	PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
	ffree((char *)D, mM[j]->C * sizeof(unsigned int));
	printf("rank = %u\n", r);
	GetReturn();
	return;
}

void NULL_SPACE_m()
{
	unsigned int j, k, m;
	MPMATm *N;

	if (scanf021(&j, &k, &m) == 0)
		return;
	N = NULL_SPACEm(mM[j], m);
	printf("null-space (mod %u) of mM[%u] = \n", m, j);
	PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
	if (N == NULL)
		printf("= {0}\n");
	else
	{
		FREEMATm(mM[k]);
		mM[k] = N;
		printf("has basis consisting of the columns of the matrix mM[%u] = \n", k);
		PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
	}
	GetReturn();
	return;
}

void COLUMN_SPACE_m()
{
	unsigned int j, k, m;
	MPMATm *N;

	if (scanf021(&j, &k, &m) == 0)
		return;
	N = COLUMN_SPACEm(mM[j], m);
	printf("Column-space (mod %u) of mM[%u] = \n", m, j);
	PRINTMATm(0, mM[j]->R - 1, 0, mM[j]->C - 1, mM[j]);
	if (N == NULL)
		printf("= {0}\n");
	else
	{
		FREEMATm(mM[k]);
		mM[k] = N;
		printf("has basis consisting of the columns of the matrix mM[%u] = \n", k);
		PRINTMATm(0, mM[k]->R - 1, 0, mM[k]->C - 1, mM[k]);
	}
	GetReturn();
	return;
}

void MATALG_m()
{
	int GoOnFlag = 1;
	char c;
	while (GoOnFlag)
	{
		c = drawmenu(11);
		switch (c)
		{
		case 'c':
			CHAR_POLY_m();
			break;
		case 'm':
			MINIMUM_POLY_m();
			break;
		case 'a':
			ADJ_m();
			break;
		case 'd':
			DET_m();
			break;
		case 'i':
			INVERSE_MAT_m();
			break;
		case 't':
			TRANSFORM_MAT_m();
			break;
		case 'l':
			SOLVE_m();
			break;
		case 'g':
			ROW_REDUCE_m();
			break;
		case 'n':
			NULL_SPACE_m();
			break;
		case 'b':
			COLUMN_SPACE_m();
			break;
		case 'p':
			PRINTm_();
			break;
		case 'x':
			INPUTm_();
			break;
		case 'q':
			FPRINTMAT_m(cmM);
			FPRINT_Pm(cPm);
			FPRINTMAT_Pm(cPmM);
			GoOnFlag = 0;
			break;
		}
	}
}
