
/* program "i1.c" */
/*
 *  A translation into C of the  pascal program ARITHMETIC,
 * from "SCIENTIFIC PASCAL" by H. FLANDERS pp. 342-357. 
 * The functions RSV(), ADD0I(), ADD0_I(), SUB0I(), SUB0_I(),
 * ADDI(), SUBI(), MULTI(), POWERI(), MULT_I(), POWER_I() are constructed.
 */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

int MAX(int i, int j)
{
	return(i >= j ? i: j);
} 

int MIN(int i, int j)
{
	return(i <= j ? i: j);
} 

int RSV(MPI *Aptr, MPI *Bptr)
/*
 * relative sizes of absolute values of MPI'S *Aptr and *Bptr.
 *				    1 if |*Aptr| > |*Bptr|
 *            RSV(Aptr, Bptr) =     0 if |*Aptr| = |*Bptr|
 *       			   -1 if |*Aptr| < |*Bptr|
 */
{
 	unsigned int j;

	if (Aptr->D > Bptr->D) 
		return (1);
	else if (Aptr->D < Bptr->D) 
		return (-1);
	else
	{
		j = Aptr->D;
		while ((Aptr->V[j] == Bptr->V[j]) && j) 
			j-- ;
		if (Aptr->V[j] > Bptr->V[j]) 
			return (1);
		else if (Aptr->V[j] < Bptr->V[j]) 
			return (-1);
		else 
			return (0);
        }
} 

MPI *ADD0I(MPI *Aptr, MPI *Bptr)
/*
 * input: Aptr and Bptr are pointers to non-negative MPI'S.
 * output: *Eptr = *Aptr + *Bptr.
 */
{
	unsigned int c, j, t, m, n;
	MPI *Eptr;
	MPI *tmp;

	m = (unsigned int)MIN((int) Aptr->D, (int) Bptr->D);
	n = (unsigned int)MAX((int) Aptr->D, (int) Bptr->D);
	c = 0;

	Eptr = BUILDMPI(n + 1);
	for (j = 0; j <= m; j++)
	{
		t = Aptr->V[j] + Bptr->V[j] + c;
		Eptr->V[j] = t & (R0 - 1);
		c = t >> T0;
	}
	if (Aptr->D > Bptr->D)
	{
		for (j = 1 + m; j <= Aptr->D; j++)
		{
			t = Aptr->V[j] + c;
			Eptr->V[j] = t & (R0 - 1);
			c = t >> T0;
		}
	}
	if (Aptr->D < Bptr->D)
	{
		for (j = 1 + m; j <= Bptr->D; j++)
		{
			t = Bptr->V[j] + c;
			Eptr->V[j] = t & (R0 - 1);
			c = t >> T0;
		}
	}
	if (c)
	{
		tmp = BANK_REALLOC(Eptr, n + 1);
		FREEMPI(Eptr);
		Eptr = tmp;
        	Eptr->V[n + 1] = c;
	}
	Eptr->S = MAX(Aptr->S, Bptr->S);

	return  Eptr;
}

MPI *ADD0_I(MPI *Aptr, USI b)
/*
 * input: Aptr is a pointer to a non-negative MPI.
 * b is a non-negative integer < R0.
 * output: *Eptr = *Aptr + b.
 */
{
	unsigned int c, j, n, t;
	MPI *Eptr, *tmp;

	if (b == 0)
		return COPYI(Aptr);
	n = Aptr->D;
	Eptr = BUILDMPI(n + 1);
	c = b;
	for (j = 0; j <= n; j++)
	{
		t = Aptr->V[j] + c;
		Eptr->V[j] = t & (R0 - 1);
		c = t >> T0;
	}
	if (c)
	{
		tmp = Eptr;
		Eptr = BANK_REALLOC(Eptr, n + 1);
		FREEMPI(tmp);
        	Eptr->V[n + 1] = c;
	}
	Eptr->S = b ? 1 : Aptr->S;

	return Eptr;
}

MPI *SUB0I(MPI *Aptr, MPI *Bptr)
/*
 * input:  Aptr, Bptr, pointers to non-negative MPI'S, *Aptr >= *Bptr.
 * output: MPI *Eptr = *Aptr - *Bptr.
 */
{

	int k;
	unsigned int c, d, e, j, m, n, t;
	MPI *Eptr, *tmptr;

	if (Bptr->S == 0)
		return COPYI(Aptr);
	n = Aptr->D;
	Eptr = BUILDMPI(n + 1);
	m = Bptr->D;
	c = 1;

	/* c = 1: no borrow, c = 0: a borrow */

	for (j = 0; j <= m; j++)
	{
		t = Aptr->V[j] - Bptr->V[j] + R0 + c - 1;
		Eptr->V[j] = t & (R0 - 1);
		c = t >> T0;
	}
	if (n > m)
	{
		for (j = m + 1; j <= n; j++)
		{
			t = Aptr->V[j] + R0 + c - 1;
			Eptr->V[j] = t & (R0 - 1);
			c = t >> T0;
		}
	}
	/* finding Eptr->S and Eptr->D */

	Eptr->S = 0;
	d = 0;
	k = n;
	while (k >= 0)
	{
		if (Eptr->V[k] != 0)
		{
			Eptr->S = 1;
			d = k;
	       	        k = -1;
		}
		else 
			k-- ;
        }	
	e = n - d;
	if (e)
	{
		tmptr = BANK_REALLOC(Eptr, d);
		FREEMPI(Eptr);
		Eptr = tmptr;
	}

	return Eptr;
}

MPI *SUB0_I(MPI *Aptr, USI b)
/*
 * input: Aptr a pointer to a non-negative MPI. b a non-negative integer,
 * b < R0, *Aptr >= b;  output: *Eptr = *Aptr - b.
 */
{
	unsigned int c, e, t, j, n;
	MPI *Eptr, *tmp;
	
	if (b == 0)
		return COPYI(Aptr);
	n = Aptr->D;
	Eptr = BUILDMPI(n + 1);
	c = 1;
	/* c = 1: no borrow, c = 0: a borrow. */
	t = Aptr->V[0] - b + R0;
	Eptr->V[0] = t & (R0 - 1);
	c = t >> T0;
	for (j = 1; j <= n; j++)
	{
		t = Aptr->V[j] + R0 + c - 1;
		Eptr->V[j] = t & (R0 - 1);
		c = t >> T0;
	}
	if (Aptr->D)
		Eptr->S = 1;
	else
		Eptr->S = Eptr->V[0] ? 1 : 0;
	e = Eptr->V[n];
	if (e == 0 && n )
	{
		tmp = Eptr;
		Eptr = BANK_REALLOC(Eptr, n - 1);
		FREEMPI(tmp);	
	}

	return Eptr;
}

MPI *ADDI(MPI *Aptr, MPI *Bptr)
/*
 * input: Aptr and Bptr, pointers to MPI'S.
 * output: MPI *Eptr = *Aptr + *Bptr.
 */
{
	MPI *Eptr;

	if (Aptr->S == Bptr->S)
	{
		Eptr = ADD0I(Aptr, Bptr);
		Eptr->S = Aptr->S;
	}
	else 
	{
		switch (RSV(Aptr, Bptr))
		{
			case -1:
			{
				Eptr = SUB0I(Bptr, Aptr);
				Eptr->S = Bptr->S;
				break;
			}
			case 0:
			{
				Eptr = ZEROI();
				break;
			}
			case 1:
			{
				Eptr = SUB0I(Aptr, Bptr);
				Eptr->S = Aptr->S;
				break;
			}
			default: Eptr = NULL;
		}
	}
	return Eptr;
}

MPI *SUBI(MPI *Aptr, MPI *Bptr)
/*
 * input: Aptr and Bptr, pointers to MPI'S.
 * output: MPI *Eptr = *Aptr - *Bptr.
 */
{
	MPI *Eptr;

	Bptr->S = -Bptr->S;
	Eptr = ADDI(Aptr, Bptr);
	Bptr->S = -Bptr->S;
	return Eptr;
}

MPI *BANK_REALLOC(MPI *Aptr, USI degree)
/*
 * Reallocs using the memory bank. The original MPI is Aptr, and the new one 
 * has the given degree.
 */
{
	int size;
	MPI *Bptr;
#ifdef _WIN32
	unsigned int k;
#endif

	if(Aptr->D <= degree)
	  size = Aptr->D;
	else
	  size = degree;
	Bptr = BUILDMPI(1 + degree);
#ifdef _WIN32
	for (k = 0; k <= size; k++)
		Bptr->V[k] = Aptr->V[k];
#else
	memcpy(Bptr->V, Aptr->V, (1 + size) * sizeof(unsigned int));
#endif
	Bptr->S = Aptr->S;

	return Bptr;
}

MPI *MULTI(MPI *Aptr, MPI *Bptr)
/*
 * input: Aptr and Bptr, pointers to MPI'S.
 * output: MPI *Eptr = *Aptr * *Bptr. 
 */
{
	register unsigned int j, bk, k, t, c = 0, m, n;
	int e;
	MPI *Eptr, *tmp;

	e = Aptr->S * Bptr->S;
	if (e == 0)
		return ZEROI();
	m = Aptr->D;
	n = Bptr->D;
	Eptr = BUILDMPI(m + n + 2);
	for (k = 0; k <= m; k++)
		Eptr->V[k] = 0;
	for (k = 0; k <= n; k++)
	{
		bk = Bptr->V[k];
		if (bk != 0)
		{
			c = 0;
			for (j = 0; j <= m; j++)
			{
				t = Aptr->V[j] * bk + Eptr->V[j + k] + c;
					/* t < R0 * R0 < 2 ^ O32 */
				Eptr->V[j + k] = t & (R0 - 1);
				c = t >> T0; 		/* c < R0 */
			}			
			Eptr->V[m + k + 1] = c;
		}
		else
			Eptr->V[m + k + 1] = 0;
	}			
	Eptr->S = e;
	if (c == 0)
	{
		tmp = BANK_REALLOC(Eptr, m + n);
		FREEMPI(Eptr);
		Eptr = tmp;
	}

	return Eptr;
}

MPI *MULT_I(MPI *Aptr, int b)
/*
 * input: Aptr is a pointer to an MPI, Aptr->D < K0.
 * b is an integer, |b| < R0. output: *Eptr = *Aptr * b. 
 */
{
	unsigned int a, t, c = 0, j, n;
	MPI *Eptr;
	MPI *tmp;
	
	if (b == 0 || Aptr->S == 0)
		return ZEROI();
	n = Aptr->D;
	Eptr = BUILDMPI(n + 1);
	Eptr->S = b > 0 ? Aptr->S : - Aptr->S;
	a = b > 0 ? b : -b;
	for (j = 0; j <= n; j++)
	{
		t = Aptr->V[j] * a + c;
		Eptr->V[j] = t & (R0 - 1);
		c = t >> T0;
	}
	if (c)
	{
		tmp = BANK_REALLOC(Eptr, n + 1);
		FREEMPI(Eptr);
		Eptr = tmp;
        	Eptr->V[n + 1] = c;
	}

	return Eptr;
}

MPI *POWERI(MPI *Aptr, USI n)
/*
 *  *Eptr = (*Aptr) ^ n, where 0 <= n < R0 * R0.
 */
{
	MPI *Btmp, *Ctmp, *Eptr;

	Eptr = ONEI();
	if (n == 0)
		return Eptr;
	Btmp = COPYI(Aptr);
	while (1)
	{
		if (n & 1)
		{
			Ctmp = Eptr;
			Eptr = MULTI(Ctmp, Btmp);
			FREEMPI(Ctmp);
			if (n == 1)
			{
				FREEMPI(Btmp);
				return Eptr;
			}
		}
		Ctmp = Btmp;
		Btmp = MULTI(Ctmp, Ctmp);
		FREEMPI(Ctmp);
		n = n >> 1;
	}
}

MPI *POWER_I(int a, USI n)
/*
 * n is a non-negative integer, n < R0 * R0.
 * a is an integer, |a| < R0.
 * *Eptr = a ^ n.
 */
{
	MPI *Eptr, *Btmp, *Ctmp;

	if (a == 0)
		return ZEROI();
	if (n == 0)
		return ONEI();
	Eptr = ONEI();
	Btmp = BUILDMPI(1);
	Btmp->S = a > 0 ? 1 : -1;
	Btmp->V[0] = a > 0 ? a : -a;
	while (1)
	{
		if (n & 1)
		{
			Ctmp = Eptr;
			Eptr = MULTI(Ctmp, Btmp);
			FREEMPI(Ctmp);
			if (n == 1)
			{
				FREEMPI(Btmp);
				return Eptr;
			}
		}
		Ctmp = Btmp;
		Btmp = MULTI(Ctmp, Ctmp);
		FREEMPI(Ctmp);
		n = n >> 1;
	}
}

MPI *MULTI3(MPI *A, MPI *B, MPI *C)
/*
 * Returns A * B * C.
 */
{
	MPI *Tmp1, *Tmp2;

	Tmp1 = MULTI(A, B);
	Tmp2 = MULTI(Tmp1, C);
	FREEMPI(Tmp1);
	return (Tmp2);
}

int COMPAREI(MPI *Aptr, MPI *Bptr)
/*
 * returns 1 if Aptr > Bptr, 0 if Aptr = Bptr, -1 if Aptr < Bptr.
 */
{
	MPI *Temp;
	int t;

	Temp = SUBI(Aptr, Bptr);
	t = Temp->S;
	FREEMPI(Temp);
	return (t);
}
