/* i3.c */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

#ifndef _WIN32
#ifdef FORBES
#define	MAX_PT	100000
static	char	*ptrs[MAX_PT];
static	int	bytes[MAX_PT];
#endif
#endif

#ifdef DEBUG
long int nettbytes = 0;
#endif

POLYI BANKI = NULL;
POLYR BANKR = NULL;
POLYCI BANKCI = NULL;
POLYCR BANKCR = NULL;
POLYm BANKm = NULL;

/*
For comments on the memory banking algorithm, please see the appropriate
routines later in this file.
	Peter Adams,  1/4/1993
*/

void *realloc1(void *ptr, size_t nbytes, long mbytes)
/*
 * Reallocates nbytes of memory using realloc().
 * mbytes = nett increase in memory. Exits if call to realloc fails.
 */
{
	register char *s;
#ifndef _WIN32
#ifdef FORBES
	int i;
#endif
#endif

	if ((s = realloc(ptr, nbytes)) == NULL)
	{
		fprintf(stderr, "out of memory!\n\
		realloc request for %u bytes failed.\n", (USI)nbytes);
		exit (1);
		
	}

#ifdef DEBUG
	nettbytes += mbytes;
#endif

#ifdef _WIN32
	return (s);
#else
#ifndef FORBES
	return(s); /* returns only if s points to a valid block of memory.*/
#endif
#endif

#ifndef _WIN32
#ifdef FORBES /*This switches on the Michael Forbes debugger.*/
	for (i = 0; i < MAX_PT; i++)
		if (!ptrs[i] || ptrs[i] == s)
		{
			ptrs[i] = s;
			bytes[i] = nbytes;
			return (s);
		}
	fprintf(stderr, "out of ptr spots\n");
	exit (1);
#endif
#endif
}

void *malloc1(size_t nbytes)
/*
 * mmalloc -- allocates nbytes of memory using malloc().
 * Exits if call to malloc fails.
 */
{
	register char *s;
#ifndef _WIN32
#ifdef FORBES
	int i;
#endif
#endif

	if ((s = malloc(nbytes)) == NULL)
	{
		fprintf(stderr, "out of memory!\n\
		malloc request for %u bytes failed.\n", nbytes);
		exit (1);
		
	}
#ifdef DEBUG
	nettbytes += nbytes;
#endif
#ifdef _WIN32
	return (s);
#else
#ifndef FORBES
	return(s); /*returns only if s points to a valid block of memory.*/
#endif
#endif

#ifndef _WIN32
#ifdef FORBES /*This switches on the Michael Forbes debugger.*/
	for (i = 0; i < MAX_PT; i++)
		if (!ptrs[i] || ptrs[i] == s)
		{
			ptrs[i] = s;
			bytes[i] = nbytes;
			return (s);
		}
	fprintf(stderr, "out of ptr spots\n");
	exit (1);
#endif
#endif
}

void *calloc1(size_t m, size_t n)
/* allocates m * n of contiguous memory using calloc().
 * Exits if call to calloc fails.
 */
{
	register char *s;
#ifndef _WIN32
#ifdef FORBES
	int i;
#endif
#endif

	if ((s = calloc(m, n)) == NULL)
	{
		fprintf(stderr, "out of memory!\n\
		calloc request for %ld bytes failed.\n",(long)(m * n));
		exit (1);
		
	}
#ifdef DEBUG
	nettbytes += m * n;
#endif

#ifdef _WIN32
	return (s);
#else
#ifndef FORBES
	return(s); /*returns only if s points to a valid block of memory.*/
#endif
#endif

#ifndef _WIN32
#ifdef FORBES /*This switches on the Michael Forbes debugger.*/
	for (i = 0; i < MAX_PT; i++)
		if (!ptrs[i] || ptrs[i] == s)
		{
			bytes[i] = m * n;
			ptrs[i] = s;
			return (s);
		}
	fprintf(stderr, "out of ptr spots\n");
	exit (1); /* returns only if s points to a valid block of memory.*/
#endif
#endif
}

void sfree(void *ptr, size_t nbytes, char *errorfile, int errorline)
/*
 * frees memory using free, keeping track of nettbytes.
 */
{
#ifndef _WIN32
#ifdef FORBES
	char aline[80];
	int i;
#endif
#endif
	free(ptr);

#ifdef DEBUG
	nettbytes -= nbytes;
#endif

#ifdef _WIN32
	return;
#else
#ifndef FORBES
	return;
#endif
#endif
#ifndef _WIN32
#ifdef FORBES /* This switches on the Michael Forbes debugger. */
	for (i = 0; i < MAX_PT; i++)
		if (ptrs[i] == ptr)
		{
			if (bytes[i] != nbytes)
			{
				fprintf (stderr, "Malloced %d Freeing %u - %ld\n", bytes[i], nbytes, (long)ptr);
				fprintf(stderr,"Error occured in file %s on line %d\n",errorfile,errorline);
				gets (aline);
			}
			else
			{
				bytes[i] = -1;
				free(ptr);
			}
			return;
		}
	fprintf (stderr, "Could not find pointer\n");
	fprintf(stderr,"Error occured in file %s on line %d\n",errorfile,errorline);
	exit (1);
#endif
#endif
}

unsigned int EQONEI(MPI *Mptr)
/*
 * returns 1 if *Mptr == 1, 0 otherwise.
 */
{
	if (Mptr->S == 1 && Mptr->D == 0 && Mptr->V[0] == 1)
		return (1);
	else
		return (0);
}

unsigned int EQZEROI(MPI *Mptr)
/*
 * returns 1 if *Mptr == 0, 0 otherwise.
 */
{
	if (Mptr->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQMINUSONEI(Mptr)
/*
 * returns 1 if *Mptr == -1, 0 otherwise.
 */
MPI *Mptr;
{
	if (Mptr->S == -1 && Mptr->D == 0 && Mptr->V[0] == 1)
		return (1);
	else
		return (0);
}

unsigned int EQONER(Mptr)
/*
 * returns 1 if *Mptr == 1, 0 otherwise.
 */
MPR *Mptr;
{
	if (EQUALI(Mptr->N, Mptr->D))
		return (1);
	else
		return (0);
}
 
unsigned int EQMINUSONER(Mptr)
/*
 * returns 1 if *Mptr == -1, 0 otherwise.
 */
MPR *Mptr;
{
	if (EQMINUSONEI(Mptr->N) * EQONEI(Mptr->D))
		return (1);
	else
		return (0);
}

unsigned int EQZEROR(Mptr)
/*
 * returns 1 if *Mptr == 0, 0 otherwise.
 */
MPR *Mptr;
{
	if (Mptr->N->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQMINUSONECR(Mptr)
/*
 * returns 1 if *Mptr == -1, 0 otherwise.
 */
MPCR *Mptr;
{
	if (EQMINUSONER(Mptr->R) && Mptr->I->N->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQONECR(Mptr)
/*
 * returns 1 if *Mptr == 1, 0 otherwise.
 */
MPCR *Mptr;
{
	if (EQONER(Mptr->R) && Mptr->I->N->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQZEROCR(Mptr)
/*
 * returns 1 if *Mptr == 0, 0 otherwise.
 */
MPCR *Mptr;
{
	if (Mptr->R->N->S == 0 && Mptr->I->N->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQMINUSONECI(Mptr)
/*
 * returns 1 if *Mptr == -1, 0 otherwise.
 */
MPCI *Mptr;
{
	if (EQMINUSONEI(Mptr->R) && Mptr->I->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQONECI(Mptr)
/*
 * returns 1 if *Mptr == 1, 0 otherwise.
 */
MPCI *Mptr;
{
	if (EQONEI(Mptr->R) && Mptr->I->S == 0)
		return (1);
	else
		return (0);
}

unsigned int EQZEROCI(Mptr)
/*
 * returns 1 if *Mptr == 0, 0 otherwise.
 */
MPCI *Mptr;
{
	if (Mptr->R->S == 0 && Mptr->I->S == 0)
		return (1);
	else
		return (0);
}

/*
The following code implements a memory banking algorithm. Previously, MPI's
and other data structures were malloc'd() and free'd() whenever they were
required, which proved to be very time consuming. About 50% of the execution
time was spent in calls to malloc() and free().

Now, rather than free'ing when we have finished using them, we place them
on a list of free MPI's (and corresponding lists for other data structures).
When another MPI is required, if there is one of an appropriate size on the
list, it can be used (rather than calling malloc()). 

Please see more comments in the following routines.
*/
 
void COUNTBANKI()
/*
This routine is only really useful for debugging. It counts the
number of TERMI's in the corresponding bank.
*/
{
	POLYI Ptr;
	int i;

	Ptr = BANKI;
	i = 0;
	while (Ptr)
	{
	  i++;
	  Ptr = Ptr->NEXT;
	}
	printf("Number of TERMI's in BANKI is %d\n", i);
}

#ifndef DEBUG
void SAVEI(Ptr)
POLYI Ptr;
/*
This routine takes a POLYI (which is a linked list of TERMI's) that is
no longer needed, and links it onto the start of the free list BANKI,
which itself is just a linked list opf TERMI's.
*/
{
	POLYI tmp;

	tmp=Ptr;
	while(tmp->NEXT)
	  tmp=tmp->NEXT;
	tmp->NEXT = BANKI;
	BANKI = Ptr;
	Ptr = NULL;
}
#endif

POLYI CREATEI()
/*
If there is anything in BANKI (the memory bank of TERMI's) then use that,
otherwise allocate a new one.
*/
{
	POLYI Ptr;
#ifndef DEBUG
	if (BANKI)
	{
		Ptr = BANKI;
		BANKI = BANKI->NEXT;
	}
	else
#endif
		Ptr = (POLYI)mmalloc(sizeof(struct TERMI));
	return (Ptr);	
}

void COUNTBANKR()
/*
As above.
*/
{
	POLYR Ptr;
	int i;

	Ptr = BANKR;
	i = 0;
	while (Ptr)
	{
	  i++;
	  Ptr = Ptr->NEXT;
	}
	printf("Number of TERMR's in BANKR is %d\n", i);
}

#ifndef DEBUG
void SAVER(Ptr)
POLYR Ptr;
/*
As above.
*/
{
	POLYR tmp;

	tmp=Ptr;
	while(tmp->NEXT)
	{
	  tmp=tmp->NEXT;
	}
	tmp->NEXT = BANKR;
	BANKR = Ptr;
	Ptr = NULL;
}
#endif

POLYR CREATER()
/*
As above.
*/
{
	POLYR Ptr;
#ifndef DEBUG
	if (BANKR)
	{
		Ptr = BANKR;
		BANKR = BANKR->NEXT;
	}
	else
#endif
		Ptr = (POLYR)mmalloc(sizeof(struct TERMR));
	return (Ptr);	
}

#ifndef DEBUG
void SAVECI(Ptr)
POLYCI Ptr;
/*
As above.
*/
{
	POLYCI tmp;

	tmp=Ptr;
	while(tmp->NEXT)
	  tmp=tmp->NEXT;
	tmp->NEXT = BANKCI;
	BANKCI = Ptr;
	Ptr = NULL;
}
#endif

POLYCI CREATECI()
/*
As above.
*/
{
	POLYCI Ptr;
#ifndef DEBUG
	if (BANKCI)
	{
		Ptr = BANKCI;
		BANKCI = BANKCI->NEXT;
	}
	else
#endif
		Ptr = (POLYCI)mmalloc(sizeof(struct TERMCI));
	return (Ptr);	
}

#ifndef DEBUG
void SAVECR(Ptr)
POLYCR Ptr;
/*
As above.
*/
{
	POLYCR tmp;

	tmp=Ptr;
	while(tmp->NEXT)
	  tmp=tmp->NEXT;
	tmp->NEXT = BANKCR;
	BANKCR = Ptr;
	Ptr = NULL;
}
#endif

POLYCR CREATECR()
/*
As above.
*/
{
	POLYCR Ptr;
#ifndef DEBUG
	if (BANKCR)
	{
		Ptr = BANKCR;
		BANKCR = BANKCR->NEXT;
	}
	else
#endif
		Ptr = (POLYCR)mmalloc(sizeof(struct TERMCR));
	return (Ptr);	
}

#ifndef DEBUG
void SAVEm(Ptr)
POLYm Ptr;
/*
As above.
*/
{
	POLYm tmp;

	tmp=Ptr;
	while(tmp->NEXT)
	  tmp=tmp->NEXT;
	tmp->NEXT = BANKm;
	BANKm = Ptr;
	Ptr = NULL;
}
#endif

POLYm CREATEm()
/*
As above.
*/
{
	POLYm Ptr;
#ifndef DEBUG
	if (BANKm)
	{
		Ptr = BANKm;
		BANKm = BANKm->NEXT;
	}
	else
#endif
		Ptr = (POLYm)mmalloc(sizeof(struct TERMm));
	return (Ptr);	
}
