/* i4CR.c */
/* matrix operations using MPCR's */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATCR *SWAP_ROWSCR(USI p, USI q, MPMATCR *Mptr)
/*
 * interchange rows p and q of *Mptr.
 */
{
	unsigned int j;
	MPCR *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		Temp = elt(Nptr, p, j);
		elt(Nptr, p, j) = elt(Nptr, q, j);
		elt(Nptr, q, j) = Temp;
	}
	return (Nptr);
}

MPMATCR *SCALAR_ROWCR(USI p, MPCR *Aptr, MPMATCR *Mptr)
/*
 * multiplying the pth row of *Mptr by *Aptr.
 */
{
	unsigned int j;
	MPCR *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		Temp = elt(Nptr, p, j);
		elt(Nptr, p, j) = MULTCR(Temp, Aptr);
		FREEMPCR(Temp);
	}
	return (Nptr);
}

MPMATCR *ADD_MULT_ROWCR(USI p, USI q, MPCR *Aptr, MPMATCR *Mptr)
/*
 * adding *Aptr times row p to row q of *Mptr.
 */
{
	unsigned int j;
	MPCR *X, *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		X = MULTCR(elt(Nptr, p, j), Aptr);
		Temp = elt(Nptr, q, j);
		elt(Nptr, q, j) = ADDCR(Temp, X);
		FREEMPCR(X);
		FREEMPCR(Temp);
	}
	return (Nptr);
}

MPMATCR *SWAP_COLSCR(USI p, USI q, MPMATCR *Mptr)
/*
 * interchange columns p and q of *Mptr.
 */
{
	unsigned int i;
	MPCR *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		Temp = elt(Nptr, i, p);
		elt(Nptr, i, p) = elt(Nptr, i, q);
		elt(Nptr, i, q) = Temp;
	}
	return (Nptr);
}

MPMATCR *SCALAR_COLCR(USI p, MPCR *Aptr, MPMATCR *Mptr)
/*
 * multiplying the pth column of *Mptr by *Aptr.
 */
{
	unsigned int i;
	MPCR *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		Temp = elt(Nptr, i, p);
		elt(Nptr, i, p) = MULTCR(Temp, Aptr);
		FREEMPCR(Temp);
	}
	return(Nptr);
}

MPMATCR *ADD_MULT_COLCR(USI p, USI q, MPCR *Aptr, MPMATCR *Mptr)
/*
 * adding *Aptr times column p to column q of *Mptr.
 */
{
	unsigned int i;
	MPCR *X, *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		X = MULTCR(elt(Nptr, i, p), Aptr);
		Temp = elt(Nptr, i, q);
		elt(Nptr, i, q) = ADDCR(Temp, X);
		FREEMPCR(X);
		FREEMPCR(Temp);
	}
	return (Nptr);
}

MPMATCR *CONJ_TRANSPOSECR(MPMATCR *Mptr)
/* 
 * returns the transpose of *Mptr.
 */
{
	MPMATCR *Nptr;
	unsigned int i, j;

	Nptr = BUILDMATCR(Mptr->C, Mptr->R);
	for (j = 0; j <= Nptr->R - 1; j++)
		for (i = 0; i <= Nptr->C - 1; i++)
			elt(Nptr, j, i) = CONJCR(elt(Mptr, i, j));
	return (Nptr);
}

MPMATCR *DELETE_ROWCR(USI r, MPMATCR *Mptr)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	r--;
	Nptr = BUILDMATCR(Mptr->R - 1, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		if (i < r)
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				elt(Nptr, i, j) = COPYCR(elt(Mptr, i, j));
		}
		else
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				elt(Nptr, i, j) = COPYCR(elt(Mptr, i + 1, j));
		}
	}
	return (Nptr);
}

MPMATCR *DELETE_COLCR(USI c, MPMATCR *Mptr)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	c--;
	Nptr = BUILDMATCR(Mptr->R, Mptr->C - 1);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		if (j < c)
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				elt(Nptr, i, j) = COPYCR(elt(Mptr, i, j));
		}
		else
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				elt(Nptr, i, j) = COPYCR(elt(Mptr, i, j + 1));
		}
	}
	return (Nptr);
}

MPMATCR *KRONECKERCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * returns the Kronecker product of *Mptr and *Nptr.
 */
{
	unsigned int i1, i2, j1, j2, m1, m2, n1, n2;
	MPMATCR *Qptr;

	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATCR(m1 * m2, n1 * n2);
	for (i1 = 0; i1 <= m1 - 1; i1++)
	{
		for (j1 = 0; j1 <= n1 - 1; j1++)
		{
			for (i2 = 0; i2 <= m2 - 1; i2++)
			{
				for (j2 = 0; j2 <= n2 - 1; j2++)
					elt(Qptr, m2 * i1 + i2, n2 * j1 + j2) = 				MULTCR(elt(Mptr, i1, j1), elt(Nptr, i2, j2));
			}
		}
	}
	return(Qptr);
}

MPMATCR *DIRECT_SUMCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * returns the direct sum of *Mptr and *Nptr.
 */
{
	unsigned int i, j, m1, m2, n1, n2;
	MPMATCR *Qptr;
	
	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATCR(m1 + m2, n1 + n2);
	for (i = 0; i <= m1 + m2 - 1; i++)
	{
		for (j = 0; j <= n1 + n2 - 1; j++)
		{
			if (i <= m1 - 1 && j <= n1 - 1)
				elt(Qptr, i, j) =  COPYCR(elt(Mptr, i, j));
			else if (m1 <= i && n1 <= j)
				elt(Qptr, i, j) = COPYCR(elt(Nptr, i - m1, j - n1));
			else
				elt(Qptr, i, j) = ZEROCR();
		}
	}
	return (Qptr);
}

MPMATCR *JORDANCR(USI n, MPCR *Aptr)
/*
 * returns the elementary Jordan matrix of size n, diagonal elt. = *Aptr.
 */
{
	unsigned int i, j;
	MPMATCR *Mptr;
	
	Mptr = BUILDMATCR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if (i == j)
				elt(Mptr, i, j) = COPYCR(Aptr);
			else if (i == j + 1)
				elt(Mptr, i, j) = ONECR();
			else
				elt(Mptr, i, j) = ZEROCR();
		}
	}
	return (Mptr);
}

MPMATCR *ZEROMNCR(USI m, USI n)
/*
 * returns the zero  m x n matrix.
 */
{
	unsigned int i, j;
	MPMATCR *Mptr;

	Mptr = BUILDMATCR(m, n);
	for (i = 0; i <= m - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			elt(Mptr, i, j) = ZEROCR();
	}
	return (Mptr);
}

MPMATCR *BANDMATCR(MPCR *Aptr, MPCR *Bptr, USI k, USI n)
/*
 * Constructs the n x n band matrix with *Aptr on the diagonal and *Bptr
 * on the k sub and super diagonals. Here n >= 2 and 1 <= k <= n - 1.
 */
{
	unsigned int i, j;
	MPMATCR *Mptr;

	Mptr = ZEROMNCR(n, n);
	for (i = 0; i < n; i++)
	{
		FREEMPCR(elt(Mptr, i, i));
		elt(Mptr, i, i) = COPYCR(Aptr);
	}
	for (j = 1; j <= k; j++)
	{
		for (i = 0; i <= n - 1 - j; i++)
		{
			FREEMPCR(elt(Mptr, i, i + j));
			elt(Mptr, i, i + j) = COPYCR(Bptr);
			FREEMPCR(elt(Mptr, i + j, i));
			elt(Mptr, i + j, i) = COPYCR(Bptr);
		}
	}
	return (Mptr);
}

MPMATCR *IDENTITYCR(USI n)
/*
 * returns the identity matrix of size n.
 */
{
	unsigned int i, j;
	MPMATCR *Mptr;

	Mptr = BUILDMATCR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( i == j)
				elt(Mptr, i, j) = ONECR();
			else
				elt(Mptr, i, j) = ZEROCR();
		}
	}
	return (Mptr);
}

MPMATCR *SUBMATROWSCR(MPMATCR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	Nptr = BUILDMATCR(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			elt(Nptr, i, j) = COPYCR(elt(Mptr, Cptr[i], j));
	return (Nptr);
}

MPMATCR *SUBMATCOLUMNSCR(MPMATCR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	Nptr = BUILDMATCR(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			elt(Nptr, i, j) = COPYCR(elt(Mptr, i, Cptr[j]));
	return (Nptr);
}
