/* i4I.c */

/* matrix operations using MPI's */

#include <stdio.h>
#include "integer.h"
#include "fun.h"


MPMATI *SWAP_ROWSI(USI p, USI q, MPMATI *Mptr)
/*
 * interchange rows p and q of *Mptr.
 */
{
	unsigned int j;
	MPI *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		Temp = elt(Nptr, p, j);
		elt(Nptr, p, j) = elt(Nptr, q, j);
		elt(Nptr, q, j) = Temp;
	}
	return (Nptr);
}

MPMATI *SCALAR_ROWI(USI p, MPI *Aptr, MPMATI *Mptr)
/*
 * multiplying the pth row of *Mptr by *Aptr.
 */
{
	unsigned int j;
	MPI *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		Temp = elt(Nptr, p, j);
		elt(Nptr, p, j) = MULTI(Temp, Aptr);
		FREEMPI(Temp);
	}
	return (Nptr);
}

MPMATI *ADD_MULT_ROWI(USI p, USI q, MPI *Aptr, MPMATI *Mptr)
/*
 * adding *Aptr times row p to row q of *Mptr.
 */
{
	unsigned int j;
	MPI *X, *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		X = MULTI(elt(Nptr, p, j), Aptr);
		Temp = elt(Nptr, q, j);
		elt(Nptr, q, j) = ADDI(Temp, X);
		FREEMPI(X);
		FREEMPI(Temp);
	}
	return (Nptr);
}

MPMATI *SWAP_COLSI(USI p, USI q, MPMATI *Mptr)
/*
 * interchange columns p and q of *Mptr.
 */
{
	unsigned int i;
	MPI *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		Temp = elt(Nptr, i, p);
		elt(Nptr, i, p) = elt(Nptr, i, q);
		elt(Nptr, i, q) = Temp;
	}
	return (Nptr);
}

MPMATI *SCALAR_COLI(USI p, MPI *Aptr, MPMATI *Mptr)
/*
 * multiplying the pth column of *Mptr by *Aptr.
 */
{
	unsigned int i;
	MPI *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		Temp = elt(Nptr, i, p);
		elt(Nptr, i, p) = MULTI(Temp, Aptr);
		FREEMPI(Temp);
	}
	return(Nptr);
}

MPMATI *ADD_MULT_COLI(USI p, USI q, MPI *Aptr, MPMATI *Mptr)
/*
 * adding *Aptr times column p to column q of *Mptr.
 */
{
	unsigned int i;
	MPI *X, *Temp;
	MPMATI *Nptr;

	Nptr = COPYMATI(Mptr);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		X = MULTI(elt(Nptr, i, p), Aptr);
		Temp = elt(Nptr, i, q);
		elt(Nptr, i, q) = ADDI(Temp, X);
		FREEMPI(X);
		FREEMPI(Temp);
	}
	return (Nptr);
}

MPMATI *TRANSPOSEI(MPMATI *Mptr)
/* 
 * returns the transpose of *Mptr.
 */
{
	MPMATI *Nptr;
	unsigned int i, j;

	Nptr = BUILDMATI(Mptr->C, Mptr->R);
	for (j = 0; j <= Nptr->R - 1; j++)
		for (i = 0; i <= Nptr->C - 1; i++)
			elt(Nptr, j, i) = COPYI(elt(Mptr, i, j));
	return (Nptr);
}

MPMATI *DELETE_ROWI(USI r, MPMATI *Mptr)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	r--;
	Nptr = BUILDMATI(Mptr->R - 1, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		if (i < r)
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				elt(Nptr, i, j) = COPYI(elt(Mptr, i, j));
		}
		else
		{
			for (j = 0; j <= Nptr->C - 1; j++)
				elt(Nptr, i, j) = COPYI(elt(Mptr, i + 1, j));
		}
	}
	return (Nptr);
}

MPMATI *DELETE_COLI(USI c, MPMATI *Mptr)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	c--;
	Nptr = BUILDMATI(Mptr->R, Mptr->C - 1);
	for (j = 0; j <= Nptr->C - 1; j++)
	{
		if (j < c)
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				elt(Nptr, i, j) = COPYI(elt(Mptr, i, j));
		}
		else
		{
			for (i = 0; i <= Nptr->R - 1; i++)
				elt(Nptr, i, j) = COPYI(elt(Mptr, i, j + 1));
		}
	}
	return (Nptr);
}

MPMATI *KRONECKERI(MPMATI *Mptr, MPMATI *Nptr)
/*
 * returns the Kronecker product of *Mptr and *Nptr.
 */
{
	unsigned int i1, i2, j1, j2, m1, m2, n1, n2;
	MPMATI *Qptr;

	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATI(m1 * m2, n1 * n2);
	for (i1 = 0; i1 <= m1 - 1; i1++)
	{
		for (j1 = 0; j1 <= n1 - 1; j1++)
		{
			for (i2 = 0; i2 <= m2 - 1; i2++)
			{
				for (j2 = 0; j2 <= n2 - 1; j2++)
					elt(Qptr, m2 * i1 + i2, n2 * j1 + j2) = MULTI(elt(Mptr, i1, j1), elt(Nptr, i2, j2));
			}
		}
	}
	return(Qptr);
}

MPMATI *DIRECT_SUMI(MPMATI *Mptr, MPMATI *Nptr)
/*
 * returns the direct sum of *Mptr and *Nptr.
 */
{
	unsigned int i, j, m1, m2, n1, n2;
	MPMATI *Qptr;
	
	m1 = Mptr->R;
	m2 = Nptr->R;
	n1 = Mptr->C;
	n2 = Nptr->C;
	Qptr = BUILDMATI(m1 + m2, n1 + n2);
	for (i = 0; i <= m1 + m2 - 1; i++)
	{
		for (j = 0; j <= n1 + n2 - 1; j++)
		{
			if (i <= m1 - 1 && j <= n1 - 1)
				elt(Qptr, i, j) =  COPYI(elt(Mptr, i, j));
			else if (m1 <= i && n1 <= j)
				elt(Qptr, i, j) = COPYI(elt(Nptr, i - m1, j - n1));
			else
				elt(Qptr, i, j) = ZEROI();
		}
	}
	return (Qptr);
}

MPMATI *JORDANI(USI n, MPI *Aptr)
/*
 * returns the elementary Jordan matrix of size n, diagonal elt. = *Aptr.
 */
{
	unsigned int i, j;
	MPMATI *Mptr;
	
	Mptr = BUILDMATI(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if (i == j)
				elt(Mptr, i, j) = COPYI(Aptr);
			else if (i == j + 1)
				elt(Mptr, i, j) = ONEI();
			else
				elt(Mptr, i, j) = ZEROI();
		}
	}
	return (Mptr);
}

MPMATI *ZEROMNI(USI m, USI n)
/*
 * returns the zero  m x n matrix.
 */
{
	unsigned int i, j;
	MPMATI *Mptr;

	Mptr = BUILDMATI(m, n);
	for (i = 0; i <= m - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			elt(Mptr, i, j) = ZEROI();
	}
	return (Mptr);
}

MPMATI *BANDMATI(MPI *Aptr, MPI *Bptr, USI k, USI n)
/*
 * Constructs the n x n band matrix with *Aptr on the diagonal and *Bptr
 * on the k sub and super diagonals. Here n >= 2 and 1 <= k <= n - 1.
 */
{
	unsigned int i, j;
	MPMATI *Mptr;

	Mptr = ZEROMNI(n, n);
	for (i = 0; i < n; i++)
	{
		FREEMPI(elt(Mptr, i, i));
		elt(Mptr, i, i) = COPYI(Aptr);
	}
	for (j = 1; j <= k; j++)
	{
		for (i = 0; i <= n - 1 - j; i++)
		{
			FREEMPI(elt(Mptr, i, i + j));
			elt(Mptr, i, i + j) = COPYI(Bptr);
			FREEMPI(elt(Mptr, i + j, i));
			elt(Mptr, i + j, i) = COPYI(Bptr);
		}
	}
	return (Mptr);
}

MPMATI *IDENTITYI(USI n)
/*
 * returns the identity matrix of size n.
 */
{
	unsigned int i, j;
	MPMATI *Mptr;

	Mptr = BUILDMATI(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( i == j)
				elt(Mptr, i, j) = ONEI();
			else
				elt(Mptr, i, j) = ZEROI();
		}
	}
	return (Mptr);
}

MPMATI *SUBMATROWSI(MPMATI *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	Nptr = BUILDMATI(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			elt(Nptr, i, j) = COPYI(elt(Mptr, Cptr[i], j));
	return (Nptr);
}

MPMATI *SUBMATCOLUMNSI(MPMATI *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATI *Nptr;

	Nptr = BUILDMATI(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			elt(Nptr, i, j) = COPYI(elt(Mptr, i, Cptr[j]));
	return (Nptr);
}
