/* i5CI.c */
/* programs for inputting and printing MPI's. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "integer.h"
#include "fun.h"

MPCI *BUILDMPCI( )
/*
 * mallocs space for an MPCI.
 */
{
	MPCI *Mptr;

	Mptr = (MPCI *)mmalloc(sizeof(MPCI));
	return (Mptr);
}

void FREEMPCI(MPCI *Mptr)
/*
 * deallocates the space alloted for the MPCI Mptr.
 */
{
	FREEMPI(Mptr->R);
	FREEMPI(Mptr->I);
	ffree ((char *)Mptr, sizeof(MPCI));
	return;
}

MPCI *COPYCI(MPCI *Aptr)
/*
 * returns a pointer to a copy of *Aptr.
 */
{
	MPCI *Bptr;

	Bptr = BUILDMPCI();
	Bptr->R = COPYI(Aptr->R);
	Bptr->I = COPYI(Aptr->I);
	return (Bptr);
}

MPCI *MINUSCI(MPCI *Aptr)
/*
 * *Bptr = -(*Aptr).
 */
{
	MPCI *Bptr;

	Bptr = BUILDMPCI();
	Bptr->R = MINUSI(Aptr->R);
	Bptr->I = MINUSI(Aptr->I);
	return (Bptr);
}

MPCI *INPUTCI(char **ptr)
/*
 * inputs the complex integer *Mptr.
 */
{
	MPCI *Mptr;
	char t;
	unsigned int u;

	Mptr = BUILDMPCI();
	while ((**ptr != '+') && (**ptr != '-') && (**ptr != 'i') && (**ptr < '0' || **ptr > '9'))
		(*ptr)++;
	t = **ptr;
	if (t == '-')
	{
		(*ptr)++;
		if (**ptr == 'i')
		{
			(*ptr)++;
			Mptr->R = ZEROI();
			Mptr->I = MINUS_ONEI();
			return (Mptr);
		}
		else
		{
			(*ptr)--;
			Mptr->R = INPUTSI(ptr, &u);
		}
	}
	else if (t == 'i')
	{
		(*ptr)++;
		Mptr->R = ZEROI();
		Mptr->I = ONEI();
		return (Mptr);
	}
	else
		Mptr->R = INPUTSI(ptr, &u);
	if (**ptr == 'i')
	{
		(*ptr)++;
		Mptr->I = Mptr->R;
		Mptr->R = ZEROI();
		return (Mptr);
	}
	else if (**ptr == '\0')
	{
		(*ptr)++;
		Mptr->I = ZEROI();
		return (Mptr);
	}
	else if (**ptr == ' ')
	{
		while (**ptr == ' ')
			(*ptr)++;
		if (**ptr == '\0')
		{
			Mptr->I = ZEROI();
			return (Mptr);
		}
		else if (**ptr == '-')
		{
			(*ptr)++;
			if (**ptr != ' ')
			{
				(*ptr)--;
				Mptr->I = ZEROI();
				return (Mptr);
			}
			else 
				(*ptr)--;
		}
		else if (((**ptr >= '0') && (**ptr <= '9')) || **ptr == 'i')
		{
			(*ptr)--;
			Mptr->I = ZEROI();
			return (Mptr);
		}
	}
	/* **ptr == '+' or **ptr == '-' (addition or subtraction). */
	t = **ptr;
	(*ptr)++;
	while (**ptr == ' ')
		(*ptr)++;
	if (**ptr == 'i')
	{
		if (t == '+')
		{
			(*ptr)++;
			Mptr->I = ONEI(); 
			return (Mptr);
		}
		else 
		{
			(*ptr)++;
			Mptr->I = MINUS_ONEI(); 
			return (Mptr);
		}
	}
	else
	{
		Mptr->I = INPUTSI(ptr, &u);
		if (t == '-')
			(Mptr->I)->S = -1;
		(*ptr)++;
	}
	return (Mptr);
}

MPCI *ZEROCI()
/*
 * Returns the MPCI ZERO.
 */
{
	MPCI *Eptr;

	Eptr = BUILDMPCI();
	Eptr->R = ZEROI();
	Eptr->I = ZEROI();
	return Eptr;
}

MPCI *ONECI()
/*
 * Returns the MPCI ONE.
 */
{
	MPCI *Eptr;

	Eptr = BUILDMPCI();
	Eptr->R = ONEI();
	Eptr->I = ZEROI();
	return Eptr;
}

MPMATCI *ZEROMNCI(USI m, USI n)
/*
 * returns the zero  m x n matrix.
 */
{
	unsigned int i, j;
	MPMATCI *Mptr;

	Mptr = BUILDMATCI(m, n);
	for (i = 0; i <= m - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			elt(Mptr, i, j) = ZEROCI();
	}
	return (Mptr);
}

MPMATCI *FINPUTMATCI(FILE *infile)
{
	MPMATCI *Temp;
	unsigned int i, j, m, n;
	char *line, *p;

	if (infile == stdin)
		printf("Enter the number of rows and number of columns\n");

	if (fscanf(infile, "%u%u", &m, &n) != 2 || m == 0 || n == 0)
	{
		fprintf(stderr, "corrupt input: returning the 1 x 1 zero matrix\n");
		if (infile != stdin)
			exit (1);
		return ZEROMNCI(1, 1);
	}
	fgetc(infile);
	Temp = ZEROMNCI(m, n);
	for (i = 0; i < m; i++)
	{
		if (infile == stdin)
			printf("Enter row %u:\n", i + 1);
		p = line = Fgets(infile);
		for (j = 0; j < n; j++)
		{
			FREEMPCI(elt(Temp, i, j));
			elt(Temp, i, j) = INPUTCI(&p);
		}
		free(line);
	}
	return Temp;
}

MPMATCI *INPUTMATCI()
/*
 * Inputting a matrix of complex integers.
 */
{
	return FINPUTMATCI(stdin);
}

void FPRINTCI(FILE *outfile, MPCI *Mptr)
/*
 * the MPCI *Mptr is printed to *outfile in decimal notation. 
 */
{
	int s;

	if ((Mptr->R)->S != 0)
	{
		FPRINTI(outfile, Mptr->R);
	}
	else
	{
		if ((Mptr->I)->S == -1 && (Mptr->I)->D == 0 && (Mptr->I)->V[0] == 1)
		{
			fprintf(outfile, "-i");
			return;
		}
		else if ((Mptr->I)->S == 1 && (Mptr->I)->D == 0 && (Mptr->I)->V[0] == 1)
		{
			fprintf(outfile, "i");
			return;
		}
		else if ((Mptr->I)->S == 0)
		{
			fprintf(outfile, "0");
			return;
		}
		else
		{
			FPRINTI(outfile, Mptr->I);
			fprintf(outfile, "i");
			return;
		}
	}
	if ((Mptr->I)->S == 0)
		return;
	else
	{
		s = (Mptr->I)->S;
		if (s == -1)
		{
			fprintf(outfile, " - ");
			(Mptr->I)->S = 1;
		}
		else 
			fprintf(outfile, " + ");
		if (!EQONEI(Mptr->I))
			FPRINTI(outfile, Mptr->I);
		if (s == -1)
			(Mptr->I)->S = -1;
		fprintf(outfile, "i");
		return;
	}
}

void PRINTCI(MPCI *Mptr)
/*
 * the MPCI *Mptr is printed to stdout in decimal notation. 
 */
{
	FPRINTCI(stdout, Mptr);
	return;
}

void SPRINTCI(char *buffer, MPCI *Mptr)
/*
 * the MPCI *Mptr is printed to buffer in decimal notation. 
 */
{
	int s;

	if ((Mptr->R)->S != 0)
	{
		SPRINTI(buffer, Mptr->R);
		buffer += strlen(buffer);
	}
	else
	{
		if ((Mptr->I)->S == -1 && (Mptr->I)->D == 0 && (Mptr->I)->V[0] == 1)
		{
			sprintf(buffer, "-i");
			return;
		}
		else if ((Mptr->I)->S == 1 && (Mptr->I)->D == 0 && (Mptr->I)->V[0] == 1)
		{
			sprintf(buffer, "i");
			return;
		}
		else if ((Mptr->I)->S == 0)
		{
			sprintf(buffer, "0");
			return;
		}
		else
		{
			SPRINTI(buffer, Mptr->I);
			buffer += strlen(buffer);
			sprintf(buffer, "i");
			return;
		}
	}
	if ((Mptr->I)->S == 0)
		return;
	else
	{
		s = (Mptr->I)->S;
		if (s == -1)
		{
			sprintf(buffer, " - ");
			buffer += strlen(buffer);
			(Mptr->I)->S = 1;
		}
		else 
		{
			sprintf(buffer, " + ");
			buffer += strlen(buffer);
		}
		if (!EQONEI(Mptr->I))
		{
			SPRINTI(buffer, Mptr->I);
			buffer += strlen(buffer);
		}
		if (s == -1)
			(Mptr->I)->S = -1;
		sprintf(buffer, "i");
		return;
	}
}

unsigned int LENGTHCI(MPCI *Aptr)
/*
 * length of a complex integer.
 */
{
	if ((Aptr->I)->S == 0)
		return (LENGTHI(Aptr->R));
	else 
	{
		if ((Aptr->R)->S == 0)
		{
			if ((Aptr->I)->S == 1 && (Aptr->I)->D == 0 && (Aptr->I)->V[0] == 1)
				return (1);
			else if ((Aptr->I)->S == -1 && (Aptr->I)->D == 0 && (Aptr->I)->V[0] == 1)
				return (2);
			else
				return (1 + LENGTHI(Aptr->I));
		}
		else
		{
			if ((Aptr->I)->S == 1 && (Aptr->I)->D == 0 && (Aptr->I)->V[0] == 1)
				return (4 + LENGTHI(Aptr->R));
			else if ((Aptr->I)->S == 1 && (Aptr->I)->D == 0 && (Aptr->I)->V[0] == 1)
				return (4 + LENGTHI(Aptr->R));
			else if ((Aptr->I)->S == 1)
				return (4 + LENGTHI(Aptr->I) + LENGTHI(Aptr->R));
			else 
				return (3 + LENGTHI(Aptr->I) + LENGTHI(Aptr->R));
		}
	}
}

void PRINTMATCI(USI i1, USI i2, USI j1, USI j2, MPMATCI *Mptr)
/*
 * prints *Mptr from rows i1 to i2 and cols j1 to j2.
 * Modified on 8/1/93 using Peter Adams' modification of August 1992.
 */
{
	char **str, *buff;
	unsigned int tmp, i, j, nrow, ncol, len, nstr = 0;
	int *colwidth;
	unsigned int ct;
	
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	if (MAX((int)nrow, (int)ncol) > 27)
	{
		printf("(The number of rows or columns to be printed exceeds 20;\n");
		printf("there is no point in printing this matrix on the screen.)\n");
		return;
	}
	str = (char **)mmalloc(nrow * ncol * sizeof(char *));
	colwidth=(int *)mmalloc(ncol*sizeof(int));

	for(i=0; i<ncol; i++)
	  colwidth[i]=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			tmp = 1 + LENGTHCI(elt(Mptr, i, j));
			buff = (char *)mmalloc(tmp * sizeof(char));
			SPRINTCI(buff, elt(Mptr, i, j));
		    	if ((len = strlen(buff)) > colwidth[j-j1])
				colwidth[j-j1] = len;
			str[nstr++] = strcpy((char *)mmalloc(len + 1), buff);
			ffree(buff, tmp * sizeof(char));
		}
	}
	ct=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
	    		printf("%*s ", colwidth[j-j1], str[ct]);
	    		ffree(str[ct], (unsigned int) (strlen(str[ct]) + 1) * sizeof(char));
	    		if ((ct % ncol) == (ncol - 1))
				printf("\n");
	    		ct++;
		}
	}
	ffree((char *) str, nrow * ncol * sizeof(char *));
	ffree((char *) colwidth, ncol*sizeof(int));
	return;
}

unsigned int EQUALCI(MPCI *Aptr, MPCI *Bptr)
/*
 * returns 1 if *Aptr = *Bptr, 0 otherwise.
 */
{
	if (EQUALI(Aptr->R, Bptr->R) * EQUALI(Aptr->I, Bptr->I))
		return (1);
	else
		return (0);
}

MPCI *ADDCI(MPCI *Aptr, MPCI *Bptr)
/*
 * *Cptr = *Aptr + *Bptr.
 */
{
	MPCI *Cptr;

	Cptr = BUILDMPCI();
	Cptr->R = ADDI(Aptr->R, Bptr->R);
	Cptr->I = ADDI(Aptr->I, Bptr->I);
	return (Cptr);
}

MPCI *MULTCI(MPCI *Aptr, MPCI *Bptr)
/*
 * *Cptr = *Aptr * *Bptr.
 */
{
	MPI *U, *V, *F, *G;
	MPCI *Cptr;

	Cptr = BUILDMPCI();
	U = MULTI(Aptr->R, Bptr->R);
	V = MULTI(Aptr->I, Bptr->I);
	F = MULTI(Aptr->R, Bptr->I);
	G = MULTI(Aptr->I, Bptr->R);
	Cptr->R = SUBI(U, V);
	Cptr->I = ADDI(F, G);
	FREEMPI(U);
	FREEMPI(V);
	FREEMPI(F);
	FREEMPI(G);
	return (Cptr);
}

MPCI *POWERCI(MPCI *Aptr, USI n)
/*
 *  *Eptr = (*Aptr) ^ n, where 0 <= n < R0 * R0.
 */
{
	MPCI *B, *TempCI, *Eptr;

	if (n == 0)
		return ONECI();
	B = COPYCI(Aptr);
	Eptr = ONECI();
	while (1)
        {
		if (n & 1)
	        {
			TempCI = Eptr;
			Eptr = MULTCI(Eptr, B);
			FREEMPCI(TempCI);
			if (n == 1)
			{
				FREEMPCI(B);	
				return (Eptr);
			}
		}
		TempCI = B;
		B = MULTCI(B, B);
		FREEMPCI(TempCI);
		n = n >> 1;
	}
}

MPCI *SUBCI(MPCI *Aptr, MPCI *Bptr)
/*
 * *Cptr = *Aptr - *Bptr.
 */
{
	MPCI *E, *Cptr;

	E = MINUSCI(Bptr);
	Cptr = ADDCI(Aptr, E);
	FREEMPCI(E);
	return (Cptr);
}
