/* i6CI.c */
/* matrix operations using MPCI's */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

MPMATCI *BUILDMATCI(USI m, USI n)
/*
 * Allocates space for an m x n matrix of MPCI's.
 */
{
	MPMATCI *N;

	N=(MPMATCI *)mmalloc(sizeof(MPMATCI));
	N->R = m;
	N->C = n;
	N->V = (MPCI **)mmalloc(m * n * sizeof(MPCI *));
	return (N);
}

MPMATCI *COPYMATCI(MPMATCI *Mptr)
/*
 * a replacement for the assignment *Nptr = *Mptr.
 */
{
	unsigned int i, t;
	MPMATCI *Nptr;

	t = Mptr->R * Mptr->C;
	Nptr = BUILDMATCI(Mptr->R, Mptr->C);
	for (i = 0; i < t; i++)
			Nptr->V[i] = COPYCI(Mptr->V[i]);
	return (Nptr);
}

void FREEMATCI(MPMATCI *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V.
 */
{
	unsigned int i, t;

	if (Mptr == NULL)
		return;
	t = Mptr->R * Mptr->C;
	for (i = 0; i < t; i++)
		FREEMPCI(Mptr->V[i]);
	ffree((char *)(Mptr->V), t *  sizeof(MPCI *));
	ffree((char *)Mptr, sizeof(MPMATCI));
	return;
}

MPCI *TRACECI(MPMATCI *Mptr)
/*
 * *Aptr is the trace of *Mptr.
 */
{
	unsigned int i;
	MPCI *Aptr, *Temp;

	Aptr = ZEROCI();
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Temp = Aptr;
		Aptr = ADDCI(Aptr, elt(Mptr, i, i));
		FREEMPCI(Temp);
	}
	return (Aptr);
}

MPMATCI *SCALARMATCI(MPMATCI *Mptr, MPCI *Aptr)
/*
 * returns (*Aptr) * (*Mptr).
 */
{
	unsigned int i, j;
	MPMATCI *Nptr;

	Nptr = BUILDMATCI(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MULTCI(elt(Mptr, i, j), Aptr);
	}
	return (Nptr);
}

MPMATCI *MINUSMATCI(MPMATCI *Mptr)
/*
 * returns -(*Mptr).
 */
{
	unsigned int i, j;
	MPMATCI *Nptr;

	Nptr = BUILDMATCI(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MINUSCI(elt(Mptr, i, j));
	}
	return (Nptr);
}

MPMATCI *DIAGCI(MPMATCI *Mptr, MPCI *Aptr)
/*
 * returns *Mptr - (*Aptr) * (Unit Matrix).
 */
{
	unsigned int i;
	MPCI *Temp;
	MPMATCI *Nptr;

	Nptr = COPYMATCI(Mptr);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
			Temp = elt(Nptr, i, i);
			elt(Nptr, i, i) = SUBCI(elt(Mptr, i, i), Aptr);
			FREEMPCI(Temp);
	}
	return (Nptr);
}

MPMATCI *ADDMATCI(MPMATCI *Mptr, MPMATCI *Nptr)
/*
 * returns *Mptr + *Nptr.
 */
{
	unsigned int i, j;
	MPMATCI *Lptr;

	Lptr = BUILDMATCI(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = ADDCI(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATCI *SUBMATCI(MPMATCI *Mptr, MPMATCI *Nptr)
/*
 * returns *Mptr - *Nptr.
 */
{
	unsigned int i, j;
	MPMATCI *Lptr;

	Lptr = BUILDMATCI(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = SUBCI(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATCI *MULTMATCI(MPMATCI *Mptr, MPMATCI *Nptr)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr).
 */
{
	MPCI *X, *Y, *Temp;
	unsigned int i, j, k;
	MPMATCI *Lptr;

	Lptr = BUILDMATCI(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			X = ZEROCI();
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				Y = MULTCI(elt(Mptr, i, k), elt(Nptr, k, j));
				Temp = X;
				X = ADDCI(X, Y);
				FREEMPCI(Temp);
				FREEMPCI(Y);
			}
			elt(Lptr, i, j) = X;
		}
	}
	return (Lptr);
}
