/* i6CR.c */
/* matrix operations using MPCR's */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

MPMATCR *BUILDMATCR(USI m, USI n)
/*
 * Allocates space for an m x n matrix of MPCR's.
 */
{
	MPMATCR *N;

	N=(MPMATCR *)mmalloc(sizeof(MPMATCR));
	N->R = m;
	N->C = n;
	N->V = (MPCR **)mmalloc(m * n * sizeof(MPCR *));
	return (N);
}

MPMATCR *COPYMATCR(MPMATCR *Mptr)
/*
 * a replacement for the assignment *Nptr = *Mptr.
 */
{
	unsigned int i, t;
	MPMATCR *Nptr;

	t = Mptr->R * Mptr->C;
	Nptr = BUILDMATCR(Mptr->R, Mptr->C);
	for (i = 0; i < t; i++)
			Nptr->V[i] = COPYCR(Mptr->V[i]);
	return (Nptr);
}

void FREEMATCR(MPMATCR *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V.
 */
{
	unsigned int i, t;

	if (Mptr == NULL)
		return;
	t = Mptr->R * Mptr->C;
	for (i = 0; i < t; i++)
		FREEMPCR(Mptr->V[i]);
	ffree((char *)(Mptr->V), t * sizeof(MPCR *));
	ffree((char *)Mptr, sizeof(MPMATCR));
	return;
}

MPMATCR *UNWRAPMATCR(MPMATCR *Mptr)
/*
 * unwraps the matrix *Mptr into a long column vector.
 */
{
	unsigned int p, k;
	MPMATCR *Nptr;

	p = Mptr->R * Mptr->C;
	Nptr = BUILDMATCR(p, 1);
	for (k = 0; k <= p - 1; k++)
	elt(Nptr, k, 0) = COPYCR(elt(Mptr, k /( Mptr->C), k % (Mptr->C)));
	return (Nptr);
}

unsigned int ZEROTESTCR(MPMATCR *Mptr)
/*
 * returns 1 if *Mptr is the ZERO matrix, 0 otherwise.
 */
{
	unsigned int i, j;

	for (i = 0; i < Mptr->R; i++)
	{
		for (j = 0; j < Mptr->C; j++)
		{
			if (!EQZEROCR(elt(Mptr, i, j)))
				return (0);
		}
	}
	return (1);
}

unsigned int PIVOTCR(USI i, USI j, MPMATCR *Mptr)
/*
 * here 0 <= i <= j <= Mptr->R - 1.
 * returns the row index k, with least k (i <= k <= Mptr->R - 1), such that 
 * Mptr->V[k][j] is non-zero; returns Mptr->R otherwise.
 */
{
	unsigned int k;
	
	k = i;

	while (k <= Mptr->R - 1 && ((elt(Mptr, k, j)->R)->N)->S == 0 && ((elt(Mptr, k, j)->I)->N)->S == 0)
		k++;
	return (k);
}

void TRANSCR(USI p, USI q, MPMATCR *Mptr)
/*
 * 0 <= p < q <= Mprt->R - 1.
 * Mptr->V[i][j] = 0 for Mptr->R - 1 >= i >= p and 0 <= j < q 
 * transposes the p-th and q-th rows of *Mptr.
 */
{
	MPCR *X;
	unsigned int j;

	for (j = p; j <= Mptr->C - 1; j++) 
	{
		X = elt(Mptr, p, j);
		elt(Mptr, p, j) = elt(Mptr, q, j);
		elt(Mptr, q, j) = X;
	}
	return;
}

void NORMCR(USI p, USI j, MPMATCR *Mptr)
/*
 * divides the p-th row by Mptr->V[p][j].
 * 0 <= p <= Mprt->R - 1.
 * 0 <= j <= Mprt->C - 1.
 * Mptr->V[p][k] = 0 if k < j.
 */
{
	MPCR *X, *TempCR;
	unsigned int k;

	X = COPYCR(elt(Mptr, p, j));
	TempCR = elt(Mptr, p, j);
	elt(Mptr, p, j) = ONECR();
	FREEMPCR(TempCR);
	for (k = j + 1; k <= Mptr->C - 1; k++) 
	{
		TempCR = elt(Mptr, p, k);
		elt(Mptr, p, k) = RATIOCR(TempCR, X);
		FREEMPCR(TempCR);
	}
	FREEMPCR(X);
	return;
}

void SUBMCR(USI p, USI q, MPCR *Aptr, USI j, MPMATCR *Mptr)
/*
 * subtracts *Aptr times the p-th row of *Mptr from the q-th,
 * for columns Mptr->C - 1 >= k > j.
 * 0 <= p, q <= Mprt->R - 1.
 * Mptr->V[p][k] = 0 if k < j, Mptr->V[p][j] = 1.
 */
{
	MPCR *Temp1CR, *Temp2CR;
	unsigned int k;

	for (k = j + 1; k <= Mptr->C - 1; k++) 
	{
		Temp1CR = MULTCR(Aptr, elt(Mptr, p, k));
		Temp2CR = elt(Mptr, q, k);
		elt(Mptr, q, k) = SUBCR(Temp2CR, Temp1CR);
		FREEMPCR(Temp2CR);
		FREEMPCR(Temp1CR);
	}
	return;
}

MPCR *TRACECR(MPMATCR *Mptr)
/*
 * *Aptr is the trace of *Mptr.
 */
{
	unsigned int i;
	MPCR *Aptr, *Temp;

	Aptr = ZEROCR();
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Temp = Aptr;
		Aptr = ADDCR(Aptr, elt(Mptr, i, i));
		FREEMPCR(Temp);
	}
	return (Aptr);
}

MPMATCR *SCALARMATCR(MPMATCR *Mptr, MPCR *Aptr)
/*
 * returns (*Aptr) * (*Mptr).
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	Nptr = BUILDMATCR(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MULTCR(elt(Mptr, i, j), Aptr);
	}
	return (Nptr);
}

MPMATCR *MINUSMATCR(MPMATCR *Mptr)
/*
 * returns -(*Mptr).
 */
{
	unsigned int i, j;
	MPMATCR *Nptr;

	Nptr = BUILDMATCR(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			elt(Nptr, i, j) = MINUSCR(elt(Mptr, i, j));
	}
	return (Nptr);
}

MPMATCR *DIAGCR(MPMATCR *Mptr, MPCR *Aptr)
/*
 * returns *Mptr - (*Aptr) * (Unit Matrix).
 */
{
	unsigned int i;
	MPCR *Temp;
	MPMATCR *Nptr;

	Nptr = COPYMATCR(Mptr);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
			Temp = elt(Nptr, i, i);
			elt(Nptr, i, i) = SUBCR(elt(Mptr, i, i), Aptr);
			FREEMPCR(Temp);
	}
	return (Nptr);
}

MPMATCR *ADDMATCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * returns *Mptr + *Nptr.
 */
{  
	unsigned int i, j;
	MPMATCR *Lptr;

	Lptr = BUILDMATCR(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = ADDCR(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATCR *SUBMATCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * returns *Mptr - *Nptr.
 */
{
	unsigned int i, j;
	MPMATCR *Lptr;

	Lptr = BUILDMATCR(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
		for (j = 0; j <= Lptr->C - 1; j++)
			elt(Lptr, i, j) = SUBCR(elt(Mptr, i, j), elt(Nptr, i, j));
	return (Lptr);
}

MPMATCR *MULTMATCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr).
 */
{
	MPCR *X, *Y, *Temp;
	unsigned int i, j, k;
	MPMATCR *Lptr;

	Lptr = BUILDMATCR(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			X = ZEROCR();
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				Y = MULTCR(elt(Mptr, i, k), elt(Nptr, k, j));
				Temp = X;
				X = ADDCR(X, Y);
				FREEMPCR(Temp);
				FREEMPCR(Y);
			}
			elt(Lptr, i, j) = X;
		}
	}
	return (Lptr);
}

MPMATCR *POWERMATCR(MPMATCR *Aptr, USI n)
/*
 * returns (*Aptr) ^ n, where 0 <= n < R0 * R0.
 */
{
	MPMATCR *Temp, *B, *Eptr;

	Eptr = IDENTITYCR(Aptr->R);
	if (n == 0)
		return (Eptr);
	B = COPYMATCR(Aptr);
	while (1)
        {
		if (n & 1) 
		{
			Temp = Eptr;
			Eptr = MULTMATCR(Temp, B);
			FREEMATCR(Temp);
			if (n == 1)
			{
				FREEMATCR(B);
				return (Eptr);
			}
		}
		Temp = B;
		B = MULTMATCR(Temp, Temp);
		FREEMATCR(Temp);
		n = n >> 1;
	}
}

MPMATCR *JCOLSMATCR(MPMATCR *Mptr, MPMATCR *Nptr)
/*
 * returns [ *Mptr | *Nptr]. 
 */
{
	unsigned int i, j;
	MPMATCR *Lptr;

	Lptr = BUILDMATCR(Mptr->R, Mptr->C + Nptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
	{
		for (j = 0; j <= Lptr->C - 1; j++)
		{
			if (j <= Mptr->C - 1)
				elt(Lptr, i, j) = COPYCR(elt(Mptr, i, j));
			else
				elt(Lptr, i, j) = COPYCR(elt(Nptr, i, j - Mptr->C));
		}
	}
	return (Lptr);
}

MPMATCR *JROWSMATCR(MPMATCR *Mptr, MPMATCR *Nptr)
/* 
 *         [ *Mptr]
 * returns [------]
 *         [ *Nptr]
 */
{
	unsigned int i, j;
	MPMATCR *Lptr;

	Lptr = BUILDMATCR(Mptr->R + Nptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
	{
		if (i <= Mptr->R - 1)
		{
			for (j = 0; j <= Lptr->C - 1; j++)
				elt(Lptr, i, j) = COPYCR(elt(Mptr, i, j));
		}
		else
		{
			for (j = 0; j <=  Lptr->C - 1; j++)
				elt(Lptr, i, j) = COPYCR(elt(Nptr, i - Mptr->R, j));
		}
	}
	return (Lptr);
}

MPMATCR *JCOLSIDENTITYCR(MPMATCR *Mptr)
/*
 * forms the partioned matrix [*Mptr | I].
 */
{
	MPMATCR *Temp, *Nptr;

	Temp = IDENTITYCR(Mptr->R);
	Nptr = JCOLSMATCR(Mptr, Temp);
	FREEMATCR(Temp);
	return Nptr;
}

MPMATCR *GAUSS_JORDANCR(MPMATCR *Mptr, USI *rptr, USI **Cptr)
/*
 * returns the reduced row-echelon form of *Mptr, 
 * *rptr = rank *Mptr. 
 */
{
	unsigned int i, j, k, m, n, p, r, *A;
	MPCR *TempCR;
	MPMATCR *Nptr;

	m = Mptr->R;
	n = Mptr->C;
	A = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	for (i = 0; i <= n - 1; i++)
		A[i] = 1;
	Nptr = COPYMATCR(Mptr);
	i = 0;
	j = 0;
	*Cptr = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	while (1) 
	{
		while ((j <= n - 1) && (PIVOTCR(i, j, Nptr) == m))
			j++;
		if (j == n)
			break;
		else 
		{
			p = PIVOTCR(i, j, Nptr);
			if (p > i)
				TRANSCR(i, p, Nptr);
			if (!EQONECR(elt(Nptr, i, j)))
				NORMCR(i, j, Nptr);
			*(*Cptr + i) = j;
			A[j] = 0;
			for (k = 0; k <= m - 1; k++)
		        {
				if (k != i) 
				{
					if (!EQZEROCR(elt(Nptr, k, j)))
					{
						SUBMCR(i, k, elt(Nptr, k, j), j, Nptr);
						TempCR = elt(Nptr, k, j);
						elt(Nptr, k, j) = ZEROCR();
						FREEMPCR(TempCR);
					}
				}
			}
			if ((i == m - 1) || (j == n - 1)) 
			{
				i++;
				break;
			}
			i++;
			j++;
		}
	}
	*rptr = r = i;
	for (j = r; j <= n - 1; j++)
	{
		for (i = 0; i <= n - 1; i++)
		{
			if (A[i])
			{
				*(*Cptr + j) = i;
				A[i] = 0;
				break;
			}
		}
	}
	ffree((char *)A, n * sizeof(unsigned int));
	return (Nptr);
 }

void EQUATIONSCR(MPMATCR *Mptr)
/*
 * solution of a system of linear equations with MPCR coefficients,
 * where *Mptr is the augmented matrix.
 */
{
	MPMATCR *N;
	MPCR *H;
	unsigned int f, i, j, n, r, s, *C;

	n = Mptr->C - 1;
	if (n == 0)
	{
		printf("augmented matrix has only one column!\n");
		return;
	}
	N = GAUSS_JORDANCR(Mptr, &r, &C);
	s = n - r;
	if (r)
	{
		if (C[r - 1] == n)
		{
			printf("the system has no solution\n");
			ffree((char *)C, (n + 1) * sizeof(unsigned int));
			FREEMATCR(N);
			return;
		}
		for (i = 0; i <= r - 1; i++)
		{
			printf("x[%u] = ", 1 + C[i]);
			if (s == 0)
			{
				PRINTCR(elt(N, i, C[n]));
				printf("\n");
				continue;
			}
			if (((elt(N, i, C[n])->R)->N)->S == 0 && ((elt(N, i, C[n])->I)->N)->S == 0)
			{
				f = 0;
				PRINTCR(elt(N, i, C[n]));
			}
			else 
				f = 1;
			for (j = r; j <= n - 1; j++)
			{

				H = MINUSCR(elt(N, i, C[j]));
				if (((H->R)->N)->S == 0 && ((H->I)->N)->S == 0)
					continue;
				else if (((H->R)->N)->S && ((H->I)->N)->S)
				{
					if (f == 0)
						printf(" + ");
					printf("(");
					PRINTCR(H);
					printf(")");
					printf("x[%u]", 1 + C[j]);
					f = 0;
				}
				else if (((H->I)->N)->S == 0)
				{
					if (((H->R)->N)->S == 1)
					{
						if (f == 0)
							printf(" + ");
						f = 0;
					}
					else
					{
						if (f == 1)
							printf(" -");
						else
							printf(" - ");
						if (!EQMINUSONER(H->R))
							((H->R)->N)->S = -(((H->R)->N)->S);
						f = 0;
					}
					if (!EQONEI((H->R)->N) && !EQMINUSONEI((H->R)->N))
						PRINTI((H->R)->N);
					printf("x[%u]", 1 + C[j]);
					if (!EQONEI((H->R)->D))
					{
						printf("/");
						PRINTI((H->R)->D);
					}
				}
				else 
				{
					if (((H->I)->N)->S == 1)
					{
						if (f == 0)
							printf(" + ");
						f = 0;
					}
					else
					{
						if (f == 1)
							printf(" -");
						else
							printf(" - ");
						if (!EQMINUSONER(H->I))
							((H->I)->N)->S = -(((H->I)->N)->S);
						f = 0;
					}
					if (!EQONEI((H->I)->N) && !EQMINUSONEI((H->I)->N))
						PRINTI((H->I)->N);
					printf("ix[%u]", 1 + C[j]);
					if (!EQONEI((H->I)->D))
					{
						printf("/");
						PRINTI((H->I)->D);
					}
				}
			}
			if (f)
				printf("0");
			printf("\n");
		}
	}
	if (s)
	{
		printf("Here ");
		for (j = r; j <= n - 1; j++)
		{
			if (j != n - 1)
				printf("x[%u], ", 1 + C[j]);
			else
				printf("x[%u] ", 1 + C[j]);
		}
		if (s == 1)
			printf("is arbitrary\n");
		else
			printf("are arbitrary\n");
	}
	ffree((char *)C, (n + 1) * sizeof(unsigned int));
	FREEMATCR(N);
	return;
}

MPMATCR *NULL_SPACECR(MPMATCR *Mptr)
/*
 * returns a matrix whose columns form a basis for N(*Mptr),
 * otherwise returns NULL.
 */
{
	MPMATCR *N, *Lptr;
	unsigned int  i, j, n, *C, r;

	n = Mptr->C;
	N = GAUSS_JORDANCR(Mptr, &r, &C);
	if (r == n)
	{
		ffree((char *)C, n * sizeof(unsigned int));
		FREEMATCR(N);
		return (NULL);
	}
	else
	{
		Lptr = BUILDMATCR(n, n - r);
		if (r)
		{
			for (i = 0; i <= r - 1; i++)
			{
				for (j = 0; j <= n - r - 1; j++)
					elt(Lptr, C[i], j) = MINUSCR(elt(N, i, C[r + j]));
			}
		}
		FREEMATCR(N);
		for (i = r; i <= n - 1; i++)
		{
			for (j = 0; j <= n - r - 1; j++)
			{
				if (j == i - r)
					elt(Lptr, C[i], j) = ONECR();
				else
					elt(Lptr, C[i], j) = ZEROCR();
			}
		}
		ffree((char *)C, n * sizeof(unsigned int));
		return (Lptr);
	}
}

MPMATCR *COLUMN_SPACECR(MPMATCR *Mptr)
/*
 * returns a matrix whose columns form a basis for C(*Mptr).
 * otherwise returns NULL.
 */
{
	MPMATCR *N, *Lptr;
	unsigned int  i, j, m, n, *C, r;

	m = Mptr->R;
	n = Mptr->C;
	N = GAUSS_JORDANCR(Mptr, &r, &C);
	FREEMATCR(N);
	if (r == 0)
	{
		ffree((char *)C, n * sizeof(unsigned int));
		return (NULL);
	}
	else
	{
		Lptr = BUILDMATCR(m, r);
		for (i = 0; i <= m - 1; i++)
		{
			for (j = 0; j <= r - 1; j++)
				elt(Lptr, i, j) = COPYCR(elt(Mptr, i, C[j]));
		}
		ffree((char *)C, n * sizeof(unsigned int));
		return (Lptr);
	}
}

MPMATCR *EQUATIONS1CR(MPMATCR *Mptr)
/*
 * solution of a system of linear equations with MPCR coefficients, where 
 * *Mptr is the augmented matrix. To be used when there is no solution, in
 * which case NULL is returned, or when there is a unique solution, in which
 * case the column vector (x[0],...x[n-1]) is returned.
 */
{
	MPMATCR *N, *Nptr;
	unsigned int i, n, r, *C;

	n = Mptr->C - 1;
	N = GAUSS_JORDANCR(Mptr, &r, &C);
	Nptr = NULL;
	if (C[r - 1] != n)
	{
		Nptr = BUILDMATCR(n, 1);
		for (i = 0; i <= n - 1; i++)
			elt(Nptr, i, 0) = COPYCR(elt(N, i, C[n]));
	}
	ffree((char *)C, (n + 1) * sizeof(unsigned int));
	FREEMATCR(N);
	return Nptr;
}
