/* i6m.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

MPMATm *BUILDMATm(USI m, USI n)
/*
 * Allocates space for an m x n matrix of unsigned ints's.
 */
{
	unsigned int i;
	MPMATm *N;

	N=(MPMATm *)mmalloc(sizeof(MPMATm));
	N->R = m;
	N->C = n;
	N->V = (unsigned int **)mmalloc(m * sizeof(unsigned int *));
	for (i = 0; i <= m - 1; i++)
		*(N->V + i) = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	return (N);
}

MPMATm *COPYMATm(MPMATm *Mptr)
/*
 * a replacement for the assignment *Nptr = *Mptr. 
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	Nptr = BUILDMATm(Mptr->R, Mptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Mptr->C - 1; j++)
			Nptr->V[i][j] = Mptr->V[i][j];
	}
	return (Nptr);
}

void FREEMATm(MPMATm *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V 
 * of unsigned ints.
 */
{
	unsigned int i;

	if (Mptr == NULL)
		return;
	for (i = 0; i <= Mptr->R - 1; i++)
		ffree((char *)(*(Mptr->V + i)), Mptr->C * sizeof(unsigned int));
	ffree((char *)Mptr->V, Mptr->R * sizeof(unsigned int *));
	ffree((char *)Mptr, sizeof(MPMATm));
	return;
}

MPMATm *UNWRAPMATm(MPMATm *Mptr)
/*
 * unwraps the matrix *Mptr of unsigned ints into a long column vector.
 */
{
	unsigned int p, k;
	MPMATm *Nptr;

	p = Mptr->R * Mptr->C;
	Nptr = BUILDMATm(p, 1);
	for (k = 0; k <= p - 1; k++)
			Nptr->V[k][0] = Mptr->V[k / Mptr->C][k % Mptr->C];
	return (Nptr);
}

unsigned int PIVOTmm(USI i, USI j, MPMATm *Mptr, USI m)
/*
 * here 0 <= i <= j <= Mptr->R - 1.
 * returns the row index k, with least k (i <= k <= Mptr->R - 1), such that 
 * Mptr->V[k][j] is relatively prime to m; returns Mptr->R otherwise.
 */
{
	unsigned int k;
	
	k = i;
	while ((k <= Mptr->R - 1) && (GCDm(Mptr->V[k][j], m) != 1))
		k++;
	return (k);
}

unsigned int PIVOTm(USI i, USI j, MPMATm *Mptr)
/*
 * here 0 <= i <= j <= Mptr->R - 1.
 * returns the row index k, with least k (i <= k <= Mptr->R - 1), such that 
 * Mptr->V[k][j] is non-=zero; returns Mptr->R otherwise.
 */
{
	unsigned int k;
	
	k = i;
	while ((k <= Mptr->R - 1) && (Mptr->V[k][j] == 0))
		k++;
	return (k);
}

void TRANSm(USI p, USI q, MPMATm *Mptr)
/*
 * 0 <= p < q <= Mprt->R - 1.
 * Mptr->V[i][j] = 0 for Mptr->R - 1 >= i >= p and 0 <= j < q 
 * transposes the p-th and q-th rows of *Mptr.
 */
{
	unsigned int j, x;

	for (j = p; j <= Mptr->C - 1; j++) 
	{
		x = Mptr->V[p][j];
		Mptr->V[p][j] = Mptr->V[q][j];
		Mptr->V[q][j] = x;
	}
	return;
}

void NORMm(USI p, USI j, MPMATm *Mptr, USI z)
/*
 * divides the p-th row by Mptr->V[p][j].
 * 0 <= p <= Mprt->R - 1.
 * 0 <= j <= Mprt->C - 1.
 * Mptr->V[p][k] = 0 if k < j.
 */
{
	unsigned int k, x;

	x = Mptr->V[p][j];
	Mptr->V[p][j] = 1;
	for (k = j + 1; k <= Mptr->C - 1; k++) 
		Mptr->V[p][k] = DIVm(Mptr->V[p][k], x, z);
	return;
}

void SUBMm(USI p, USI q, USI a, USI j, MPMATm *Mptr, USI z)
/*
 * subtracts *Aptr times the p-th row of *Mptr from the q-th,
 * for columns Mptr->C - 1 >= k > j.
 * 0 <= p, q <= Mprt->R - 1.
 * Mptr->V[p][k] = 0 if k < j, Mptr->V[p][j] = 1.
 */
{
	unsigned int k, y;

	for (k = j + 1; k <= Mptr->C - 1; k++) 
	{
		y = MULTm(a, Mptr->V[p][k], z);
		Mptr->V[q][k] = SUBm(Mptr->V[q][k], y, z);
	}
	return;
}

MPMATm *GAUSS_JORDANm(MPMATm *Mptr, USI *rptr, USI *Cptr[], USI z)
/*
 * returns the reduced row-echelon form of *Mptr. 
 * *rptr = rank *Mptr and *Cptr[0],...,*Cptr[rank-1] are the columns in which 
 * the leading entries 1 occur.
 */
{
	MPMATm *Nptr;
	unsigned int i, j, k, m, n, p, r, *A;

	m = Mptr->R;
	n = Mptr->C;
	A = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	for (i = 0; i <= n - 1; i++)
		A[i] = 1;
	Nptr = COPYMATm(Mptr);
	i = 0;
	j = 0;

	*Cptr = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	while (1) 
	{
		while ((j <= n - 1) && (PIVOTm(i, j, Nptr) == m))
			j++;
		if (j == n)
			break;
		else 
		{
			p = PIVOTm(i, j, Nptr);
			if (p > i)
				TRANSm(i, p, Nptr);
			if (Nptr->V[i][j] != 1)
				NORMm(i, j, Nptr, z);
			*(*Cptr + i) = j;
			A[j] = 0;
			for (k = 0; k <= m - 1; k++) 
			{
				if (k != i)
				{
					if (Nptr->V[k][j])
					{
						SUBMm(i, k, Nptr->V[k][j], j, Nptr, z);
						Nptr->V[k][j] = 0;
					}
				}
			}
			if ((i == m - 1) || (j == n - 1))
			{
				i++;
				break;
			}
			i++;
			j++;
		}
	}
	*rptr = r = i;
	for (j = r; j <= n - 1; j++)
	{
		for (i = 0; i <= n - 1; i++)
		{
			if (A[i])
			{
				*(*Cptr + j) = i;
				A[i] = 0;
				break;
			}
		}
	}
	ffree((char *)A, n * sizeof(unsigned int));
	/*for (k = 0; k <= n - 1; k++)
		printf("C[%u] = %u\n", k, C[k]);*/
	return (Nptr);
 }

void EQUATIONSm(MPMATm *Mptr, USI z)
/*
 * solution of a system of linear equations with mod z coefficients,
 * where *Mptr is the augmented matrix.
 */
{
	MPMATm *N;
	unsigned int f, h, i, j, n, r, s, *C;

	n = Mptr->C - 1;
	if (n == 0)
	{
		printf("augmented matrix has only one column!\n");
		return;
	}
	N = GAUSS_JORDANm(Mptr, &r, &C, z);
	s = n - r;
	if (r)
	{
		if (C[r - 1] == n)
		{
			printf("the system has no solution\n");
			ffree((char *)C, (n + 1) * sizeof(unsigned int));
			FREEMATm(N);
			return;
		}
		for (i = 0; i <= r - 1; i++)
		{
			printf("x[%u] = ", 1 + C[i]);
			if (s == 0)
			{
				printf("%u\n", N->V[i][C[n]]);
				continue;
			}
			if (N->V[i][C[n]])
			{
				f = 0;
				printf("%u", N->V[i][C[n]]);
			}
			else if (N->V[i][C[n]] == 0)
				f = 1;
			for (j = r; j <= n - 1; j++)
			{
				h = MINUSm(N->V[i][C[j]], z);
				if (h == 0)
					continue;
				else 
				{
					if (f == 0)
						printf(" + ");
					f = 0;
				}
				if (h != 1)
					printf("%u", h);
				printf("x[%u]", 1 + C[j]);
			}
			if (f)
				printf("0");
			printf("\n");
		}
	}
	if (s)
	{
		printf("Here ");
		for (j = r; j <= n - 1; j++)
		{
			if (j != n - 1)
				printf("x[%u], ", 1 + C[j]);
			else
				printf("x[%u] ", 1 + C[j]);
		}
		if (s == 1)
			printf("is arbitrary.\n");
		else
			printf("are arbitrary.\n");
	}
	ffree((char *)C, (n + 1) * sizeof(unsigned int));
	FREEMATm(N);
	return;
}

MPMATm *EQUATIONS1m(MPMATm *Mptr, USI m)
/*
 * solution of a system of linear equations with mod m coefficients, where 
 * *Mptr is the augmented matrix. To be used when there is no solution, in
 * which case NULL is returned, or when there is a unique solution, in which
 * case the column vector (x[0],...x[n-1]) is returned.
 */
{
	MPMATm *N, *Nptr;
	unsigned int i, n, r, *C;

	n = Mptr->C - 1;
	N = GAUSS_JORDANm(Mptr, &r, &C, m);
	Nptr = NULL;
	if (C[r - 1] != n)
	{
		Nptr = BUILDMATm(n, 1);
		for (i = 0; i <= n - 1; i++)
			Nptr->V[i][0] = N->V[i][C[n]];
	}
	ffree((char *)C, (n + 1) * sizeof(unsigned int));
	FREEMATm(N);
	return Nptr;
}

MPMATm *NULL_SPACEm(MPMATm *Mptr, USI z)
/*
 * returns a matrix mod z whose columns form a basis for N(*Mptr),
 * otherwise returns NULL.
 */
{
	MPMATm *N, *Lptr;
	unsigned int i, j, n, *C, r;

	n = Mptr->C;
	N = GAUSS_JORDANm(Mptr, &r, &C, z);
	if (r == n)
	{
		ffree((char *)C, n * sizeof(unsigned int));
		FREEMATm(N);
		return (NULL);
	}
	else
	{
		Lptr = BUILDMATm(n, n - r);
		if (r)
		{
			for (i = 0; i <= r - 1; i++)
			{
				for (j = 0; j <= n - r - 1; j++)
					Lptr->V[C[i]][j] = MINUSm(N->V[i][C[r + j]], z);
			}
		}
		FREEMATm(N);
		for (i = r; i <= n - 1; i++)
		{
			for (j = 0; j <= n - r - 1; j++)
			{
				if (j == i - r)
					Lptr->V[C[i]][j] = 1;
				else
					Lptr->V[C[i]][j] = 0;
			}
		}
		ffree((char *)C, n * sizeof(unsigned int));
		return (Lptr);
	}
}

MPMATm *COLUMN_SPACEm(MPMATm *Mptr, USI z)
/*
 * returns a matrix mod z whose columns form a basis for C(*Mptr).
 * otherwise returns NULL.
 */
{
	MPMATm *N, *Lptr;
	unsigned int  i, j, m, n, *C, r;

	m = Mptr->R;
	n = Mptr->C;
	N = GAUSS_JORDANm(Mptr, &r, &C, z);
	FREEMATm(N);
	if (r == 0)
	{
		ffree((char *)C, n * sizeof(unsigned int));
		return (NULL);
	}
	else
	{
		Lptr = BUILDMATm(m, r);
		for (i = 0; i <= m - 1; i++)
		{
			for (j = 0; j <= r - 1; j++)
				Lptr->V[i][j] = Mptr->V[i][C[j]];
		}
		ffree((char *)C, n * sizeof(unsigned int));
		return (Lptr);
	}
}
unsigned int TRACEm(MPMATm *Mptr, USI z)
/*
 * returns the trace of the mod z matrix *Mptr. 
 */
{
	unsigned int a, i;

	a = 0;
	for (i = 0; i <= Mptr->R - 1; i++)
		a = ADDm(a, Mptr->V[i][i], z);
	return a;
}

MPMATm *SCALARMATm(MPMATm *Mptr, USI a, USI z)
/*
 * returns a * (*Mptr).
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	Nptr = BUILDMATm(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			Nptr->V[i][j] = MULTm(Mptr->V[i][j], a, z);
	}
	return (Nptr);
}

MPMATm *MINUSMATm(MPMATm *Mptr, USI z)
/*
 * returns -(*Mptr) mod z. 
 */
{
	unsigned int i, j;
	MPMATm *Nptr;

	Nptr = BUILDMATm(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			Nptr->V[i][j] = MINUSm(Mptr->V[i][j], z);
	}
	return (Nptr);
}

MPMATm *DIAGm(MPMATm *Mptr, USI a, USI z)
/*
 * *Nptr = *Mptr - a * (Unit Matrix). 
 */
{
	unsigned int i, x;
	MPMATm *Nptr;

	Nptr = COPYMATm(Mptr);
	x = MINUSm(a, z);
	for (i = 0; i <= Mptr->R - 1; i++)
			Nptr->V[i][i] = ADDm(Nptr->V[i][i], x, z);
	return Nptr;
}

MPMATm *ADDMATm(MPMATm *Mptr, MPMATm *Nptr, USI z)
/*
 * returns *Mptr + *Nptr mod z.
 */
{
	unsigned int i, j;
	MPMATm *Lptr;

	Lptr = BUILDMATm(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
	{
		for (j = 0; j <= Lptr->C - 1; j++)
			Lptr->V[i][j] = ADDm(Mptr->V[i][j], Nptr->V[i][j], z);
	}
	return (Lptr);
}

MPMATm *SUBMATm(MPMATm *Mptr, MPMATm *Nptr, USI z)
/*
 * returns *Mptr - *Nptr mod z.
 */
{
	unsigned int i, j;
	MPMATm *Lptr;

	Lptr = BUILDMATm(Mptr->R, Mptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
	{
		for (j = 0; j <= Lptr->C - 1; j++)
			Lptr->V[i][j] =	SUBm(Mptr->V[i][j], Nptr->V[i][j], z);
	}
	return (Lptr);
}

MPMATm *MULTMATm(MPMATm *Mptr, MPMATm *Nptr, USI z)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr) (mod z).
 */
{
	MPMATm *Qptr;
	unsigned int i, j, k, x, y;

	Qptr = BUILDMATm(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			x = 0;
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				y = MULTm(Mptr->V[i][k], Nptr->V[k][j], z);
				x = ADDm(x, y, z);
			}
			Qptr->V[i][j] = x;
		}
	}
	return (Qptr);
}

MPMATm *POWERMATm(MPMATm *Aptr, USI n, USI z)
/*
 *  returns (*Aptr) ^ n, mod z, where 0 <= n < R0 * R0.
 */
{
	MPMATm *Temp, *B, *Eptr;

	Eptr = IDENTITYm(Aptr->R);
	if (n == 0)
		return (Eptr);
	B = COPYMATm(Aptr);
	while (1)
        {
		if (n & 1) 
		{
			Temp = Eptr;
			Eptr = MULTMATm(Temp, B, z);
			FREEMATm(Temp);
			if (n == 1)
			{
				FREEMATm(B);
				return (Eptr);
			}
		}
		Temp = B;
		B = MULTMATm(Temp, Temp, z);
		FREEMATm(Temp);
		n = n >> 1;
	}
}

unsigned int DETm(MPMATm *Mptr, USI z)
/*
 *returns the determinant of a matrix (mod z). 
 */
{
	MPMATm *A;
	unsigned int DETA, i, k, p, n;

	A = COPYMATm(Mptr);
	n = Mptr->R;
	DETA = 1;
	for (i = 0; i <= n - 1; i++)
	{
		p = PIVOTm(i, i, A);
		if (p == n)
		{
			FREEMATm(A);
			return 0;
		}
		else	
		{
			if (p > i)
			{
				TRANSm(i, p, A);
				DETA = MINUSm(DETA, z);
			}
			DETA = MULTm(DETA, A->V[i][i], z);
			if (A->V[i][i] != 1) 
				NORMm(i, i, A, z);
			for (k = i + 1; k <= n - 1; k++)
			{
					if (A->V[k][i])
					{
						SUBMm(i, k, A->V[k][i], i, A, z);
						A->V[k][i] = 0;
					}
			}
		}
	}
	FREEMATm(A);
	return DETA;
}

MPMATm *JCOLSMATm(MPMATm *Mptr, MPMATm *Nptr)
/*
 * returns [ *Mptr | *Nptr]. 
 */
{
	unsigned int i, j;
	MPMATm *Lptr;

	Lptr = BUILDMATm(Mptr->R, Mptr->C + Nptr->C);
	for (i = 0; i <= Lptr->R - 1; i++)
	{
		for (j = 0; j <= Lptr->C - 1; j++)
		{
			if (j <= Mptr->C - 1)
				Lptr->V[i][j] = Mptr->V[i][j];
			else
				Lptr->V[i][j] = Nptr->V[i][j - Mptr->C];
		}
	}
	return (Lptr);
}

MPMATm *JROWSMATm(MPMATm *Mptr, MPMATm *Nptr)
/*
 *         [ *Mptr]
 * returns [------]
 *         [ *Nptr] 
 */
{
	unsigned int i, j;
	MPMATm *Lptr;

		Lptr = BUILDMATm(Mptr->R + Nptr->R, Mptr->C);
		for (i = 0; i <= Lptr->R - 1; i++)
		{
			if (i <= Mptr->R - 1)
			{
				for (j = 0; j <= Lptr->C - 1; j++)
					Lptr->V[i][j] = Mptr->V[i][j];
			}
			else
			{
				for (j = 0; j <=  Lptr->C - 1; j++)
					Lptr->V[i][j] = Nptr->V[i - Mptr->R][j];
			}
		}
		return (Lptr);
}

MPMATm *JCOLSIDENTITYm(MPMATm *Mptr)
/*
 * forms the partioned matrix [*Mptr | I]. 
 */
{
	MPMATm *Temp, *Nptr;

	Temp = IDENTITYm(Mptr->R);
	Nptr = JCOLSMATm(Mptr, Temp);
	FREEMATm(Temp);
	return Nptr;
}

MPMATm *INVERSEMATm(MPMATm *Mptr, USI z)
/*
 * returns the inverse of a matrix mod z if the inverse exists,
 * returns NULL otherwise.
 */
{
	MPMATm *A, *Temp;
	unsigned int j, k, p, i, n, t;


	A = JCOLSIDENTITYm(Mptr);
	n = Mptr->R;
	t = n << 1;
	i = 0;
	j = 0;
	while (1)
	{
		while ((j <= t - 1) && (PIVOTmm(i, j, A, z) == n))
			j++;
		if (i < j)
		{
			FREEMATm(A);
			return NULL;
		}
		if (j == t)
			break;
		else {
			p = PIVOTmm(i, j, A, z);
			if (p > i)
				TRANSm(i, p, A);
			if (A->V[i][j] != 1)
				NORMm(i, j, A, z);
			for (k = 0; k <= n - 1; k++)
			{
				if (k != i)
				{
					if (A->V[k][j])
					{
						SUBMm(i, k, A->V[k][j], j, A, z);
						A->V[k][j] = 0;
					}
				}
			}
			if ((i == n - 1) || (j == t - 1)) 
				break;
			i++;
			j++;
		}
	}

	Temp = BUILDMATm(n, n);
	for (i = 0; i <= n - 1; i++) 
	{
		for (j = 0; j <= n - 1; j++) 
			Temp->V[i][j] = A->V[i][j + n];	
	}
	FREEMATm(A);
	return Temp;
}

MPMATm *ADJOINTm(MPMATm *Mptr, USI z)
/*
 * returns the adjoint of a matrix mod z.
 */
{
	unsigned int i, j, k, l, r, rank, n, *C, det, cofactor, multiple;
	unsigned int minor;
	MPMATm *Nptr, *Temp, *Temp1, *Temp2;

	n = Mptr->C;
	Nptr = GAUSS_JORDANm(Mptr, &rank, &C, z);
	FREEMATm(Nptr);
	ffree((char *)C, n * sizeof(unsigned int));
	if (rank <= n - 2)
	{
		Nptr = ZEROMNm(n, n);
		return Nptr;
	}
	else if (rank == n)
	{
		Temp = INVERSEMATm(Mptr, z);
		det = DETm(Mptr, z);
		Nptr = SCALARMATm(Temp, det, z);
		FREEMATm(Temp);
	}
	if (rank == n - 1)
	{
		Nptr = ZEROMNm(n, n);
		Temp = NULL_SPACEm(Mptr, z);
		i = 0;
		while (1)
		{
			k = Temp->V[i][0];
			if (k != 0)
				break;
			i++;
		}
		Temp1 = DELETE_COLm(i + 1, Mptr);
		l = (i % 2) ? z - 1 : 1;
		for (j = 0; j <= n - 1; j++)
		{
			Temp2 = DELETE_ROWm(j + 1, Temp1);
			minor = DETm(Temp2, z);
			FREEMATm(Temp2);
			if (minor)
			{
				cofactor = MULTm(minor, l, z);
				multiple = DIVm(cofactor, k, z);
				for (r = 0; r <= n - 1; r++)
					if (Temp->V[r][0])
						Nptr->V[r][j] = MULTm(multiple, Temp->V[r][0], z);
			}
			l = MINUSm(l, z);
		}
		FREEMATm(Temp);
		FREEMATm(Temp1);
	}
	return Nptr;
}
