/* i7R.c */
#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATR *ROWVECTORR(MPMATR *Mptr, USI i)
/*
 * returns the i-th row vector of *Mptr.
 */
{
	MPMATR *Temp;
	unsigned int j;
	
	Temp = BUILDMATR(1, Mptr->C);
	for (j = 0; j < Mptr->C; j++)
		elt(Temp, 0, j) = COPYR(elt(Mptr, i, j));
	return (Temp);
}

MPMATR *ColVectorR(MPMATR *Mptr, USI j)
/*
 * returns the j-th column vector of *Mptr.
 */
{
	MPMATR *Temp;
	unsigned i;
	Temp = BUILDMATR(Mptr->R, 1);
	for (i = 0; i < Mptr->R; i++)
		elt(Temp, i, 0) = COPYR(elt(Mptr, i, j));
	return Temp;
}

MPR *DOTR(MPMATR *Mptr, MPMATR *Nptr)
/*
 * *Aptr is the dot product of row/column vectors *Mptr and *Nptr.
 */
{
	unsigned int i, j;
	MPR *H, *TempR, *Aptr;
	
	Aptr = ZEROR();
	for (i = 0; i <= Mptr->R - 1; i++)
		for (j = 0; j <= Mptr->C - 1; j++)
		{
			H = MULTR(elt(Mptr, i, j), elt(Nptr, i, j));
			TempR = Aptr;
			Aptr = ADDR(Aptr, H);
			FREEMPR(TempR);
		}
	return (Aptr);
}

MPMATR *GRAM_SCHMIDTR(MPMATR *Mptr)
/*
 * *Mptr is a non-zero matrix. The left to right basis is found for the
 * column-space. The Gram-Schmidt process is then applied to get an orthogonal
 * basis for the column-space of *Mptr.
 */
{
	MPR *G, *H, *I, *TempR;
	MPMATR *M1, *R1, *R2, *Tmp;
	unsigned int i, j, k, n, r;

	n = Mptr->R;
	M1 = COLUMN_SPACER(Mptr);
	r = M1->C;
	for (j = 1; j <= r - 1; j++)
	{
		R1 = ColVectorR(M1, j);
		for (k = 0; k < j; k++)
		{
			R2 = ColVectorR(M1, k);
			G = DOTR(R2, R1);
			H = DOTR(R2, R2);
			TempR = G;
			G = MINUSR(G);
			FREEMPR(TempR);
			I = RATIOR(G, H);
			Tmp = R2;
			R2 = SCALARMATR(Tmp, I);
			FREEMATR(Tmp);
			Tmp = R1;
			R1 = ADDMATR(Tmp, R2);
			FREEMATR(R2);
			FREEMATR(Tmp);
			FREEMPR(G);
			FREEMPR(H);
			FREEMPR(I);
		}
		for (i = 0; i <= n - 1; i++)
			elt(M1, i, j) = COPYR(elt(R1, i, 0));
		FREEMATR(R1);
	}
	return (M1);
}

MPR *DOTLR(MPMATR *Mptr)
/*
 * calculates the square of the length of *Mptr.
 */
{
	return DOTR(Mptr, Mptr);
}
