/* menu.c */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#ifndef _WIN32
#include "cmat.h"
#endif
#include "integer.h"
#include "fun.h"

#define MEMINCR 256

#ifndef _WIN32
pid_t spawnl(int parallel, char *path,...);
#endif

extern unsigned int VERSION;

int keyvalid(char c, struct MENUITEM *p)
/* Returns 1 if character c is in list of MENUITEMS starting with p, 
0 otherwise. */
{
	int flag;
	flag = 0;
	while (p != NULL && !flag)
		if (p->c == c)
			flag = 1;
		else
			p = p->next;
	return flag;
}

void putnchar(char c, int n)
{
	int i;
	for (i = 0; i < n; i++)
		putchar(c);
}

struct MENU *menulist = NULL;

int menuinit(char *menufile)
/*
 * This function reads in all the menus from the file given.  It stores
 * them in an linked list sort of a set up.
 * Not a lot of (ie almost none) error checking is done on the file.
 * If, however, an error is detected, such as menu.dat not existing or being
 * corrupt in some way, then a non-zero value is returned.  Else, 0 is returned.
 */
{
	struct MENUITEM *last, *new;
	struct MENU *lastmenu = NULL, *newmenu;
	char fname[50];
	FILE *f;
	int longest,sl, c;

	if (menulist != NULL)
		return 0;	/* No need to re-initialize */
#ifdef _WIN32
	sprintf(fname, "%s", menufile);
	if ((f = fopen(fname, "r")) == NULL)
#else
	sprintf(fname, "%s/%s", DATPATH, menufile);
	if ((f = fopen(fname, "r")) == NULL)
#endif
	{
		fprintf(stderr, "could not open menu file %s\n", fname);
		GetReturn();
		return 1;
	}
	c = fgetc(f);
	while (c != EOF) {
		if (c != '#') {
			fclose(f);
			return 1;
		}
		newmenu = (struct MENU *)malloc(sizeof(struct MENU));
		newmenu->next = NULL;
		fscanf(f, "%d", &newmenu->number);
		c = fgetc(f);
		if (c == ',') {
			fscanf(f, "%79[^\n]\n", newmenu->title);
			longest=strlen(newmenu->title)+5;
			newmenu->titleflag = 1;
		} else {
			newmenu->titleflag = 0;
			longest=0;
		}
		newmenu->head = NULL;
		last = NULL;
		while ((c = fgetc(f)) != EOF && c != '#') {
			new = (struct MENUITEM *)malloc(sizeof(struct MENUITEM));
			new->next = NULL;
			new->c = c;
			fscanf(f, ",%79[^\n]\n", new->s);
			sl=strlen(new->s);
			if(sl > longest)
			  longest=sl;
			if (last == NULL)
				newmenu->head = new;
			else
				last->next = new;
			last = new;
		}
		newmenu->longest=longest;
		if (lastmenu == NULL)
			menulist = newmenu;
		else
			lastmenu->next = newmenu;
		lastmenu = newmenu;
	}
	fclose(f);
	return 0;
}

void menuend()
/*
 * This function deallocates all the memory previously allocated with
 * menuinit() and then sets the number of menus to 0.
 */
{
	struct MENU *mp, *mq;
	struct MENUITEM *p, *q;

	mp = menulist;
	while (mp != NULL)
	{
		p = mp->head;
		while (p) {
			q = p->next;
			free((char *)p);
			p = q;
		}
		mq = mp->next;
		free((char *)mp);
		mp = mq;
	}
	menulist = NULL;
}

char drawmenu(int n)
/*
 * Basically, all this function does is to draw the menu
 * with number n.  Very simple.
 * Returns character pressed (guaranteed valid) if menu was found, else
 * exit(1) (a serious error).
 */
{
	struct MENUITEM *p;
	struct MENU *menucursor;
	int longest, l;
	char found, format[15], c;

	/* Work out which element of menu[] we are using */
	menucursor = menulist;
	c = '0';
	found = 0;
	while (menucursor != NULL && !found)
		if (menucursor->number == n)
			found = 1;
		else
			menucursor = menucursor->next;
	if (!found) {
		fprintf(stderr, "Menu %d not found\n",n);
		exit(1);
	}
	longest=menucursor->longest;
	sprintf(format, "* %%c %%-%ds *\n", longest);
#ifndef MENUDEBUG
	clearscreen();
#endif /* MENUDEBUG */
	putnchar('*', longest + 6);
	putchar('\n');
	if (menucursor->titleflag) {
		printf("* ");
		l = strlen(menucursor->title);
		putnchar(' ', (longest - l) / 2);
		printf("%s", menucursor->title);
		if(n<10)
		  putnchar(' ', longest - 1 - l + (l - longest) / 2);
		else
		  putnchar(' ', longest - 2 - l + (l - longest) / 2);
		if(VERSION==1)
		  printf(" R%d *\n",n);
		else if (VERSION==2)
		  printf("CR%d *\n",n);
		else if (VERSION==3)
		  printf(" m%d *\n",n);
		else
		  printf("  %d *\n",n);
		putnchar('*', longest + 6);
		putchar('\n');
	}
	p = menucursor->head;
	while (p) {
		printf(format, p->c, p->s);
		p = p->next;
	}
	putnchar('*', longest + 6);
	printf("\n");
#ifndef MENUDEBUG
	SelOpt();
	scanf(" %c", &c);
	while (!keyvalid(c, menucursor->head)) {
		printf("re-enter option (and integers, if relevant)\n");
		SelOpt();
		Flush();
		scanf(" %c", &c);
	}
#endif /* MENUDEBUG */
	return c;	/* all went okay */
}

char *Gets()
/*
 * As for Fgets() (see below) but from standard input.
 */
{
	return Fgets(stdin);
}

char *Fgets(FILE *f)
/*
 * Reads one line from file into a string allocated by malloc.  Return NULL on
 * error, otherwise a pointer to the string.
 * Byte, June 1988, page 316.
 */
{
	char *r;
	unsigned int n, m;
	int c;

	n = 0;
	m = MEMINCR;
	r = (char *)malloc((m + 1) * sizeof(unsigned int));
	while ((c = fgetc(f)) != '\n') {
		if (c == EOF)
			return NULL;
		if (--m == 0) {
			if ((r = (char*)realloc(r, n + MEMINCR + 1)) == NULL)
				return NULL;
			m = MEMINCR;
		}
		r[n++] = c;
	}
	r[n] = '\0';
	if ((r = (char*)realloc(r, n + 1)) == NULL)
		return NULL;
	return r;
}

void SelOpt()
/*
 * This function simply prints the "SELECT OPTION: " prompt.  It is here only
 * for consistency, ie whenever the words SELECT OPTION are printed it can be
 * guaranteed that they are coming from this function, and the way they are
 * printed will be exactly the same every time.
 */
{
	printf("\nSELECT OPTION: ");
}

unsigned int GetYN()
/*
 * Gets a character in from the keyboard, making sure it's a y or an n (either
 * case).  If at first the user doesn't succeed, he tries, tries again.
 * 0 is returned if n, 1 if y.
 */
{
	int c;

	printf("Enter y or n : ");
	c = GetCharFlush();
	while (c != 'y' && c != 'Y' && c != 'n' && c != 'N') {
		printf("Try again. Enter y or n: ");
		c = GetCharFlush();
	}
	return c == 'y' || c == 'Y';
}

int whitespace(FILE *f)
/* removes spaces or newlines from a stream. */
{
	int ch;

	while ((ch = fgetc(f)) != EOF && ( ch == ' ' || ch == '\n'))
		;
	return ch;
}

void FFlush(FILE *f)
{
	int ch;
	while ((ch = fgetc(f)) != EOF && ch != '\n')
		;
}

void Flush()
{
	int ch;
	while ((ch = getchar()) != EOF && ch != '\n')
		;
}

int GetCharFlush()
{
	char ch;
	scanf(" %c", &ch);
	Flush();
	return ch;
}

void GetReturn()
{
	printf("HIT RETURN: ");
	Flush();
}

void TryAgain()
{
	printf("Hit return and Try again: ");
	Flush();
}


void clearscreen()
{
#ifndef _WIN32
	system("/usr/bin/clear");
#else
	system("cls");
#endif
}

int scanf01(USI *jptr)
/*
 * returns 1 if unsigned int *jptr < M0 is successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u", jptr);
	Flush();
	if (n == 1)
	{
		if (*jptr < M0)
			return 1;
		else
			printf("integer %u exceeded array bound %u, try again:\n", *jptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf011(USI *jptr, USI *kptr)
/*
 * returns 1 if unsigned int *jptr < M0 and unsigned int *kptr are successfully
 * entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u", jptr, kptr);
	Flush();
	if (n == 2)
	{
		if (*jptr < M0)
			return 1;
		else
			printf("first integer %u exceeded array bound %u, try again:\n", *jptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf02(USI *jptr, USI *kptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr < M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u", jptr, kptr);
	Flush();
	if (n == 2)
	{
		if (*jptr < M0 && *kptr < M0)
			return 1;
		else
			printf("first integer %u or second integer %u exceeded array bound %u, try again:\n", *jptr, *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf021(USI *jptr, USI *kptr, USI *mptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, each less than M0 
 * and unsigned int *mptr are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", jptr, kptr, mptr);
	Flush();
	if (n == 3)
	{
		if (*jptr < M0 && *kptr < M0)
			return 1;
		else
			printf("first integer %u or second integer %u exceeded array bound %u, try again:\n", *jptr, *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf21(USI *jptr, USI *kptr, USI *mptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, with *mptr < M0 are successfully 
 * entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", jptr, kptr, mptr);
	Flush();
	if (n == 3)
	{
		if (*mptr < M0)
			return 1;
		else
			printf("third integer %u exceeded array bound %u, try again:\n", *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf121(USI *hptr, USI *jptr, USI *kptr, USI *mptr)
/*
 * returns 1 if arbitrary unsigned ints *hptr, *mptr and 
 * unsigned ints *jptr, *kptr, each less than M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", hptr, jptr, kptr, mptr);
	Flush();
	if (n == 4)
	{
		if (*jptr < M0 && *kptr < M0)
			return 1;
		else
			printf("second integer %u or third integer %u exceeded array bound %u, try again:\n", *jptr, *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf221(USI *hptr, USI *iptr, USI *jptr, USI *kptr, USI *mptr)
/*
 * returns 1 if unsigned ints *hptr, *iptr, *mptr and
 * unsigned ints *jptr, *kptr, each less than M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u", hptr, iptr, jptr, kptr, mptr);
	Flush();
	if (n == 5)
	{
		if (*jptr < M0 && *kptr < M0)
			return 1;
		else
			printf("third integer %u or fourth integer %u exceeded array bound %u, try again:\n", *jptr, *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf321(USI *gptr, USI *hptr, USI *iptr, USI *jptr, USI *kptr, USI *mptr)
/*
 * returns 1 if unsigned ints *gptr, *hptr, *iptr, *mptr and
 * unsigned ints *jptr, *kptr, each less than M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u%u", gptr, hptr, iptr, jptr, kptr, mptr);
	Flush();
	if (n == 6)
	{
		if (*jptr < M0 && *kptr < M0)
			return 1;
		else
			printf("fourth integer %u or fifth integer %u exceeded array bound %u, try again:\n", *jptr, *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf03(USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, each less than M0 are
 * successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", jptr, kptr, lptr);
	Flush();
	if (n == 3)
	{
		if (*jptr < M0 && *kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf031(USI *jptr, USI *kptr, USI *lptr, USI *mptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, each less than M0 and
 * unsigned int *mptr are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", jptr, kptr, lptr, mptr);
	Flush();
	if (n == 4)
	{
		if (*jptr < M0 && *kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf131(USI *iptr, USI *jptr, USI *kptr, USI *lptr, USI *mptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, each less than M0 and
 * unsigned int *mptr, *iptr are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u", iptr, jptr, kptr, lptr, mptr);
	Flush();
	if (n == 5)
	{
		if (*jptr < M0 && *kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf05(USI *iptr, USI *jptr, USI *kptr, USI *lptr, USI *mptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr, *mptr with
 * *iptr, *jptr, *kptr, *lptr and *mptr < M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u", iptr, jptr, kptr, lptr, mptr);
	Flush();
	if (n == 5)
	{
		if (*iptr < M0 && *jptr < M0 && *kptr < M0 && *lptr < M0 && *mptr < M0)
			return 1;
		else
			printf("one of %u, %u, %u, %u and %u exceeded array bound %u, try again:\n", *iptr, *jptr, *kptr, *lptr, *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf051(USI *iptr, USI *jptr, USI *kptr, USI *lptr, USI *mptr, USI *nptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr, *mptr, *nptr with,
 * *iptr, *jptr, *kptr, *lptr and *mptr < M0 are successfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u%u", iptr, jptr, kptr, lptr, mptr, nptr);
	Flush();
	if (n == 6)
	{
		if (*iptr < M0 && *jptr < M0 && *kptr < M0 && *lptr < M0 && *mptr < M0)
			return 1;
		else
			printf("one of %u, %u, %u, %u and %u exceeded array bound %u, try again:\n", *iptr, *jptr, *kptr, *lptr, *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf10(USI *jptr)
/*
 * returns 1 if unsigned int *jptr successfully, 0 otherwise.
 */
{
	int n;

	n = scanf("%u", jptr);
	Flush();
	if (n == 1)
		return 1;
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanfd10(int *jptr)
/*
 * returns 1 if int *jptr successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%d", jptr);
	Flush();
	if (n == 1)
		return 1;
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf11(USI *jptr, USI *kptr)
/*
 * returns 1 if  unsigned ints *jptr, *kptr with *kptr < M0 are successfully
 * entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u", jptr, kptr);
	Flush();
	if (n == 2)
	{
		if (*kptr < M0)
			return 1;
		else
			printf("second integer %u exceeded array bound %u, try again:\n", *kptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf12(USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr with *kptr and *lptr < M0
 * are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", jptr, kptr, lptr);
	Flush();
	if (n == 3)
	{
		if (*kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("second integer %u or third integer %u exceeded array bound %u, try again:\n", *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf012(USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr with *jptr < M0 are 
 * successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", jptr, kptr, lptr);
	Flush();
	if (n == 3)
	{
		if (*jptr < M0)
			return 1;
		else
			printf("first integer %u exceeded array bound %u, try again:\n", *jptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf0112(USI *jptr, USI *kptr, USI *lptr, USI *mptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, *mptr with
 * jptr < M0, *lptr < M0, *mptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", jptr, kptr, lptr, mptr);
	Flush();
	if (n == 4)
	{
		if (*jptr < M0 && *lptr < M0 && *mptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *lptr, *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf0122(USI *jptr, USI *kptr, USI *lptr, USI *mptr, USI *nptr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, *mptr, *nptr with
 * *jptr < M0, *mptr < M0, *nptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u", jptr, kptr, lptr, mptr, nptr);
	Flush();
	if (n == 5)
	{
		if (*jptr < M0 && *mptr < M0 && *nptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *mptr, *nptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf01221(USI *jptr, USI *kptr, USI *lptr, USI *mptr, USI *nptr, USI *optr)
/*
 * returns 1 if unsigned ints *jptr, *kptr, *lptr, *mptr, *nptr, *optr
 * with *jptr < M0, *mptr < M0, *nptr < M0 successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u%u", jptr, kptr, lptr, mptr, nptr, optr);
	Flush();
	if (n == 6)
	{
		if (*jptr < M0 && *mptr < M0 && *nptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *mptr, *nptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf22(USI *iptr, USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr with *kptr and
 * *lptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", iptr, jptr, kptr, lptr);
	Flush();
	if (n == 4)
	{
		if (*kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("third integer %u or fourth integer %u exceeded array bound %u, try again:\n", *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf20(USI *iptr, USI *jptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u", iptr, jptr);
	Flush();
	if (n == 2)
		return 1;
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf30(USI *iptr, USI *jptr, USI *kptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr are succesfully entered,
 * 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u", iptr, jptr, kptr);
	Flush();
	if (n == 3)
		return 1;
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf40(USI *iptr, USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr are successfully 
 * entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", iptr, jptr, kptr, lptr);
	Flush();
	if (n == 4)
		return 1;
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf23(USI *iptr, USI *jptr, USI *kptr, USI *lptr, USI *mptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr, *mptr with
 * *kptr, *lptr and *mptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u", iptr, jptr, kptr, lptr, mptr);
	Flush();
	if (n == 5)
	{
		if (*kptr < M0 && *lptr < M0 && *mptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *kptr, *lptr, *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf231(USI *iptr, USI *jptr, USI *kptr, USI *lptr, USI *mptr, USI *nptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr, *mptr, *nptr with
 * *kptr, *lptr and *mptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u%u%u", iptr, jptr, kptr, lptr, mptr, nptr);
	Flush();
	if (n == 6)
	{
		if (*kptr < M0 && *lptr < M0 && *mptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *kptr, *lptr, *mptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

int scanf13(USI *iptr, USI *jptr, USI *kptr, USI *lptr)
/*
 * returns 1 if unsigned ints *iptr, *jptr, *kptr, *lptr with
 * *jptr, *kptr and *lptr < M0 are successfully entered, 0 otherwise.
 */
{
	int n;

	n = scanf("%u%u%u%u", iptr, jptr, kptr, lptr);
	Flush();
	if (n == 4)
	{
		if (*jptr < M0 && *kptr < M0 && *lptr < M0)
			return 1;
		else
			printf("one of %u, %u and %u exceeded array bound %u, try again:\n", *jptr, *kptr, *lptr, M0 - 1);
	}
	else
		printf("try again:\n");
	GetReturn();
	return 0;
}

FILE *Fopen_r(char *filename)
/*
 * An attempt is made at opening the file specified for reading. If successful,
 * a pointer to the file is returned, otherwise a null pointer is returned.
 */
{
	FILE *f;
	if ((f = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "cmat: ");
		perror(filename);
		exit(1);
	}
	return f;
}

void printfile(char *file)
{
#ifdef _WIN32
	char buf[512];
        printf("filename is %s\n");
	sprintf(buf, "type %s > PRN", file);
	system(buf);
#else
/*	spawnl(0, LPR, "lpr", file, NULL);*/
	char buf[512];
	sprintf(buf, "lpr %s", file);
	system(buf);

#endif
}

void fatal_error(char *fmt, ...)
/*
 * Prints error message with formatting.
 * Then exits with error condition.
 */
{
	va_list ap;

	va_start(ap, fmt);
	fputs("cmat: ", stderr);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
	exit(1);
}
