/* p1CR.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

MPMATPCR *BUILDMATPCR(USI m, USI n)
/*
 * Allocates space for an m x n matrix of POLYCR's.
 */
{
	unsigned int i;
	MPMATPCR *N;

	N=(MPMATPCR *)mmalloc(sizeof(MPMATPCR));
	N->R = m;
	N->C = n;
	N->V = (POLYCR **)mmalloc(m * sizeof(POLYCR *));
	for (i = 0; i <= m - 1; i++)
		*(N->V + i) = (POLYCR *)mmalloc(n * sizeof(POLYCR));
	return (N);
}

MPMATPCR *COPYMATPCR(MPMATPCR *Mptr)
/*
 * a replacement for the assignment *Nptr = *Mptr. 
 */
{
	unsigned int i, j;
	MPMATPCR *Nptr;

	Nptr = BUILDMATPCR(Mptr->R, Mptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Mptr->C - 1; j++)
			Nptr->V[i][j] = COPYPCR(Mptr->V[i][j]);
	}
	return (Nptr);
}

void FREEMATPCR(MPMATPCR *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V.
 */
{
	unsigned int i, j;

	if (Mptr == NULL)
		return;
	for (i = 0; i < Mptr->R; i++)
	{
		for (j = 0; j < Mptr->C; j++)
			DELETE_PCR(Mptr->V[i][j]);
		ffree((char *)(*(Mptr->V + i)), Mptr->C * sizeof(POLYCR));
	}
	ffree((char *)(Mptr->V), Mptr->R * sizeof(POLYCR *));
	ffree((char *)Mptr, sizeof(MPMATPCR));
}

MPMATPCR *INPUTMATPCR()
/*
 * inputting a matrix of POLYCR'S.
 */
{
	unsigned int u, i, j, m, n;
	MPMATPCR *Mptr;

	printf("Enter the number of rows and number of columns\n");
	if (scanf("%u%u", &m, &n) != 2 || m == 0 || n == 0)
	{
		fprintf(stderr, "corrupt input: returning the 1 x 1 zero matrix\n");
		return ZEROMNPCR(1, 1);
	}
	getchar();
	Mptr = ZEROMNPCR(m, n);
	for (i = 0; i <= m - 1; i++) 
        {
		printf("Enter row %u:\n", i + 1);
		for (j = 0; j <= n - 1; j++) 
		{
			printf("Entering polynomial[%u][%u]:\n", i + 1, j + 1);
			Mptr->V[i][j] = INPUTPCR(&u);
			if (u == 0)
			{
				printf("matrix input prematurely interrupted\n");
				GetReturn();
				return Mptr;
			}
			printf("Polynomial[%u][%u] = ", i + 1, j + 1);
			PRINTPCR(Mptr->V[i][j]);
			printf("\n");
		}
	}
	return (Mptr);
}

POLYCR *SMITHPCR(MPMATPCR *Mptr, MPMATPCR **Pptr, MPMATPCR **Qptr, USI *ptr)
/*
 * returns the invariant factors of *Mptr.
 * **Pptr and **Qptr are invertible matrices such that
 * **Pptr x *Mptr x **Qptr is the Smith canonical form *Nptr below.
 * *ptr is the number of invariant factors.
 * See Rings, Modules and Linear Algebra by B. Hartley and T. O. Hawkes,
 * Chapman and Hall, 1970.
 */
{
	unsigned int i, k, l, m, n, p, q, t, z;
	MPCR *G, *TempCR;
	POLYCR Q, *N, Tmp;
	MPMATPCR *Nptr, *Temp;
	
	m = Mptr->R;
	n = Mptr->C;
	Nptr = COPYMATPCR(Mptr);
	*Pptr = IDENTITYPCR(m);
	*Qptr = IDENTITYPCR(n);
	z = MIN((int) m - 1, (int) n - 1);
	N = (POLYCR *)mmalloc((z + 1) * sizeof(POLYCR));
	for (i = 0; i <= z; i++)
	{
		ZEROTESTPCR(0, Nptr, &p, &q);
		if (p == Nptr->R && q == Nptr->C)
			break;
		if (p > 0)
		{
			Nptr = SWAP_ROWSPCR1(0, p, Nptr);
			*Pptr = SWAP_ROWSPCR1(i, p + i, *Pptr);
		}
		if (q > 0)
		{
			Nptr = SWAP_COLSPCR1(0, q, Nptr);
			*Qptr = SWAP_COLSPCR1(i, q + i, *Qptr);
		}
		/* Now Nptr->V[0][0] != NULL. */
	while (1)
	{
		while (1)
		{
			l = ROWSEEKPCR(0, Nptr);
			if (l == Nptr->C)
				break;
			else
			{
				/* Case 1, p. 112. */
				Q = DIVPCR(Nptr->V[0][l], Nptr->V[0][0]);
				Nptr = COLSUBPCR1(0, l, Q, Nptr); 
				*Qptr = COLSUBPCR1(i, l + i, Q, *Qptr); 
				DELETE_PCR(Q);
				Nptr = SWAP_COLSPCR1(0, l, Nptr); 
				*Qptr = SWAP_COLSPCR1(i, l + i, *Qptr); 
			}
		}
/* Now all elts to the right of Nptr->V[0][0] are divisible by it. */
		while (1)
		{
			k = COLSEEKPCR(0, Nptr);
			if (k == Nptr->R)
				break;
			else
			{
				/* Case 2, p. 112. */
				Q = DIVPCR(Nptr->V[k][0], Nptr->V[0][0]);
				Nptr = ROWSUBPCR1(0, k, Q, Nptr);
				*Pptr = ROWSUBPCR1(i, k + i, Q, *Pptr);
				DELETE_PCR(Q);
				Nptr = SWAP_ROWSPCR1(0, k, Nptr); 
				*Pptr = SWAP_ROWSPCR1(i, k + i, *Pptr); 
				/* This reduces the degree of Nptr->V[0][0]. */
			}
		}
/* Now (Case 3, p.112) Nptr->V[0][0] divides each element to its right as well as below it.*/
		for (l = 1; l < Nptr->C; l++)
		{
			if (Nptr->V[0][l] != NULL)
			{
				Q = DIVPCR(Nptr->V[0][l], Nptr->V[0][0]);
				Nptr = COLSUBPCR1(0, l, Q, Nptr); 
				*Qptr = COLSUBPCR1(i, l + i, Q, *Qptr); 
				DELETE_PCR(Q);
			}
		}
		for (k = 1; k < Nptr->R; k++)
		{
			if ( Nptr->V[k][0] != NULL)
			{
				Q = DIVPCR(Nptr->V[k][0], Nptr->V[0][0]);
				*Pptr = ROWSUBPCR1(i, k + i, Q, *Pptr);
				DELETE_PCR(Q);
				DELETE_PCR(Nptr->V[k][0]);
				Nptr->V[k][0] = NULL;
			}
		}
		/* Now each element to the right of and below 
		Nptr->V[0][0] is zero. */
		k = MATSEEKPCR(0, Nptr);
		if (k != Nptr->R)
		{
			for (t = 0; t < Nptr->C; t++)
			{
				if (Nptr->V[k][t] != NULL)
				{
					Tmp = Nptr->V[0][t];
					Nptr->V[0][t] = ADDPCR(Nptr->V[0][t], Nptr->V[k][t]);
					DELETE_PCR(Tmp);
				}
			}
			for (t = 0; t < m; t++)
			{
				if ((*Pptr)->V[k+i][t] != NULL)
				{
					Tmp = (*Pptr)->V[i][t];
					(*Pptr)->V[i][t] = ADDPCR((*Pptr)->V[i][t], (*Pptr)->V[k+i][t]);
					DELETE_PCR(Tmp);
				}
			}
		}
		else
		{
/* Now Nptr->V[0][0] divides all elements Nptr->V[k][l], k,l > 0. */
			G = LEADPCR(Nptr->V[0][0]);
			TempCR = G;
			G = INVERSECR(G);
			FREEMPCR(TempCR);
			Temp = *Pptr;
			*Pptr = SCALAR_ROWPCR(i, G, Temp);
			FREEMATPCR(Temp);
			FREEMPCR(G);
			Tmp = Nptr->V[0][0];
			Nptr->V[0][0] = MONICPCR(Nptr->V[0][0]);
			DELETE_PCR(Tmp);
			N[i] = COPYPCR(Nptr->V[0][0]);
			break;
		}
	}
		if (i < z)
		{
			Temp = Nptr;
			Nptr = DELETE_ROWPCR(1, Nptr);
			FREEMATPCR(Temp);
			Temp = Nptr;
			Nptr = DELETE_COLPCR(1, Nptr);
			FREEMATPCR(Temp);
		}
	}
	FREEMATPCR(Nptr);
	*ptr = i++;
	return (N);
}

POLYCR *SMITH1PCR(MPMATPCR *Mptr, USI *ptr)
/*
 * returns the invariant factors of *Mptr.
 * See Rings, Modules and Linear Algebra by B. Hartley and T. O. Hawkes,
 * Chapman and Hall, 1970.
 */
{
	unsigned int i, k, l, m, n, p, q, t, z;
	POLYCR Q, *N, Tmp;
	MPMATPCR *Nptr, *Temp;
	
	m = Mptr->R;
	n = Mptr->C;
	Nptr = COPYMATPCR(Mptr);
	z = MIN((int) m - 1, (int) n - 1);
	N = (POLYCR *)mmalloc((z + 1) * sizeof(POLYCR));
	for (i = 0; i <= z; i++)
	{
		ZEROTESTPCR(0, Nptr, &p, &q);
		if (p == Nptr->R && q == Nptr->C)
			break;
		if (p > 0)
			Nptr = SWAP_ROWSPCR1(0, p, Nptr);
		if (q > 0)
			Nptr = SWAP_COLSPCR1(0, q, Nptr);
		/* Now Nptr->V[0][0] != NULL. */
	while (1)
	{
		while (1)
		{
			l = ROWSEEKPCR(0, Nptr);
			if (l == Nptr->C)
				break;
			else
			{
				/* Case 1, p. 112. */
				Q = DIVPCR(Nptr->V[0][l], Nptr->V[0][0]);
				Nptr = COLSUBPCR1(0, l, Q, Nptr); 
				DELETE_PCR(Q);
				Nptr = SWAP_COLSPCR1(0, l, Nptr); 
			}
		}
/* Now all elts to the right of Nptr->V[0][0] are divisible by it. */
		while (1)
		{
			k = COLSEEKPCR(0, Nptr);
			if (k == Nptr->R)
				break;
			else
			{
				/* Case 2, p. 112. */
				Q = DIVPCR(Nptr->V[k][0], Nptr->V[0][0]);
				Nptr = ROWSUBPCR1(0, k, Q, Nptr);
				DELETE_PCR(Q);
				Nptr = SWAP_ROWSPCR1(0, k, Nptr); 
				/* This reduces the degree of Nptr->V[0][0]. */
			}
		}
/* Now (Case 3, p.112) Nptr->V[0][0] divides each element to its right as well as below it.*/
		for (l = 1; l < Nptr->C; l++)
		{
			if (Nptr->V[0][l] != NULL)
			{
				Q = DIVPCR(Nptr->V[0][l], Nptr->V[0][0]);
				Nptr = COLSUBPCR1(0, l, Q, Nptr); 
				DELETE_PCR(Q);
			}
		}
		for (k = 1; k < Nptr->R; k++)
		{
			if ( Nptr->V[k][0] != NULL)
			{
				DELETE_PCR(Nptr->V[k][0]);
				Nptr->V[k][0] = NULL;
			}
		}
		/* Now each element to the right of and below 
		Nptr->V[0][0] is zero. */
		k = MATSEEKPCR(0, Nptr);
		if (k != Nptr->R)
		{
			for (t = 0; t < Nptr->C; t++)
			{
				if (Nptr->V[k][t] != NULL)
				{
					Tmp = Nptr->V[0][t];
					Nptr->V[0][t] = ADDPCR(Nptr->V[0][t], Nptr->V[k][t]);
					DELETE_PCR(Tmp);
				}
			}
		}
		else
		{
			/* Now Nptr->V[0][0] divides all elements
			Nptr->V[k][l], k,l > 0. */
			Tmp = MONICPCR(Nptr->V[0][0]);
			DELETE_PCR(Nptr->V[0][0]);
			Nptr->V[0][0] = Tmp;
			N[i] = COPYPCR(Nptr->V[0][0]);
			break;
		}
	}
		if (i < z)
		{
			Temp = Nptr;
			Nptr = DELETE_ROWPCR(1, Nptr);
			FREEMATPCR(Temp);
			Temp = Nptr;
			Nptr = DELETE_COLPCR(1, Nptr);
			FREEMATPCR(Temp);
		}
	}
	FREEMATPCR(Nptr);
	*ptr = i++;
	return (N);
}
