/* yacc parsing for CMAT. */
/* See Kernighan and Pike, the Unix programming environment.*/
%{
#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"
#define YYSTYPE MPIPTR
static unsigned int i, j;
MPI *retval;
%}
%token NUMBER
%left '+' '-'
%left '*' '/'
%right '^'
%%
mpi:    list			{ retval = $$; }
	;
list:
	| list expr		{ $$ = $2; }
	;
expr:	  NUMBER		{ $$ = $1; }
	| 'i'			{ $$ = CHANGE(i); }
	| 'j'			{ $$ = CHANGE(j); }
	| expr '+' expr		{ $$ = ADDI($1, $3); FREEMPI($1); FREEMPI($3); }
	| expr '-' expr		{ $$ = SUBI($1, $3); FREEMPI($1); FREEMPI($3); }
	| expr '*' expr		{ $$ = MULTI($1, $3); FREEMPI($1); FREEMPI($3); }
	| expr '/' expr		{ $$ = INTI($1, $3); FREEMPI($1); FREEMPI($3); }
	| expr '^' expr		{ $$ = POWERI($1, CONVERTI($3)); FREEMPI($1); FREEMPI($3); }
	| '(' expr ')'		{ $$ = $2; }
	;
%%

#include <ctype.h>

static char *p;

MPI *Parse(char *s, int v1, int v2)
{
	i = v1;
	j = v2;
	p = s;
	yyparse();
	return retval;
}

int yylex()
{
	MPI *Temp;
	char c;
	while ((c = *p++) == ' ' || c == '\t')
		;
	if (c == '\0')
		return 0;
	if (isdigit(c))
	{
		yylval = CHANGE(c - '0');
		while (isdigit(*p))
		{
			Temp = yylval;
			yylval = MULT_I(yylval, 10);
			FREEMPI(Temp);
			Temp = yylval;
			yylval = ADD0_I(yylval, *p++ - '0');
			FREEMPI(Temp);
		}
		return NUMBER;
	}
	return c;
}

yyerror(s)
char *s;
{
	fprintf(stderr, "test: %s\n", s);
}
