/* cyclo.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"
#define S0 65536

void NEXTPHI(int p, int *gptr, int *b, int *coef)
{
	int cycle, next, i,j, lastgrad, l, grad, w;

	cycle=p;
	grad = *gptr;
	next=grad;
	for(i=grad; i>=0; i--)
	{
	  if(cycle==p)
	  {
	    cycle=1;
	    b[i]=coef[next];
	    next=next-1;
	  }
	  else
	  {
	    cycle++;
	    b[i]=0;
	  }
	}
	lastgrad=grad;
	grad=grad*(p-1);

	for(i=grad; i>= (grad/2); i--)
	/* PHI = PHI(x^p)/PHI.  First half of the coefficients. */
	{
	  l = b[lastgrad];
	  coef[i]=l;
	  for(j=lastgrad-1; j>=0; j--)
	  {
	    w = l*(coef[j]);
	    b[j+1]=b[j]-w;
/*
		if (abs(w) >= S0 || abs(b[j+1]) >= S0)
		{
			fprintf(stderr, "coefficient >= 65536\n");
			exit (1);
		}
*/
	  }
	  if(cycle==p)
	  {
	    cycle=1;
	    b[0]=coef[next];
	    next--;
	  }
	  else
	  {
	    cycle++;
	    b[0]=0;
	  }
	}

	for(j=0; j<=(grad/2)-1; j++)
	  coef[j]=coef[grad-j];
	*gptr = grad;
}

int NOTYET(int n, int i, int *qptr)
{
	int q;

	q = n / i;
	*qptr = q;
	if ((n > q * i) && (i <= q))
		return 1;
	else
		return 0;
}

int LPF(int n)
{
	int c, i, q;

	if (n % 2 == 0) return (2);
	if (n % 3 == 0) return (3);
	else
	{
		i = 5;
		c = 0;
		while (NOTYET(n, i, &q))
		{
			i = (c == 0) ? i + 2 : i + 4;
			c = 1 - c;
		}
		if (q < i)
			return (n);
		else
			return (i);	
	}
}

POLYI CYCLOTOMIC(int n)
{
	int i, n1, socn, p, grad, *b, *coef;
	POLYI P;
	MPI *H;

	b = (int *)mmalloc(n * sizeof(int));
	coef = (int *)mmalloc(n * sizeof(int));
	n1 = n;
	if (n1 == 1)
	{
		grad = 1;
		coef[0] = -1;
		coef[1] = 1;
	}
	else
	{
		p = LPF(n1);
		socn = p;
		while (n1 % p == 0)
			n1 = n1 / p;
		grad = p - 1;
		for (i = 0; i <= grad; i++)
			coef[i] = 1;
		/* PHI is the pth cyclotomic polynomial. */
			
		while (n1 > 1)
		{
			p = LPF(n1);
			socn = socn * p;
			while (n1 % p == 0)
				n1 = n1 / p;
			NEXTPHI(p, &grad, b, coef);
		/* PHI is the socnth cyclotomic polynomial. */		
		}
		n1 = n / socn;
		if (n1 > 1)
		{
		/* PHI = PHI(x^n). */
			grad = n1 * grad;
			for (i = grad; i >= 0; i--)
			{
				if (i % n1 == 0)
					coef[i] = coef[i/ n1];
				else
					coef[i] = 0;
			}
		}	
	}
	ffree((char *)b, n * sizeof(int));
	P = NULL;
	for (i = 0; i <= grad; i++)
	{
		if (coef[i])
		{
			H = CHANGEI(coef[i]);
			PINSERTPI(i, H, &P, 0);
			FREEMPI(H);
		}
	}
	ffree((char *)coef, n * sizeof(int));
	return (P);
}
