/* p2m.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "integer.h"
#include "fun.h"

void ZEROTESTPm(USI i, MPMATPm *Mptr, USI *p, USI *q)
/*
 * returns the least indices *p >= i, *q >= i such that Mptr->V[*p][*q] != 0.
 * otherwise *p = Mptr->R and *q = Mptr->C.
 */
{
	unsigned int k, l;
	
	
	for (k = i; k <= Mptr->R - 1; k++)
	{
		for (l = i; l <= Mptr->C - 1; l++)
		{
			if (Mptr->V[k][l] != NULL)
			{
				*p = k;
				*q = l;
				return;
			}
		}
	}
	*p = Mptr->R;
	*q = Mptr->C;
	return;
}

unsigned int MATSEEKPm(USI i, MPMATPm *Mptr, USI p)
/*
 * returns the least row index k, i < k <= Mptr->R - 1, such that 
 *Mptr->V[k][l] is not divisible mod p by Mptr->V[i][i], for some l, 
 * i < l <= Mptr->C - 1; otherwise returns Mptr->R.
 */
{
	unsigned int k, l;
	POLYm P;
	
	for (k = i + 1; k <= Mptr->R - 1; k++)
	{
		for (l = i + 1; l <= Mptr->C - 1; l++)
		{
			if ((P = MODPm(Mptr->V[k][l], Mptr->V[i][i], p)) != NULL)
			{
				DELETE_Pm(P);
				return (k);
			}
		}
	}
	return (k);
}

unsigned int ROWSEEKPm(USI i, MPMATPm *Mptr, USI p)
/*
 * returns the least integer l > i such that Mptr->V[i][l] is not divisible
 * (mod p) by Mptr->V[i][i], otherwise returns Mptr->C.
 */
{
	unsigned int l;
	POLYm P;

	for (l = i + 1; l <= Mptr->C - 1; l++)
	{
		if ((P = MODPm(Mptr->V[i][l], Mptr->V[i][i], p)) != NULL)
		{
			DELETE_Pm(P);
			return (l);
		}
	}
	return (l);
}

unsigned int COLSEEKPm(USI i, MPMATPm *Mptr, USI p)
/*
 * returns the least integer k > i such that Mptr->V[k][i] is not divisible
 * (mod p) by Mptr->V[i][i], otherwise returns Mptr->C.
 */
{
	unsigned int k;
	POLYm P;

	for (k = i + 1; k <= Mptr->R - 1; k++)
	{
		if ((P = MODPm(Mptr->V[k][i], Mptr->V[i][i], p)) != NULL)
		{
			DELETE_Pm(P);
			return (k);
		}
	}
	return (k);
}

MPMATPm *ROWSUBPm(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * subtracts (mod z) *Aptr times the p-th row of *Mptr from the q-th.
 * 0 <= p, q <= Mprt->R - 1.
 */
{
	POLYm Y, Temp;
	unsigned int l;
	MPMATPm *Nptr;
	Nptr = COPYMATPm(Mptr, z);
	if (Aptr != NULL)
	{
		for (l = 0; l <= Nptr->C - 1; l++) 
		{
			if (Nptr->V[p][l] != NULL)
			{
				Y = MULTPm(Aptr, Nptr->V[p][l], z);
				Temp = Nptr->V[q][l];
				Nptr->V[q][l] = SUBPm(Nptr->V[q][l], Y, z);
				DELETE_Pm(Y);
				DELETE_Pm(Temp);
			}
		}
	}
	return (Nptr);
}

MPMATPm *COLSUBPm(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * subtracts *Aptr times the p-th column of *Mptr from the q-th.
 * 0 <= p, q <= Mprt->C - 1.
 */
{
	POLYm Y, Temp;
	unsigned int k;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, z);
	if (Aptr != NULL)
	{
		for (k = 0; k <= Nptr->R - 1; k++) 
		{
			if (Nptr->V[k][p] != NULL)
			{
				Y = MULTPm(Aptr, Nptr->V[k][p], z);
				Temp = Nptr->V[k][q];
				Nptr->V[k][q] = SUBPm(Nptr->V[k][q], Y, z);
				DELETE_Pm(Y);
				DELETE_Pm(Temp);
			}
		}
	}
	return (Nptr);
}

MPMATPm *ROWSUBPm1(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * subtracts (mod z) *Aptr times the p-th row of *Mptr from the q-th.
 * 0 <= p, q <= Mprt->R - 1.
 */
{
	POLYm Y, Temp;
	unsigned int l;
 
	if (Aptr != NULL)
	{
		for (l = 0; l <= Mptr->C - 1; l++) 
		{
			if (Mptr->V[p][l] != NULL)
			{
				Y = MULTPm(Aptr, Mptr->V[p][l], z);
				Temp = Mptr->V[q][l];
				Mptr->V[q][l] = SUBPm(Mptr->V[q][l], Y, z);
				DELETE_Pm(Y);
				DELETE_Pm(Temp);
			}
		}
	}
	return (Mptr);
}

MPMATPm *COLSUBPm1(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * subtracts *Aptr times the p-th column of *Mptr from the q-th.
 * 0 <= p, q <= Mprt->C - 1.
 */
{
	POLYm Y, Temp;
	unsigned int k;

	if (Aptr != NULL)
	{
		for (k = 0; k <= Mptr->R - 1; k++) 
		{
			if (Mptr->V[k][p] != NULL)
			{
				Y = MULTPm(Aptr, Mptr->V[k][p], z);
				Temp = Mptr->V[k][q];
				Mptr->V[k][q] = SUBPm(Mptr->V[k][q], Y, z);
				DELETE_Pm(Y);
				DELETE_Pm(Temp);
			}
		}
	}
	return (Mptr);
}

MPMATPm *ADD_MULT_ROWPm(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * adds (mod z) *Aptr times the p-th row of *Mptr to the q-th.
 * 0 <= p, q <= Mprt->R - 1.
 */
{
	POLYm Y, Tmp;
	unsigned int l;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, z);
	if (Aptr != NULL)
	{
		for (l = 0; l <= Nptr->C - 1; l++) 
		{
			Y = MULTPm(Aptr, Nptr->V[p][l], z);
			Tmp = Nptr->V[q][l];
			Nptr->V[q][l] = ADDPm(Tmp, Y, z);
			DELETE_Pm(Y);
			DELETE_Pm(Tmp);
		}
	}
	return (Nptr);
}

MPMATPm *ADD_MULT_COLPm(USI p, USI q, POLYm Aptr, MPMATPm *Mptr, USI z)
/*
 * adds (mod z) *Aptr times the p-th column of *Mptr to the q-th.
 * 0 <= p, q <= Mprt->R - 1.
*/
{
	POLYm Y, Tmp;
	unsigned int l;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, z);
	if (Aptr != NULL)
	{
		for (l = 0; l <= Nptr->R - 1; l++) 
		{
			Y = MULTPm(Aptr, Nptr->V[l][p], z);
			Tmp = Nptr->V[l][q];
			Nptr->V[l][q] = ADDPm(Tmp, Y, z);
			DELETE_Pm(Y);
			DELETE_Pm(Tmp);
		}
	}
	return (Nptr);
}

MPMATPm *SWAP_ROWSPm(USI p, USI q, MPMATPm *Mptr, USI z)
/*
 * 0 <= p < q <= Mprt->R - 1.
 * transposes the p-th and q-th rows of *Mptr.
 */
{
	POLYm X;
	unsigned int l;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, z);
	for (l = 0; l <= Nptr->C - 1; l++) 
	{
		X = Nptr->V[p][l];
		Nptr->V[p][l] = Nptr->V[q][l];
		Nptr->V[q][l] = X;
	}
	return (Nptr);
}

MPMATPm *SWAP_COLSPm(USI p, USI q, MPMATPm *Mptr, USI z)
/*
 * 0 <= p < q <= Mprt->C - 1.
 * transposes the p-th and q-th columns of *Mptr.
 */
{
	POLYm Temp;
	unsigned int k;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, z);
	for (k = 0; k <= Nptr->R - 1; k++)
        {
		Temp = Nptr->V[k][p];
		Nptr->V[k][p] = Nptr->V[k][q];
		Nptr->V[k][q] = Temp;
	}
	return (Nptr);
}

MPMATPm *SWAP_ROWSPm1(USI p, USI q, MPMATPm *Mptr, USI z)
/*
 * 0 <= p < q <= Mprt->R - 1.
 * transposes the p-th and q-th rows of *Mptr.
 */
{
	POLYm X;
	unsigned int l;

	for (l = 0; l <= Mptr->C - 1; l++) 
	{
		X = Mptr->V[p][l];
		Mptr->V[p][l] = Mptr->V[q][l];
		Mptr->V[q][l] = X;
	}
	return (Mptr);
}

MPMATPm *SWAP_COLSPm1(USI p, USI q, MPMATPm *Mptr, USI z)
/*
 * 0 <= p < q <= Mprt->C - 1.
 * transposes the p-th and q-th columns of *Mptr.
 */
{
	POLYm Temp;
	unsigned int k;

	for (k = 0; k <= Mptr->R - 1; k++)
        {
		Temp = Mptr->V[k][p];
		Mptr->V[k][p] = Mptr->V[k][q];
		Mptr->V[k][q] = Temp;
	}
	return (Mptr);
}

MPMATPm *DIAGXPm(MPMATm *Mptr, USI z)
/*
 * returns the polynomial matrix xI - *Mptr mod z.
 */
{
	unsigned int g, i, j, m;
	MPMATPm *Nptr;

	m = Mptr->R;
	Nptr = BUILDMATPm(m, m);
	for (i = 0; i <= m - 1; i++)
	{
		for (j = 0; j <= m - 1; j++)
		{
			if (i == j)
			{
				g = MINUSm(Mptr->V[i][i], z);
				Nptr->V[i][i] = LINEARPm(g, z);
			}
			else
			{
				g = MINUSm(Mptr->V[i][j], z);
				Nptr->V[i][j] = CONSTANTPm(g);
			}
		}
	}
	return (Nptr);
}

MPMATPm *IDENTITYPm(USI n)
/*
 * returns the identity polynomial matrix of size n.
 */
{
	unsigned int i, j;
	MPMATPm *Mptr;

	Mptr = BUILDMATPm(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( i == j)
				Mptr->V[i][i] =  CONSTANTPm(1);
			else
				Mptr->V[i][j] =  NULL;
		}
	}
	return (Mptr);
}

MPMATPm *SCALAR_ROWPm(USI p, USI a, MPMATPm *Mptr, USI z)
/*
 * multiplies row p (mod z) by the scalar a.
 */
{
	unsigned int l;
	MPMATPm *Nptr;
	POLYm Tmp;

	Nptr = COPYMATPm(Mptr, z);
	for (l = 0; l <= Nptr->C - 1; l++)
	{
		Tmp = Nptr->V[p][l];
		Nptr->V[p][l] = SCALARPm(a, Tmp, z);
		DELETE_Pm(Tmp);
	}
	return (Nptr);
}

MPMATPm *SCALAR_COLPm(USI p, USI a, MPMATPm *Mptr, USI z)
/*
 * multiplies column p (mod z) by the scalar a.
 */
{
	unsigned int l;
	MPMATPm *Nptr;
	POLYm Tmp;

	Nptr = COPYMATPm(Mptr, z);
	for (l = 0; l <= Nptr->R - 1; l++)
	{
		Tmp = Nptr->V[l][p];
		Nptr->V[l][p] = SCALARPm(a, Tmp, z);
		DELETE_Pm(Tmp);
	}
	return (Nptr);
}

POLYm INPUT1Pm(FILE *infile)
/*
 * Inputs a polynomial from a file created by FPRINT1Pm below.
 */
{
	unsigned NoTerms, i, COEF;
	int k;
	POLYm Temp;

	fscanf(infile, "%u", &NoTerms);
	Temp = NULL;
	if (NoTerms)
	{
		for (i = 0; i < NoTerms; i++)
		{
			fscanf(infile, "%d", &k);
			fscanf(infile, "%u", &COEF);
			PINSERTPm(k, COEF, &Temp, 0, 1);
		}
	}
	return Temp;
}

unsigned int CARDPm(POLYm Pptr)
/*
 * returns the number of terms in Pptr (0 if Pptr = NULL).
 */
{
	unsigned int i = 0;
	POLYm CURSOR;

	if (Pptr == NULL)
		return (0);
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		i++;
		CURSOR = CURSOR->NEXT;
	}
	return (i);
}

void FPRINT1Pm(FILE *outfile, POLYm Pptr)
{
	unsigned int l;
	POLYm CURSOR;

	l = CARDPm(Pptr);
	fprintf(outfile, "%u ", l);
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		fprintf(outfile, "%u ", CURSOR->DEG);
		fprintf(outfile, "%u ", CURSOR->COEF);
		CURSOR = CURSOR->NEXT;
	}
}

unsigned int LENGTHPm(POLYm Pptr)
{
	POLYm CURSOR;
	unsigned int d = 0;
	int k;

	if (Pptr == NULL)
	{
		d++;
		return (d);
	}
	else
	{
		CURSOR = Pptr;
		k = DEGREEPm(Pptr);
		while (CURSOR != NULL)
		{
			if (CURSOR->DEG != k)
				d += 2;
			if (CURSOR->DEG == 0 || CURSOR->COEF != 1)
				d += LENGTHm(CURSOR->COEF);
			if (CURSOR->DEG == 1)
				d++;
			if (CURSOR->DEG > 1)
				d += 2 + CURSOR->DEG;
			if (CURSOR->NEXT != NULL)
				d++;
			CURSOR = CURSOR->NEXT;
		}
		return (d);
	}
}

void FPRINTPm(FILE *outfile, POLYm Pptr)
{
	POLYm CURSOR;
	int k;

	if (Pptr == NULL)
	{
		fprintf(outfile, "0");
		return;
	}
	else
	{
		CURSOR = Pptr;
		k = DEGREEPm(Pptr);
		while (CURSOR != NULL)
		{
			if (CURSOR->DEG != k)
				fprintf(outfile, "+ ");
			if (CURSOR->DEG == 0 || CURSOR->COEF != 1)
				fprintf(outfile, "%u", CURSOR->COEF);
			if (CURSOR->DEG == 1)
				fprintf(outfile, "x");
			if (CURSOR->DEG > 1)
				fprintf(outfile, "x^%d", CURSOR->DEG);
			if (CURSOR->NEXT != NULL)
				fprintf(outfile, " ");
			CURSOR = CURSOR->NEXT;
		}
		return;
	}
}

void SPRINTPm(char *buffer, POLYm Pptr)
{
	POLYm CURSOR;
	int k;

	if (Pptr == NULL)
	{
		sprintf(buffer, "0");
		return;
	}
	else
	{
		CURSOR = Pptr;
		k = DEGREEPm(Pptr);
		while (CURSOR != NULL)
		{
			if (CURSOR->DEG != k)
			{
				sprintf(buffer, "+ ");
				buffer += 2;
			}
			if (CURSOR->DEG == 0 || CURSOR->COEF != 1)
			{
				sprintf(buffer, "%u", CURSOR->COEF);
				buffer += strlen(buffer);
			}
			if (CURSOR->DEG == 1)
			{
				sprintf(buffer, "x");
				buffer++;
			}
			if (CURSOR->DEG > 1)
			{
				sprintf(buffer, "x^%d", CURSOR->DEG);
				buffer += strlen(buffer);
			}
			if (CURSOR->NEXT != NULL)
			{
				sprintf(buffer, " ");
				buffer++;
			}
			CURSOR = CURSOR->NEXT;
		}
		return;
	}
}

void FPRINTMATPm(FILE *outfile, USI i1, USI i2, USI j1, USI j2, MPMATPm *Mptr)
/*
 * Modified 8/1/93 using Peter Adams' improvement from, August 1992.
 */
{
	char **str, *buff;
	unsigned int tmp, i, j, nrow, ncol, len, nstr = 0;
	int *colwidth;
	unsigned int ct;
	
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	str = (char **)mmalloc(nrow * ncol * sizeof(char *));
	colwidth=(int *)mmalloc(ncol*sizeof(int));

	for(i=0; i<ncol; i++)
	  colwidth[i]=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			tmp = 1 + LENGTHPm(Mptr->V[i][j]);
			buff = (char *)mmalloc(tmp * sizeof(char));
			SPRINTPm(buff, Mptr->V[i][j]);
		    	if ((len = strlen(buff)) > colwidth[j-j1])
				colwidth[j-j1] = len;
			str[nstr++] = strcpy((char *)mmalloc(len + 1), buff);
			ffree(buff, tmp * sizeof(char));
		}
	}
	if (outfile != stdout)
		fprintf(outfile, "%u %u\n", nrow, ncol);
	ct=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
	    		fprintf(outfile, "%*s ", colwidth[j-j1], str[ct]);
	    		ffree(str[ct], (unsigned int) (strlen(str[ct]) + 1) * sizeof(char));
	    		if ((ct % ncol) == (ncol - 1))
				fprintf(outfile, "\n");
	    		ct++;
		}
	}
	ffree((char *) str, nrow * ncol * sizeof(char *));
	ffree((char *) colwidth, ncol*sizeof(int));
	return;
}

void PRINTMATPm(USI i1, USI i2, USI j1, USI j2, MPMATPm *Mptr)
{
	unsigned int i, j;

	i = i2 - i1 + 1;
	j = j2 - j1 + 1;
	if (MAX((int)i, (int)j) > 20)
	{
		printf("(The number of rows or columns to be printed exceeds 20;\n");
		printf("there is no point in printing this matrix on the screen.)\n");
		return;
	}
	FPRINTMATPm(stdout, i1, i2, j1, j2, Mptr);
}

void FPRINTMAT1Pm(FILE *outfile, USI i1, USI i2, USI j1, USI j2, MPMATPm *Mptr)
{
	unsigned i, j;

	fprintf(outfile, "%u %u\n", i2 - i1 + 1, j2 - j1 + 1);
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			FPRINT1Pm(outfile, Mptr->V[i][j]);
			fputc('\n', outfile);
		}
	}
}

MPMATPm *FINPUTMATPm(FILE *infile)
{
	unsigned i, j, m, n;
	MPMATPm *Temp;

	fscanf(infile, "%u%u\n", &m, &n);
	Temp = BUILDMATPm(m, n);
	for (i = 0; i < m; i++)
		for (j = 0; j < n; j++)
			Temp->V[i][j] = INPUT1Pm(infile);
	return Temp;
}
