/* p3R.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

POLYR EUCLIDPR(POLYR Pptr, POLYR Qptr, POLYR *Hptr, POLYR *Kptr)
/*
 * gcd(Pptr, Qptr) = Hptr * Pptr + Kptr * Qptr.
 */
{
	POLYR Q, A, B, C, H1, H2, K1, K2, L1, L2, Tmp, Tmp1, Tmp2;
	MPR *G, *TempR;

	if (Qptr == NULL)
	{
		if (Pptr != NULL)
		{
			G = LEADPR(Pptr);
			TempR = G;
			G = INVERSER(G);
			FREEMPR(TempR);
			*Hptr = CONSTANTPR(G);
			*Kptr = NULL;
			FREEMPR(G);
			return (MONICPR(Pptr));
		}
		else
		{
			*Hptr = NULL;
			*Kptr = NULL;
			return (NULL);
		}
	}
	A = COPYPR(Pptr);
	B = COPYPR(Qptr);
	C = MODPR(A, B);
	if (C == NULL)
	{
		*Hptr = NULL;
		G = LEADPR(Qptr);
		TempR = G;
		G = INVERSER(G);
		FREEMPR(TempR);
		*Kptr = CONSTANTPR(G);
		DELETE_PR(A);
		DELETE_PR(B);
		FREEMPR(G);
		return (MONICPR(Qptr));
	}
	TempR = ONER();
	L1 = CONSTANTPR(TempR);
	K1 = NULL;
	L2 = NULL;
	K2 = CONSTANTPR(TempR);
	FREEMPR(TempR);
	while (C != NULL)
	{
		Q = DIVPR(A, B);
		DELETE_PR(A);
		A = B;
		B = C;
		C = MODPR(A, B);
		Tmp1 = MULTPR(Q, K1);
		Tmp2 = MULTPR(Q, K2);
		DELETE_PR(Q);
		H1 = SUBPR(L1, Tmp1);
		H2 = SUBPR(L2, Tmp2);
		DELETE_PR(Tmp1);
		DELETE_PR(Tmp2);
		Tmp = L1;
		L1 = K1;
		DELETE_PR(Tmp);
		Tmp = L2;
		L2 = K2;
		DELETE_PR(Tmp);
		K1 = COPYPR(H1);
		K2 = COPYPR(H2);
		DELETE_PR(H1);
		DELETE_PR(H2);
	}
	G = LEADPR(B);
	TempR = G;
	G = INVERSER(G);
	FREEMPR(TempR);
	*Hptr = SCALARPR(G, K1);
	*Kptr = SCALARPR(G, K2);
	FREEMPR(G);
	DELETE_PR(K1);
	DELETE_PR(K2);
	DELETE_PR(L1);
	DELETE_PR(L2);
	DELETE_PR(A);
	DELETE_PR(C);
	Tmp = MONICPR(B);
	DELETE_PR(B);
	return (Tmp);
}

MPMATPR *MULTMATPR(MPMATPR *Mptr, MPMATPR *Nptr)
/*
 * Here Mptr->C = Nptr->R
 * returns (*Mptr) * (*Nptr)
 */
{
	POLYR X, Y, Tmp;
	MPMATPR *Qptr;
	unsigned int i, j, k;

	Qptr = BUILDMATPR(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			X = NULL;
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				Y = MULTPR(Mptr->V[i][k], Nptr->V[k][j]);
				Tmp = X;
				X = ADDPR(Tmp, Y);
				DELETE_PR(Tmp);
				DELETE_PR(Y);
			}
			Qptr->V[i][j] = X;
		}
	}
	return (Qptr);
}

POLYR TRACEPR(MPMATPR *Mptr)
/*
 * returns the trace of *Mptr
 */
{
	POLYR A, Tmp;
	unsigned int i;

	A = NULL;
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Tmp = A;
		A = ADDPR(Tmp, Mptr->V[i][i]);
		DELETE_PR(Tmp);
	}
	return (A);
}

MPMATPR *MINUSMATPR(MPMATPR *Mptr)
/*
 * returns -(*Mptr)
 */
{
	unsigned int i, j;
	MPMATPR *Nptr;
	MPR *TempR;

	TempR = MINUS_ONER();
	Nptr = BUILDMATPR(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			Nptr->V[i][j] = SCALARPR(TempR, Mptr->V[i][j]);
	}
	FREEMPR(TempR);
	return (Nptr);
}

MPMATPR *DIAGPR(MPMATPR *Mptr, POLYR Aptr)
/*
 * returns *Mptr - Aptr * (Unit Matrix)
 */
{
	unsigned int i;
	POLYR Tmp;
	MPMATPR *Nptr;

	Nptr = COPYMATPR(Mptr);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Tmp = Nptr->V[i][i];
		Nptr->V[i][i] = SUBPR(Tmp, Aptr);
		DELETE_PR(Tmp);
	}
	return (Nptr);
}

POLYR LEVERRIER_PR(MPMATPR *Mptr, MPMATPR **Nptr)
/*
 * The Faddeeva/Leverrier method for finding the 
 * adjoint and determinant,  applied to a polynomial matrix *Mptr.
 * returns the determinant while *Nptr is the adjoint.
 * See "Computational methods in linear algebra", by V. N. Faddeeva, Dover
 * 1959.
 */
{
	unsigned int i;
	MPMATPR *Qptr, *M;
	POLYR P, D, Tmp;
	MPI *G;
	MPR *H, *TempR;

	D = NULL;
	Qptr = COPYMATPR(Mptr);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		P = TRACEPR(Qptr);
		G = CHANGE(i + 1);
		H = BUILDMPR();
		H->N = ONEI();
		H->D = G;
		Tmp = P;
		P = SCALARPR(H, Tmp);
		FREEMPR(H);
		DELETE_PR(Tmp);
		M = DIAGPR(Qptr, P);
		FREEMATPR(Qptr);
		if (i == Mptr->R - 1)
			D = COPYPR(P);
		DELETE_PR(P);
		if (i == Mptr->R - 2)
		{
			if (i % 2 == 0)
				*Nptr = MINUSMATPR(M);
			else
				*Nptr = COPYMATPR(M);
		}  
		Qptr =	MULTMATPR(Mptr, M);
		FREEMATPR(M);
	}
	FREEMATPR(Qptr);
	if (i % 2 == 0)
	{
		Tmp = D;
		TempR = MINUS_ONER();
		D = SCALARPR(TempR, Tmp);
		FREEMPR(TempR);
		DELETE_PR(Tmp);
	}
	return (D);
}

POLYR EUCLIDSPR(POLYR Pptr[], POLYR *Hptr[], USI n)
/*
 * returns the gcd of polynomials Pptr[0],...,Pptr[n - 1]
 * and expresses it as Hptr[0] * Pptr[0] + ... Hptr[n - 1] * Pptr[n - 1].
 */
{
	unsigned int i;
	POLYR B, H, K, L, Tmp;
	MPR *TempR;

	*Hptr = (POLYR *)mmalloc(n * sizeof(POLYR));
	for (i = 0; i <= n - 1; i++)
	{
		*(*Hptr + i) = COPYPR(Pptr[i]);
	}
	for (i = 1; i <= n - 1; i++)
	{
		Tmp = *(*Hptr + i);
		*(*Hptr + i) = GCDPR(Tmp, *(*Hptr + (i - 1)));
		DELETE_PR(Tmp);
	}
	Tmp = *(*Hptr + (n - 1));
	B = COPYPR(Tmp);
	DELETE_PR(Tmp);
	*(*Hptr + (n - 1)) = COPYPR(Pptr[n - 1]);
	TempR = ONER();
	L = CONSTANTPR(TempR);
	FREEMPR(TempR);
	for (i = n - 1; i >= 1; i--)
	{
		Tmp = EUCLIDPR(*(*Hptr + i), *(*Hptr + (i - 1)), &H, &K);
		DELETE_PR(Tmp);
		DELETE_PR(*(*Hptr + i));
		*(*Hptr + i) = MULTPR(H, L);
		DELETE_PR(H);
		if (i == 1)
			break;
		Tmp = L;
		L = MULTPR(Tmp, K);
		DELETE_PR(K);
		DELETE_PR(Tmp);
		DELETE_PR(*(*Hptr + (i - 1)));
		*(*Hptr + (i - 1)) = COPYPR(Pptr[i - 1]);
	}
	DELETE_PR(**Hptr);
	**Hptr  = MULTPR(L, K);
	DELETE_PR(L);
	DELETE_PR(K);
	return (B);
}

MPMATR *VALMATR(POLYR Pptr, MPMATR *Mptr)
/*
 * Evaluates the polynomial Pptr at the square matrix *Mptr.
 */
{
	unsigned int i, n;
	int k;
	MPMATR *Aptr, *Temp1, *Temp, *ID;
	POLYR CURSOR;
	MPR *TempR;

	n = Mptr->R;
	if (Pptr == NULL)
		return ZEROMNR(n, n);
	else
	{
		ID = IDENTITYR(n);
		Aptr = SCALARMATR(ID, Pptr->COEF);
		FREEMATR(ID);
	}
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		if (CURSOR->NEXT == NULL)
			k = DEGREEPR(CURSOR);
		else
			k = DEGREEPR(CURSOR) - DEGREEPR(CURSOR->NEXT);
		Temp1 = POWERMATR(Mptr, (unsigned int)k);
		Temp = Aptr;
		Aptr = MULTMATR(Temp1, Temp);
		FREEMATR(Temp);
		FREEMATR(Temp1);
		CURSOR = CURSOR->NEXT;
		if (CURSOR != NULL)
		{
			for (i = 0; i < n; i++)
			{
				TempR = elt(Aptr, i, i);
				elt(Aptr, i, i) = ADDR(elt(Aptr, i, i), CURSOR->COEF);
				FREEMPR(TempR);
			}
		}
	}
	return Aptr;
}
