/* p3m.c */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATPm *MULTMATPm(MPMATPm *Mptr, MPMATPm *Nptr, USI m)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr) (mod m).
 */
{
	POLYm X, Y, Tmp;
	MPMATPm *Qptr;
	unsigned int i, j, k;

	Qptr = BUILDMATPm(Mptr->R, Nptr->C);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
		{
			X = NULL;
			for (k = 0; k <= Mptr->C - 1; k++)
			{
				Y = MULTPm(Mptr->V[i][k], Nptr->V[k][j], m);
				Tmp = X;
				X = ADDPm(Tmp, Y, m);
				DELETE_Pm(Tmp);
				DELETE_Pm(Y);
			}
			Qptr->V[i][j] = X;
		}
	}
	return (Qptr);
}

POLYm TRACEPm(MPMATPm *Mptr, USI m)
/*
 * returns the trace of *Mptr (mod m).
 */
{
	POLYm A, Tmp;
	unsigned int i;

	A = NULL;
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Tmp = A;
		A = ADDPm(Tmp, Mptr->V[i][i], m);
		DELETE_Pm(Tmp);
	}
	return (A);
}

MPMATPm *MINUSMATPm(MPMATPm *Mptr, USI m)
/*
 * returns -(*Mptr) (mod m). 
 */
{
	unsigned int i, j;
	MPMATPm *Nptr;

	Nptr = BUILDMATPm(Mptr->R, Mptr->C);
	for (i = 0; i <= Nptr->R - 1; i++)
	{
		for (j = 0; j <= Nptr->C - 1; j++)
			Nptr->V[i][j] = SCALARPm(m - 1, Mptr->V[i][j], m);
	}
	return (Nptr);
}

MPMATPm *DIAGPm(MPMATPm *Mptr, POLYm Aptr, USI m)
/*
 * returns *Mptr - Aptr * (Unit Matrix), (mod m). 
 */
{
	unsigned int i;
	POLYm Tmp;
	MPMATPm *Nptr;

	Nptr = COPYMATPm(Mptr, m);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Tmp = Nptr->V[i][i];
		Nptr->V[i][i] = SUBPm(Tmp, Aptr, m);
		DELETE_Pm(Tmp);
	}
	return (Nptr);
}

POLYm LEVERRIER_Pm(MPMATPm *Mptr, MPMATPm **Nptr, USI m)
/*
 * The Faddeeva/Leverrier method for finding the
 * adjoint and determinant,  applied to a polynomial matrix *Mptr (mod m).
 * returns the determinant while *Nptr is the adjoint.
 * See "Computational methods in linear algebra", by V. N. Faddeeva, Dover
 * 1959.
 */
{
	unsigned int i, h, n, j;
	MPMATPm *Qptr, *M;
	MPMATPI *Temp1PI, *Temp2PI;
	POLYm P, D, Tmp, DETPm;
	POLYI DETPI;

	D = NULL;
	n = Mptr->R;
	if (m <= n)
	{
		Temp1PI = BUILDMATPI(n, n);
		for (i = 0; i <= n - 1; i++)
		{
			for (j = 0; j <= n - 1; j++)
				Temp1PI->V[i][j] = CHANGETOPI(Mptr->V[i][j]);
		}
		DETPI = LEVERRIER_PI(Temp1PI, &Temp2PI);
		FREEMATPI(Temp1PI);
		*Nptr = BUILDMATPm(n, n);
		for (i = 0; i <= n - 1; i++)
		{
			for (j = 0; j <= n - 1; j++)
				(*Nptr)->V[i][j] = CHANGETOPm(Temp2PI->V[i][j], m);	
		}
		FREEMATPI(Temp2PI);
		DETPm = CHANGETOPm(DETPI, m);
		DELETE_PI(DETPI);
		return (DETPm);
	}
	Qptr = COPYMATPm(Mptr, m);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		P = TRACEPm(Qptr, m);
		h =  INVERSEm(i + 1, m);
		Tmp = P;
		P = SCALARPm(h, Tmp, m);
		DELETE_Pm(Tmp);
		M = DIAGPm(Qptr, P, m);
		FREEMATPm(Qptr);
		if (i == Mptr->R - 1)
			D = COPYPm(P, m);
		DELETE_Pm(P);
		if (i == Mptr->R - 2)
		{
			if (i % 2 == 0)
				*Nptr = MINUSMATPm(M, m);
			else
				*Nptr = COPYMATPm(M, m);
		}  
		Qptr =	MULTMATPm(Mptr, M, m);
		FREEMATPm(M);
	}
	FREEMATPm(Qptr);
	if (i % 2 == 0)
	{
		Tmp = D;
		D = SCALARPm(m - 1, Tmp, m);
		DELETE_Pm(Tmp);
	}
	return (D);
}

MPMATm *VALMATm(POLYm Pptr, MPMATm *Mptr, USI m)
/*
 * Evaluates the polynomial Pptr at the square matrix *Mptr (mod m).
 */
{
	unsigned int i, n;
	int k;
	MPMATm *Aptr, *Temp1, *Temp, *ID;
	POLYm CURSOR;

	n = Mptr->R;
	if (Pptr == NULL)
		return ZEROMNm(n, n);
	else
	{
		ID = IDENTITYm(n);
		Aptr = SCALARMATm(ID, Pptr->COEF, m);
		FREEMATm(ID);
	}
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		if (CURSOR->NEXT == NULL)
			k = DEGREEPm(CURSOR);
		else
			k = DEGREEPm(CURSOR) - DEGREEPm(CURSOR->NEXT);
		Temp1 = POWERMATm(Mptr, (unsigned int)k, m);
		Temp = Aptr;
		Aptr = MULTMATm(Temp1, Temp, m);
		FREEMATm(Temp);
		FREEMATm(Temp1);
		CURSOR = CURSOR->NEXT;
		if (CURSOR != NULL)
		{
			for (i = 0; i < n; i++)
				Aptr->V[i][i] = ADDm(CURSOR->COEF, Aptr->V[i][i], m);
		}
	}
	return Aptr;
}
