/* p4CR.c */

/* matrix operations using POLYCR's */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATPCR *DELETE_ROWPCR(USI r, MPMATPCR *Mptr)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPCR *Nptr;

	r--;
	Nptr = BUILDMATPCR(Mptr->R - 1, Mptr->C);
	for (i = 0; i < Nptr->R; i++)
	{
		if (i < r)
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPCR(Mptr->V[i][j]);	
		}
		else
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPCR(Mptr->V[i + 1][j]);	
		}
	}
	return (Nptr);
}

MPMATPCR *DELETE_COLPCR(USI c, MPMATPCR *Mptr)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPCR *Nptr;

	c--;
	Nptr = BUILDMATPCR(Mptr->R, Mptr->C - 1);
	for (j = 0; j < Nptr->C; j++)
	{
		if (j < c)
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPCR(Mptr->V[i][j]);	
		}
		else
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPCR(Mptr->V[i][j + 1]);	
		}
	}
	return (Nptr);
}

MPMATCR *COMPANIONCR(POLYCR Pptr)
/*
 * returns the companion matrix corresponding to the polynomial Pptr.
 */
{
	unsigned int j, n;
	MPMATCR *Mptr;
	int i;
	POLYCR CURSOR;

	n = DEGREEPCR(Pptr);
	CURSOR = Pptr;
	Mptr = BUILDMATCR(n, n);
	CURSOR = CURSOR->NEXT;
	for (i = n - 1; i >= 0; i--)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( j == n - 1)
			{
				if (CURSOR != NULL)
				{
					if (CURSOR->DEG == i)
					{
						elt(Mptr, i, j) = MINUSCR(CURSOR->COEF);
						CURSOR = CURSOR->NEXT;
					}
					else
						elt(Mptr, i, j) = ZEROCR();
				}
				else
					elt(Mptr, i, j) = ZEROCR();
			}
			else if (i == j + 1)
				elt(Mptr, i, j) = ONECR();
			else
				elt(Mptr, i, j) = ZEROCR();
		}
	}
	return (Mptr);
}

POLYCR LEVERRIER_CR(MPMATCR *Mptr, MPMATCR **Nptr, MPCR **Dptr)
/*
 * The Faddeeva/Leverrier method for finding the characteristic polynomial,
 * adjoint and determinant of a matrix of complex rationals.
 * See "Computational methods in linear algebra", by V. N. Faddeeva, Dover 1959
 */
{
	POLYCR Pptr;
	POLYCI CURSOR, PCI;
	MPMATCI *M, *MCI;
	MPMATI *M1, *M2;
	MPMATCR *N;
	MPI *A, *A1, *A2, *T, *TempI;
	MPCI *DCI;
	MPCR *B, *TempCR;
	unsigned int i, j, n;

	n = Mptr->R;
	M1 = BUILDMATI(n,n);
	M2 = BUILDMATI(n,n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			elt(M1, i, j) = COPYI((elt(Mptr, i, j)->R)->D);
			elt(M2, i, j) = COPYI((elt(Mptr, i, j)->I)->D);
		}
	}
	A1 = LCM_MATRIX(M1);
	A2 = LCM_MATRIX(M2);
	FREEMATI(M1);
	FREEMATI(M2);
	A = LCM(A1, A2);
	FREEMPI(A1);
	FREEMPI(A2);
	B = BUILDMPCR();
	B->R = BUILDMPR();
	B->R->N = COPYI(A);
	B->R->D = ONEI();
	B->I = ZEROR();
	N = SCALARMATCR(Mptr, B);/* N->V[i][j] is really a complex integer. */
	FREEMPCR(B);
	M = BUILDMATCI(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			elt(M, i, j) = BUILDMPCI();
			elt(M, i, j)->R = COPYI(elt(N, i, j)->R->N);
			elt(M, i, j)->I = COPYI(elt(N, i, j)->I->N);
		}
	}
	FREEMATCR(N);
	PCI = LEVERRIER_CI(M, &MCI, &DCI);
	FREEMATCI(M);
	T = ONEI();
	Pptr = NULL;
	CURSOR = PCI->NEXT;
	while (CURSOR != NULL)
	{
		TempI = T;
		T = POWERI(A, (unsigned)(n - CURSOR->DEG));
		FREEMPI(TempI);
		TempCR = BUILDMPCR();
		TempCR->R = RATIOI((CURSOR->COEF)->R, T);
		TempCR->I = RATIOI((CURSOR->COEF)->I, T);
		PINSERTPCR(CURSOR->DEG, TempCR, &Pptr, 0);
		FREEMPCR(TempCR);
		CURSOR = CURSOR->NEXT;
	}
	TempCR = ONECR();
	PINSERTPCR((int)n, TempCR, &Pptr, 0);
	FREEMPCR(TempCR);
	DELETE_PCI(PCI);
	*Dptr = BUILDMPCR();
	(*Dptr)->R = RATIOI(DCI->R, T);
	(*Dptr)->I = RATIOI(DCI->I, T);
	TempI = T;
	T = INT0(T, A);
	FREEMPI(TempI);
	*Nptr = BUILDMATCR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			elt((*Nptr), i, j) = BUILDMPCR();
			elt((*Nptr), i, j)->R = RATIOI(elt(MCI, i, j)->R, T);
			elt((*Nptr), i, j)->I = RATIOI(elt(MCI, i, j)->I, T);
		}
	}
	FREEMATCI(MCI);
	FREEMPI(T);
	FREEMPI(A);
	FREEMPCI(DCI);
	return (Pptr);
}

MPMATCR *INVERSEMATCR(MPMATCR *Mptr)
/*
 * returns the inverse of *Mptr, if it exists, NULL otherwise.
 */
{
	MPMATCR *Nptr, *adj;
	MPCR *det;
	unsigned int i, j, n;
	POLYCR A;

	A = LEVERRIER_CR(Mptr, &adj, &det);
	DELETE_PCR(A);
	if (EQZEROCR(det))
	{
		FREEMPCR(det);
		FREEMATCR(adj);
		return(NULL);
	}
	n = Mptr->R;
	Nptr = BUILDMATCR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		elt(Nptr, i, j) = RATIOCR(elt(adj, i, j), det);
	}
	FREEMPCR(det);
	FREEMATCR(adj);
	return (Nptr);
}

POLYCR MINIMUM_POLYCR(MPMATCR *Mptr)
/*
 * returns the minimum polynomial of the square matrix *Mptr.
 */
{
	MPMATCR *A, *B, *C, *Temp, *Temp1;
	POLYCR Pptr;
	unsigned int i, j, n;
	MPCR *TempCR;

	n = Mptr->R;
	A = IDENTITYCR(n);
	B = UNWRAPMATCR(A);
	for (i = 1; i <= n; i++)
	{
		Temp = A;
		A = MULTMATCR(Temp, Mptr);
		FREEMATCR(Temp);
		Temp = B;
		Temp1 = UNWRAPMATCR(A);
		B = JCOLSMATCR(Temp, Temp1);
		FREEMATCR(Temp);
		FREEMATCR(Temp1);
		C = EQUATIONS1CR(B);
		if (C != NULL)
		{
			Pptr = NULL;
			for (j = 0; j <= i - 1; j++)
			{
				TempCR = MINUSCR(elt(C, j, 0));
				PINSERTPCR((int)j, TempCR, &Pptr, 0);
				FREEMPCR(TempCR);
			}
			TempCR = ONECR();
			PINSERTPCR((int)i, TempCR, &Pptr, 0);
			FREEMPCR(TempCR);
			PURGEPCR(&Pptr);
			FREEMATCR(C);
			break;
		}
	}
	FREEMATCR(A);
	FREEMATCR(B);
	return Pptr;
}

MPMATPCR *SUBMATROWSPCR(MPMATPCR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPCR *Nptr;

	Nptr = BUILDMATPCR(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			Nptr->V[i][j] = COPYPCR(Mptr->V[Cptr[i]][j]);
	return Nptr;
}

MPMATPCR *SUBMATCOLUMNSPCR(MPMATPCR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPCR *Nptr;

	Nptr = BUILDMATPCR(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			Nptr->V[i][j] = COPYPCR(Mptr->V[i][Cptr[j]]);
	return Nptr;
}
