/* p4R.c */

/* matrix operations using POLYR's */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

MPMATPR *DELETE_ROWPR(USI r, MPMATPR *Mptr)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPR *Nptr;

	r--;
	Nptr = BUILDMATPR(Mptr->R - 1, Mptr->C);
	for (i = 0; i < Nptr->R; i++)
	{
		if (i < r)
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPR(Mptr->V[i][j]);	
		}
		else
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPR(Mptr->V[i + 1][j]);	
		}
	}
	return (Nptr);
}

MPMATPR *DELETE_COLPR(USI c, MPMATPR *Mptr)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPR *Nptr;

	c--;
	Nptr = BUILDMATPR(Mptr->R, Mptr->C - 1);
	for (j = 0; j < Nptr->C; j++)
	{
		if (j < c)
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPR(Mptr->V[i][j]);	
		}
		else
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPR(Mptr->V[i][j + 1]);	
		}
	}
	return (Nptr);
}

MPMATR *COMPANIONR(POLYR Pptr)
/*
 * returns the companion matrix corresponding to the polynomial Pptr.
 */
{
	unsigned int j, n;
	MPMATR *Mptr;
	int i;
	POLYR CURSOR;

	n = DEGREEPR(Pptr);
	CURSOR = Pptr;
	Mptr = BUILDMATR(n, n);
	CURSOR = CURSOR->NEXT;
	for (i = n - 1; i >= 0; i--)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( j == n - 1)
			{
				if (CURSOR != NULL)
				{
					if (CURSOR->DEG == i)
					{
						elt(Mptr, i, j) = MINUSR(CURSOR->COEF);
						CURSOR = CURSOR->NEXT;
					}
					else
						elt(Mptr, i, j) = ZEROR();
				}
				else
					elt(Mptr, i, j) = ZEROR();
			}
			else if (i == j + 1)
				elt(Mptr, i, j) = ONER();
			else
				elt(Mptr, i, j) = ZEROR();
		}
	}
	return (Mptr);
}

POLYR LEVERRIER_R(MPMATR *Mptr, MPMATR **Nptr, MPR **Dptr)
/*
 * The characteristic polynomial is returned, along with the adjoint and
 * determinant of a matrix, using the Faddeeva/Leverrier method. See
 * "Computational methods in linear algebra", by V. N. Faddeeva, Dover 1959.
 */
{
	POLYR Pptr;
	POLYI CURSOR, PI;
	MPMATI *M, *MI;
	MPMATR *N;
	MPI *A, *DI, *T, *TempI;
	MPR *B, *TempR;
	unsigned int i, j, n;

	n = Mptr->R;
	M = BUILDMATI(n,n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			elt(M, i, j) = COPYI(elt(Mptr, i, j)->D);
	}
	A = LCM_MATRIX(M);
	B = BUILDMPR();
	B->N = COPYI(A);
	B->D = ONEI();
	N = SCALARMATR(Mptr, B);
	FREEMPR(B);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			TempI = elt(M, i, j);
			elt(M, i, j) = COPYI(elt(N, i, j)->N);
			FREEMPI(TempI);
		}
	}
	PI = LEVERRIER_I(M, &MI, &DI);
	T = ONEI();
	Pptr = NULL;
	CURSOR = PI->NEXT;
	while (CURSOR != NULL)
	{
		TempI = T;
		T = POWERI(A, (unsigned)(n - CURSOR->DEG));
		FREEMPI(TempI);
		TempR = RATIOI(CURSOR->COEF, T);
		PINSERTPR(CURSOR->DEG, TempR, &Pptr, 0);
		FREEMPR(TempR);
		CURSOR = CURSOR->NEXT;
	}
	TempR = ONER();
	PINSERTPR((int)n, TempR, &Pptr, 0);
	FREEMPR(TempR);
	DELETE_PI(PI);
	*Dptr = RATIOI(DI, T);
	FREEMPI(DI);
	TempI = T;
	T = INT0(T, A);
	FREEMPI(TempI);
	*Nptr = BUILDMATR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= Mptr->C - 1; j++)
			elt((*Nptr), i, j) = RATIOI(elt(MI, i, j), T);
	}
	FREEMPI(T);
	FREEMPI(A);
	FREEMATI(M);
	FREEMATI(MI);
	FREEMATR(N);
	return (Pptr);
}

MPMATR *CROSS_PRODR(MPMATR *Mptr)
/*
 * Returns the cross-product of n - 1 row vectors of dimension n.
 */
{
	MPMATR *M, *Tmp1, *Tmp2;
	POLYR P;
	unsigned int n, j;
	MPR *TempR;

	n = Mptr->C;
	M = BUILDMATR(1, n);
	for (j = 0; j <= n - 1; j++)
	{
		Tmp1 = DELETE_COLR(j + 1, Mptr);
		P = LEVERRIER_R(Tmp1, &Tmp2, &elt(M, 0, j));
		if (j % 2)
		{
			TempR = elt(M, 0, j);
			elt(M, 0, j) = MINUSR(elt(M, 0, j));
			FREEMPR(TempR);
		}
		FREEMATR(Tmp1);
		FREEMATR(Tmp2);
		DELETE_PR(P);
	}
	return (M);
}

MPMATR *INVERSEMATR(MPMATR *Mptr)
/*
 * returns the inverse of *Mptr, if it exists, NULL otherwise.
 */
{
	MPMATR *Nptr, *adj;
	MPR *det;
	unsigned int i, j, n;
	POLYR A;

	A = LEVERRIER_R(Mptr, &adj, &det);
	DELETE_PR(A);
	if (EQZEROR(det))
	{
		printf("det = ");PRINTR(det);printf("\n");
		FREEMPR(det);
		FREEMATR(adj);
		return(NULL);
	}
	n = Mptr->R;
	Nptr = BUILDMATR(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
			elt(Nptr, i, j) = RATIOR(elt(adj, i, j), det);
	}
	FREEMPR(det);
	FREEMATR(adj);
	return (Nptr);
}

POLYR MINIMUM_POLYR(MPMATR *Mptr)
/*
 * returns the minimum polynomial of the square matrix *Mptr.
 */
{
	MPMATR *A, *B, *C, *Temp, *Temp1;
	POLYR Pptr;
	unsigned int i, j, n;
	MPR *TempR;

	n = Mptr->R;
	A = IDENTITYR(n);
	B = UNWRAPMATR(A);
	for (i = 1; i <= n; i++)
	{
		Temp = A;
		A = MULTMATR(Temp, Mptr);
		FREEMATR(Temp);
		Temp = B;
		Temp1 = UNWRAPMATR(A);
		B = JCOLSMATR(Temp, Temp1);
		FREEMATR(Temp);
		FREEMATR(Temp1);
		C = EQUATIONS1R(B);
		if (C != NULL)
		{
			Pptr = NULL;
			for (j = 0; j <= i - 1; j++)
			{
				TempR = MINUSR(elt(C, j, 0));
				PINSERTPR((int)j, TempR, &Pptr, 0);
				FREEMPR(TempR);
			}
			TempR = ONER();
			PINSERTPR((int)i, TempR, &Pptr, 0);
			FREEMPR(TempR);
			PURGEPR(&Pptr);
			FREEMATR(C);
			break;
		}
	}
	FREEMATR(A);
	FREEMATR(B);
	return Pptr;
}

MPMATPR *SUBMATROWSPR(MPMATPR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPR *Nptr;

	Nptr = BUILDMATPR(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			Nptr->V[i][j] = COPYPR(Mptr->V[Cptr[i]][j]);
	return Nptr;
}

MPMATPR *SUBMATCOLUMNSPR(MPMATPR *Mptr, USI *Cptr, USI n)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPR *Nptr;

	Nptr = BUILDMATPR(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			Nptr->V[i][j] = COPYPR(Mptr->V[i][Cptr[j]]);
	return Nptr;
}
