/* p4m.c */

/* matrix operations using POLYm's */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

void DUMPPOLYm(POLYm Pptr)
{
	POLYm CURSOR;

	CURSOR = Pptr;
	printf("Pptr ought to be ");
	PRINTPm(Pptr);
	printf("\n");
	if (Pptr == NULL)
	{
		printf("Pptr = NULL");
		return;
	}
	while (CURSOR != NULL)
	{
		printf("However---CURSOR->DEG = %d\n", CURSOR->DEG);
		printf("----------CURSOR->COEF = %u\n", CURSOR->COEF);
		CURSOR = CURSOR->NEXT;
	}
}

MPMATPm *DELETE_ROWPm(USI r, MPMATPm *Mptr, USI m)
/*
 * deletes row r of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPm *Nptr;

	r--;
	Nptr = BUILDMATPm(Mptr->R - 1, Mptr->C);
	for (i = 0; i < Nptr->R; i++)
	{
		if (i < r)
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPm(Mptr->V[i][j], m);	
		}
		else
		{
			for (j = 0; j < Nptr->C; j++)
				Nptr->V[i][j] = COPYPm(Mptr->V[i + 1][j], m);
		}
	}
	return (Nptr);
}

MPMATPm *DELETE_COLPm(USI c, MPMATPm *Mptr, USI m)
/*
 * deletes  column c of *Mptr.
 */
{
	unsigned int i, j;
	MPMATPm *Nptr;

	c--;
	Nptr = BUILDMATPm(Mptr->R, Mptr->C - 1);
	for (j = 0; j < Nptr->C; j++)
	{
		if (j < c)
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPm(Mptr->V[i][j], m);	
		}
		else
		{
			for (i = 0; i < Nptr->R; i++)
				Nptr->V[i][j] = COPYPm(Mptr->V[i][j + 1], m);	
		}
	}
	return (Nptr);
}

MPMATm *COMPANIONm(POLYm Pptr, USI m)
/*
 * returns the companion matrix corresponding to the polynomial Pptr.
 */
{
	unsigned int j, n;
	MPMATm *Mptr;
	int i;
	POLYm CURSOR;

	n = Pptr->DEG;
	CURSOR = Pptr;
	Mptr = BUILDMATm(n, n);
	CURSOR = CURSOR->NEXT;
	for (i = n - 1; i >= 0; i--)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( j == n - 1)
			{
				if (CURSOR != NULL)
				{
					if (CURSOR->DEG == i)
					{
						Mptr->V[i][j] = MINUSm(CURSOR->COEF, m);
						CURSOR = CURSOR->NEXT;
					}
					else
						Mptr->V[i][j] = 0;
				}
				else
					Mptr->V[i][j] = 0;
			}
			else if (i == j + 1)
				Mptr->V[i][j] = 1;
			else
				Mptr->V[i][j] = 0;
		}
	}
	return (Mptr);
}

POLYm MINIMUM_POLYm(MPMATm *Mptr, USI m)
/*
 * returns the minimum polynomial of the square matrix *Mptr (mod m).
 */
{
	MPMATm *A, *B, *C, *Temp, *Temp1;
	POLYm Pptr;
	unsigned int i, j, n, COEF;

	n = Mptr->R;
	A = IDENTITYm(n);
	B = UNWRAPMATm(A);
	for (i = 1; i <= n; i++)
	{
		Temp = A;
		A = MULTMATm(Temp, Mptr, m);
		FREEMATm(Temp);
		Temp = B;
		Temp1 = UNWRAPMATm(A);
		B = JCOLSMATm(Temp, Temp1);
		FREEMATm(Temp);
		FREEMATm(Temp1);
		C = EQUATIONS1m(B, m);
		if (C != NULL)
		{
			Pptr = NULL;
			for (j = 0; j <= i - 1; j++)
			{
				COEF = MINUSm(C->V[j][0], m);
				PINSERTPm((int)j, COEF, &Pptr, 0, m);
			}
			PINSERTPm((int)i, 1, &Pptr, 0, m);
			PURGEPm(&Pptr);
			FREEMATm(C);
			break;
		}
	}
	FREEMATm(A);
	FREEMATm(B);
	return Pptr;
}

MPMATPm *SUBMATROWSPm(MPMATPm *Mptr, USI *Cptr, USI n, USI m)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from rows Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPm *Nptr;

	Nptr = BUILDMATPm(n, Mptr->C);
	for (i = 0; i < n; i++)
		for (j = 0; j < Mptr->C; j++)
			Nptr->V[i][j] = COPYPm(Mptr->V[Cptr[i]][j], m);
	return Nptr;
}

MPMATPm *SUBMATCOLUMNSPm(MPMATPm *Mptr, USI *Cptr, USI n, USI m)
/*
 * Cptr is an array of n (distinct) non-negative integers, n < Mptr->R.
 * Output: the submatrix formed from columns Cptr[0],...,Cptr[n-1].
 */
{
	unsigned int i, j;
	MPMATPm *Nptr;

	Nptr = BUILDMATPm(Mptr->R, n);
	for (j = 0; j < n; j++)
		for (i = 0; i < Mptr->R; i++)
			Nptr->V[i][j] = COPYPm(Mptr->V[i][Cptr[j]], m);
	return Nptr;
}
