/* pCI.c */

#include <stdio.h>
#include <stdlib.h>
#include "integer.h"
#include "fun.h"

int DEGREEPCI(POLYCI Pptr)
/*
 * returns the degree of the polynomial Pptr.
 */
{
	if (Pptr == NULL)
		return (-1);
	else
		return (Pptr->DEG);
	
}

MPCI *LEADPCI(POLYCI Pptr)
/*
 * Aptr is the  leading coefficient of the polynomial Pptr.
 */
{
	if (Pptr == NULL)
		return ZEROCI();
	else
		return COPYCI(Pptr->COEF);
}

void DELETE_PCI(POLYCI Pptr)
/*
 * returns back to the system storage dynamically allocated to the linked list
 * which Pptr points to.
 */
{
	POLYCI Qtmp, orig;
	
	if(!Pptr) return;
	orig = Pptr;
	while (Pptr != NULL)
	{
 		Qtmp = Pptr->NEXT;
		FREEMPCI(Pptr->COEF);
#ifdef DEBUG
		SAVEPCI(Pptr);
#endif
		Pptr= Qtmp;
	}
#ifndef DEBUG
	SAVEPCI(orig);
#endif
	return;
}

void PINSERTPCI(int DEG, MPCI *COEF, POLYCI *Pptr, USI OPT)
/*
 * inserts a copy of a single term *Qptr into polynomial *Pptr.
 * It has two options that control the insertion in case a term of *Pptr already
 * has the degree of *Qptr, If OPT = 0, then the term is replaced; if OPT = 1,
 * then the coefficient of *Qptr is added to that of the corresponding term of
 *  *Pptr.
 */
{
	POLYCI CURSOR, Ttmp;
	MPCI *TempCI;

	if (*Pptr == NULL)
	{
		*Pptr = CREATECI();
		(*Pptr)->DEG = DEG;
		(*Pptr)->COEF = COPYCI(COEF);
		(*Pptr)->NEXT = NULL;
		return;
	}
	else
	{
		CURSOR = *Pptr;
		while ((CURSOR->DEG > DEG) && (CURSOR->NEXT != NULL))
			CURSOR = CURSOR->NEXT;
		if (CURSOR->DEG < DEG)
		{
			/* Qptr inserted before CURSOR */
			Ttmp = CREATECI();
			Ttmp->DEG = CURSOR->DEG;
			Ttmp->COEF = CURSOR->COEF;
			Ttmp->NEXT = CURSOR->NEXT;
			CURSOR->DEG = DEG;
			CURSOR->COEF = COPYCI(COEF);
			CURSOR->NEXT = Ttmp;
			return;
		}
		else if (CURSOR->DEG == DEG)
		{
			if (OPT == 0)
				CURSOR->COEF = COPYCI(COEF);
			else
			{
				TempCI = CURSOR->COEF;
				CURSOR->COEF = ADDCI(CURSOR->COEF, COEF);
				FREEMPCI(TempCI);
			}
			return;
		}
		else
		{
			/* CURSOR->DEG > Qptr->DEG and CURSOR->NEXT === NULL */
			/* Qptr inserted at end */
			Ttmp = CREATECI();
			Ttmp->DEG = DEG;
			Ttmp->COEF = COPYCI(COEF);
			Ttmp->NEXT = NULL;
			CURSOR->NEXT = Ttmp;
			return;
		}
	}
}

void PURGEPCI(POLYCI *Pptr)
/*
 * gets rid of any zero coefficients in the polynomial Pptr.
 */
{
	POLYCI R1tmp, R2tmp;

	if (*Pptr != NULL)
	{
		R1tmp = *Pptr;
		R2tmp = (*Pptr)->NEXT;
		while (R2tmp != NULL)
		{
			/* purge non-leading terms */
			if (EQZEROCI(R2tmp->COEF))
			{
				R1tmp->NEXT = R2tmp->NEXT;
				R2tmp->NEXT = NULL;
				DELETE_PCI(R2tmp);
				R2tmp = R1tmp->NEXT;
			}
			else
			{
				R1tmp = R2tmp;
				R2tmp = R2tmp->NEXT;
			}
		}
		if (EQZEROCI((*Pptr)->COEF))
		{
			/* purge leading term */
			R1tmp = *Pptr;
			*Pptr = (*Pptr)->NEXT;
			R1tmp->NEXT = NULL;
			DELETE_PCI(R1tmp);
		}
	}
	return;
}

POLYCI INPUTPCI()
/*
 * Inputting a polynomial Pptr from the terminal.
 */
{
	POLYCI Pptr;
	int k;
	char *line, *p;
	MPCI *TempCI;
	
	Pptr = NULL;
	do
	{
		printf("Enter the degree of new term, (-1 to terminate): ");
		if (scanfd10(&k) == 0)
			k = -1;
		if (k >= 0)
		{
			printf("Enter the coefficient of x^%d: ", k);
			p = line = Gets();
		        TempCI = INPUTCI(&p); 
			PINSERTPCI(k, TempCI, &Pptr, 0);
			FREEMPCI(TempCI);
			free(line);
		}
	}
	while (k >= 0);
	PURGEPCI(&Pptr);
	return Pptr;
}

POLYCI COPYPCI(POLYCI Pptr)
/*
 * makes a copy of the polynomial Pptr.
 */
{
	POLYCI Rtmp, Sptr;

	Sptr = NULL;
	Rtmp = Pptr;
	while (Rtmp != NULL)
	{
		PINSERTPCI(Rtmp->DEG, Rtmp->COEF, &Sptr, 0);
		Rtmp = Rtmp->NEXT;
	}
	return (Sptr);
}
 
void PRINTPCI(POLYCI Pptr)
/*
 * outputs the terms of the polynomial Pptr.
 */
{
	POLYCI CURSOR;
	int k, l;

	if (Pptr == NULL)
	{
		printf("0");
		return;
	}
	else
	{
		CURSOR = Pptr;
		k = DEGREEPCI(Pptr);
		while (CURSOR != NULL)
		{
			if (((CURSOR->COEF)->R)->S && ((CURSOR->COEF)->I)->S)
			{
				if (CURSOR->DEG != k)
					printf("+ ");
				printf("(");
				PRINTCI(CURSOR->COEF);
				printf(")");
			}
			if (((CURSOR->COEF)->R)->S && ((CURSOR->COEF)->I)->S == 0)
			{
				l = ((CURSOR->COEF)->R)->S;
				if (l == -1)
				{
					((CURSOR->COEF)->R)->S = 1;
					printf("- ");
				}
				else
				{
					if (CURSOR->DEG != k)
						printf("+ ");
			}
			if (CURSOR->DEG == 0 ||(!EQONEI((CURSOR->COEF)->R)))
				PRINTI((CURSOR->COEF)->R);
			if (l == -1)
				((CURSOR->COEF)->R)->S = -1;
			}
			if (((CURSOR->COEF)->R)->S == 0 && ((CURSOR->COEF)->I)->S)
			{
				l = ((CURSOR->COEF)->I)->S;
				if (l == -1)
				{
					((CURSOR->COEF)->I)->S = 1;
					printf("- ");
				}
				else
				{
					if (CURSOR->DEG != k)
						printf("+ ");
				}
				if (CURSOR->DEG == 0 || !EQONEI((CURSOR->COEF)->I))
					PRINTCI(CURSOR->COEF);
				if (EQONEI((CURSOR->COEF)->I) && CURSOR->DEG)
					printf("i");
				if (l == -1)
					((CURSOR->COEF)->I)->S = -1;
			}
			if (CURSOR->DEG == 1)
				printf("x");
			if (CURSOR->DEG > 1)
				printf("x^%d", CURSOR->DEG);
			if (CURSOR->NEXT != NULL)
				printf(" ");
			CURSOR = CURSOR->NEXT;
		}
		return;
	}
}

POLYCI SCALARPCI(MPCI *Cptr, POLYCI Pptr)
/*
 * multiplying the polynomial Pptr by the MPCI *Cptr.
 */
{
	POLYCI Sptr, Rtmp;
	MPCI *TempCI;

	if (Pptr == NULL || ((Cptr->R)->S == 0 && (Cptr->I)->S == 0))
		Sptr = NULL;
	else
	{
		Sptr = COPYPCI(Pptr);
		Rtmp = Sptr;
		while (Rtmp != NULL)
		{
			TempCI = Rtmp->COEF;
			Rtmp->COEF = MULTCI(Rtmp->COEF, Cptr);
			FREEMPCI(TempCI);
			Rtmp = Rtmp->NEXT;
		}
	}
	return (Sptr);
}

POLYCI ADDPCI(POLYCI Pptr, POLYCI Qptr)
/*
 * returns the sum of two polynomials.
 */
{
	POLYCI CURSOR, Sptr;

	Sptr = COPYPCI(Pptr);
	CURSOR = Qptr;
	while (CURSOR != NULL)
	{
		PINSERTPCI(CURSOR->DEG, CURSOR->COEF, &Sptr, 1);
		CURSOR = CURSOR->NEXT;
	}
	PURGEPCI(&Sptr);
	return (Sptr);
}

POLYCI SUBPCI(POLYCI Pptr, POLYCI Qptr)
/*
 * returns the difference Pptr - Qptr of two polynomials.
 */
{
	POLYCI CURSOR, Sptr;
	MPCI *TempCI;

		Sptr = COPYPCI(Pptr);
		if (Qptr == NULL)
			return (Sptr);
		CURSOR = Qptr;
		while (CURSOR != NULL)
		{
			TempCI = MINUSCI(CURSOR->COEF);
			PINSERTPCI(CURSOR->DEG, TempCI, &Sptr, 1);
			FREEMPCI(TempCI);
			CURSOR = CURSOR->NEXT;
		}
		PURGEPCI(&Sptr);
		return (Sptr);
}

POLYCI MULTPCI(POLYCI Pptr, POLYCI Qptr)
/*
 * returns the product of two polynomials.
 */
{
	POLYCI Sptr, CURSOR_P, CURSOR_Q;
	int k;
	MPCI *TempCI;

	if (Pptr == NULL || Qptr == NULL)
		return (NULL);
	Sptr = NULL;
	CURSOR_P = Pptr;
	while (CURSOR_P != NULL)
	{
		CURSOR_Q = Qptr;
		while (CURSOR_Q != NULL)
		{
			k = CURSOR_P->DEG + CURSOR_Q->DEG;
			TempCI = MULTCI(CURSOR_P->COEF, CURSOR_Q->COEF);
			PINSERTPCI(k, TempCI, &Sptr, 1);
	printf("Sptr = ");PRINTPCI(Sptr);printf("\n");
			FREEMPCI(TempCI);
			CURSOR_Q = CURSOR_Q->NEXT;
		}
		CURSOR_P = CURSOR_P->NEXT;
	}
	PURGEPCI(&Sptr);
	return (Sptr);
}

MPCI *VALPCI(POLYCI Pptr, MPCI *Cptr)
/*
 * Evaluating the polynomial Pptr at the MPCI *Cptr.
 */
{
	int k;
	MPCI *B, *Aptr, *TempCI;
	POLYCI CURSOR;

	if (Pptr == NULL)
		return ZEROCI();
	else
		Aptr = COPYCI(Pptr->COEF);
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		if (CURSOR->NEXT == NULL)
			k = DEGREEPCI(CURSOR);
		else
			k = DEGREEPCI(CURSOR) - DEGREEPCI(CURSOR->NEXT);
		B = POWERCI(Cptr, (unsigned int)k);
		TempCI = Aptr;
		Aptr = MULTCI(Aptr, B);
		FREEMPCI(B);
		FREEMPCI(TempCI);
		CURSOR = CURSOR->NEXT;
		if (CURSOR != NULL)
		{
			TempCI = Aptr;
			Aptr = ADDCI(Aptr, CURSOR->COEF);
			FREEMPCI(TempCI);
		}
	}
	return (Aptr);
}

POLYCI DERIVPCI(POLYCI Pptr)
/*
 * returns the derivative of a polynomial.
 */
{
	POLYCI CURSOR, Sptr;
	MPI *TempI;

	if (DEGREEPCI(Pptr) == 0)
		return (NULL);
	else
	{
		Sptr = COPYPCI(Pptr);
  		CURSOR = Sptr;
		while (CURSOR != NULL)
		{
			TempI = (CURSOR->COEF)->R;
			(CURSOR->COEF)->R = MULT_I((CURSOR->COEF)->R, CURSOR->DEG);
			FREEMPI(TempI);
			TempI = (CURSOR->COEF)->I;
			(CURSOR->COEF)->I = MULT_I((CURSOR->COEF)->I, CURSOR->DEG);
			FREEMPI(TempI);
			CURSOR->DEG = CURSOR->DEG - 1;
			if (CURSOR->NEXT == NULL)
				break;
			else if ((CURSOR->NEXT)->DEG == 0)
			{
				DELETE_PCI(CURSOR->NEXT);
				CURSOR->NEXT = NULL;
				break;
			}
			else
				CURSOR = CURSOR->NEXT;
		}
		return (Sptr);
	}
}
